#include <dce/pthread.h>
#include <iostream.hxx>
#include "daerpc.h"
#include "types.hxx"
#include "rio_edescript.hxx"
#include "useful_macros.hxx"
#include "daerequest.hxx"
#include "dae1srequest.hxx"
#include "daerequestlist.hxx"
#include "riolibrary.hxx"
#include "daepacketheader.hxx"
#include "daepacket.hxx"
#include "dce_status.hxx"
#include <dce/dce_error.h>


#include "riolibrary_debug.hxx"

int RioLibrary::MAX_IO_WORDS = 2000;
int RioLibrary::MAX_MEM_WORDS = 2000;

#define CHECK_DAE_OPEN	if (!connected) { return_status(NO_DAE_OPEN); }

RioLibrary::RioLibrary()
{
    requestList = new (DaeRequestList);
    if (requestList == 0) { cerr << "Error creating request list in RioLibrary" << endl; }
    global_section = 0;
    currentDae = "";
    daeContext_h = NULL;
    connected = False;    
// Determine words to use
    const char *max_dae_words = getenv("MAX_DAE_WORDS");
    if (max_dae_words != NULL)
    {
	MAX_IO_WORDS = atoi(max_dae_words);
	MAX_MEM_WORDS = atoi(max_dae_words);
	cerr << "Found logical name MAX_DAE_WORDS --- using use supplied default" << endl;
	cerr << "Maximum number of 16 bit words per IO space request = " << MAX_IO_WORDS << endl;
	cerr << "Maximum number of 16 bit words per MEMORY request = " << MAX_MEM_WORDS << endl;
    }
}

RioLibrary::~RioLibrary()
{    
    if (connected)
    {
	cerr << "Riolibrary going out of scope while connected to " << currentDae << " --- calling closeDae()" << endl;
	closeDae();
    }
    delete requestList;
}

Status RioLibrary::sendRequests()
{
    Status status;
    DaePacket packet;
    DaeRequestPointer a_request;
    if (debug_on(DEBUG_G_RIO_REQUESTS)) { cdebug << "Sending " << requestList->count() << " requests\n"; }
    while(status.ok() && !requestList->isempty())    
    {
// Read from the front of list, so send in same order as added to list
	a_request = requestList->shift();
	packet.setContent(a_request);
        status = packet.transfer(daeBinding_h, daeContext_h);
	delete a_request;
    }
    if (!requestList->isempty()) 
    { 
	cerr << "RioLibrary::sendRequests to " << currentDae << " terminated with " 
             << requestList->count() << " requests remaining\n";
	clearRequests();
    }
    return status;
}


Status RioLibrary::useIO(fort_i4 bus_addr, fort_i2* buffer, fort_i4 nwrds, TransferMode direction)
{
    fort_i4 ntransfer;	// number of 16 bit words sent
    DaeRequestPointer request;
    while(nwrds > 0) 
    {
	ntransfer = (nwrds > MAX_IO_WORDS ? MAX_IO_WORDS : nwrds);
	request = new Dae1sRequest(direction, bus_addr, buffer, ntransfer, AccessIO);
	NEW_CHECK(request);
	requestList->push(request);
	if (debug_on(DEBUG_G_RIO_REQUESTS)) { cdebug << "IO space request for " << ntransfer*2 << " bytes added to list\n"; }
	buffer += ntransfer;
	nwrds -= ntransfer;
    }
    return Status();
}

Status RioLibrary::useMem(fort_i4 bus_addr, fort_i2* buffer, fort_i4 nwrds, TransferMode direction)
{
    fort_i4 ntransfer;	// number of 16 bit words sent
    DaeRequestPointer request;
    while(nwrds > 0)
    {
	ntransfer = (nwrds > MAX_MEM_WORDS ? MAX_MEM_WORDS : nwrds);
	request = new Dae1sRequest(direction, bus_addr, buffer, ntransfer, AccessMem);
	NEW_CHECK(request);
	requestList->push(request);
	if (debug_on(DEBUG_G_RIO_REQUESTS)) { cdebug << "Memory access request for " << ntransfer*2 << " bytes added to list\n"; }
	buffer += ntransfer;
	nwrds -= ntransfer;
	bus_addr += (ntransfer * sizeof(fort_i2));
    }
    return Status();
}

Status RioLibrary::readMem(fort_i4 bus_addr, fort_i2* buffer, fort_i4 nwrds)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_READ_MEM)) 
    { 
	cdebug << "Reading Memory: Address = " << bus_addr << ", Size = " << nwrds*sizeof(fort_i2) << " bytes ...\n"; 
    }
    Status status = useMem(bus_addr, buffer, nwrds, Receive);
    if (status.ok()) { status = sendRequests(); }
    return status;
}

Status RioLibrary::readIO(fort_i4 bus_addr, fort_i2 *buffer, fort_i4 nwrds)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_READ_IO)) 
    { 
	cdebug << "Reading IO: Address = " << bus_addr << ", Size = " << nwrds*sizeof(fort_i2) << " bytes ...\n"; 
    }
    Status status = useIO(bus_addr, buffer, nwrds, Receive);
    if (status.ok()) { status = sendRequests(); }
    return status;
}

Status RioLibrary::writeMem(fort_i4  bus_addr, fort_i2 *buffer, fort_i4 nwrds)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_WRITE_MEM)) 
    { 
	cdebug << "Writing Memory: Address = " << bus_addr << ", Size = " << nwrds*sizeof(fort_i2) << " bytes ...\n"; 
    }
    Status status = useMem(bus_addr, buffer, nwrds, Send);
    if (status.ok()) { status = sendRequests(); }
    return status;
}

Status RioLibrary::writeIO (fort_i4 bus_addr, fort_i2 *buffer, fort_i4 nwrds)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_WRITE_IO)) 
    { 
	cdebug << "Writing IO: Address = " << bus_addr << ", Size = " << nwrds*sizeof(fort_i2) << " bytes ...\n"; 
    }
    Status status = useIO(bus_addr, buffer, nwrds, Send);
    if (status.ok()) { status = sendRequests(); }
    return status;
}

Status RioLibrary::resetDae(fort_i4 nn)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_RESET)) { cdebug << "Resetting " << currentDae << " ...\n"; }
    if (nn != 3)
    {
	return Status(BAD_WORD_LENGTH, FCODE, nn);
    }
    DaeRequestPointer request = new Dae1sRequest(Reset, nn, 0, 0, AccessMem);
    NEW_CHECK(request);
// *** should error check this
    requestList->push(request);
    if (debug_on(DEBUG_G_RIO_REQUESTS)) { cdebug << "Reset request added to list\n"; }
    Status status = sendRequests();
    return status;
}

Status RioLibrary::openDae(const char *dae_name)
{
    if (connected) { return Status(A_DAE_OPEN, FCODE, dae_name, currentDae); }
    if (debug_on(DEBUG_G_RIO_OPEN)) { cdebug << "Opening connection to " << dae_name << " ...\n"; } 
    rpc_ns_handle_t import_context;
    error_status_t rpc_status;
    unsigned32 status; 
    boolean32 temp;

    rpc_ns_binding_import_begin(rpc_c_ns_syntax_default, (unsigned_char_t*)dae_name, daerpcif_v1_0_c_ifspec,
				NULL, &import_context, &status);
//    if (status != rpc_s_ok) return_DCE_status(status);
    RETURN_RPC_STATUS(status, "can't begin import");
    rpc_ns_binding_import_next(import_context, &daeBinding_h, &status);
    RETURN_RPC_STATUS(status, "can't do import");
    rpc_ns_binding_import_done(&import_context, &status);
    RETURN_RPC_STATUS(status, "can't end import");
    rpc_ep_resolve_binding(daeBinding_h, daerpcif_v1_0_c_ifspec, &status);
    RETURN_RPC_STATUS(status, "can't resolve binding");
    temp = rpc_mgmt_is_server_listening(daeBinding_h, &status);
    RETURN_RPC_STATUS(status, "can't check if server is listening");
    if (temp == FALSE)
    {
	return_status(SERVER_NOT_LISTENING);
    }
    daeContext_h = open_dae_connection(daeBinding_h, &rpc_status);
    RETURN_RPC_STATUS(rpc_status, "Error creating server context");
    if (daeContext_h == 0) { return_status(INVALID_CONTEXT_HANDLE); }
    connected = True;
    currentDae = dae_name;
    return Status();
}

Status RioLibrary::closeDae()
{
    CHECK_DAE_OPEN;
    unsigned32 status;
    error_status_t rpc_status;
    if (debug_on(DEBUG_G_RIO_CLOSE)) { cdebug << "Terminating connection to " << currentDae << " ...\n"; }
    if (!requestList->isempty())
    {
	cerr << "RioLibrary: " << requestList->count() << " unsent requests for " << currentDae << " being discarded\n";
    }
    clearRequests();
    connected = False;
    currentDae = "";
    free(global_section);
    global_section = 0;
    close_dae_connection(daeBinding_h, &daeContext_h, &rpc_status);
    if (daeContext_h != 0) { rpc_ss_destroy_client_context(&daeContext_h); }
    rpc_binding_free(&daeBinding_h, &status);
//    RETURN_RPC_STATUS(status, "can't free binding handle");
    return Status();
} 

Status RioLibrary::readMemVI(fort_i4 bus_addr, fort_i4* buffer, fort_i4 length, fort_i4 wrd_len)
{
    CHECK_DAE_OPEN;
    if (debug_on(DEBUG_G_RIO_READ_MEM)) 
    { 
	cdebug << "Reading Memory: Address = " << bus_addr << ", Size = " << length << " * " << wrd_len << " bytes ...\n"; 
    }
    if (wrd_len != 4)
    {
	return Status(BAD_WORD_LENGTH, FCODE, wrd_len);
    }
    fort_i4 nwrds = length * wrd_len / 2;
    fort_i2 *temp_buffer = new fort_i2[nwrds];
    NEW_CHECK(temp_buffer);
    Status status = readMem(bus_addr, temp_buffer, nwrds);
    if (!status.ok()) { delete []temp_buffer; return status; };
/*
 * note: these are copied with no sign extension
 */
    switch(wrd_len)
    {
	case 4:
	    {
	        fort_i4 *base = (fort_i4*) temp_buffer;
    	        for(int i=0; i<length; i++) { buffer[i] = base[i]; }
	    }
	    break;
	case 2:
	    {
                ufort_i2 *base = (ufort_i2*) temp_buffer;
    	        for(int i=0; i<length; i++) { buffer[i] = base[i]; }
	    }
	    break;
	case 1:
	    {
	        unsigned char *base = (unsigned char*) temp_buffer;
    	        for(int i=0; i<length; i++) { buffer[i] = base[i]; }
	    }
	    break;
	default:
	    break;
    }
    delete []temp_buffer;
    return status;
}

// fort_i4 start[size]
Status RioLibrary::readCrpt(fort_i4*& start, fort_i4 crpt_section, fort_i4& size_in_words)
{
    CHECK_DAE_OPEN;  
    free(global_section);
    global_section = 0;
    if (debug_on(DEBUG_G_RIO_READ_CRPT)) { cdebug << "Reading global section, part = " << crpt_section << "\n"; }
// Perform RPC call which also allocates storage
    error_status_t rpc_status = rpc_s_ok;
    get_global_section(daeBinding_h, crpt_section, &global_section, &rpc_status); 	
    RETURN_RPC_STATUS(rpc_status, "can't get global section");
    if (global_section == 0) { return Status(RPC_ERROR, FCODE, "get_global_section = 0"); }
    start = global_section->data;
    size_in_words = global_section->length;
    if (debug_on(DEBUG_G_RIO_READ_CRPT))
	{ cdebug << "Mapped CRPT section: address = " << (void*)start << ", length = " << size_in_words*4 << " bytes\n"; }
    return Status();
}

Status RioLibrary::clearRequests()
{
    while(!requestList->isempty())
    {
	delete requestList->pop();
    }
    requestList->reset();
    return Status();
}
