/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 * Module:	colourmap
 * Author:	Freddie Akeroyd, ISIS, 3/6/96
 * Purpose:	Give GENIE control over colours allocated on the display
 */
#include "geniedefs.h"
#include <stdlib.h>
#include <math.h>
#include <fstream.h>
#include "gnuregex.h"
#include "g_types.h"
#include "Pginclude.h"
#include "colourmap.h"
#if defined(__VMS) && __VMS_VER < 70000000
#include "strcasecmp.h"
#endif /* defined(__VMS) && __VMS_VER < 70000000 */

static const char* rcsid = "@(#)$Id: colourmap.cxx,v 1.11 1997/06/17 21:24:03 faa Exp $";
USE_RCSID(rcsid);

#define FROM_PGPLOT(__val)	(int)(GENIE_COLOUR_RANGE * __val + 0.5F)
#define TO_PGPLOT(__val)	(float)(__val / GENIE_COLOUR_RANGE)

ColourMap::ColourMap() : m_ci_low(0), m_ci_high(-1), m_map(0)
{
    initialize();
}

void ColourMap::initialize()
{
    pg_float r, g, b;
// determine colours available on device
    pgqcol(m_ci_low, m_ci_high);
    m_map = new Colour[m_ci_high + 1];
// Read in pre-assigned values, and mark first 16 as unavailable as they are pre-assigned by PGPLOT
    for(int i=m_ci_low; i<=m_ci_high; i++)
    {
	pgqcr(i, r, g, b);
	m_map[i].r = FROM_PGPLOT(r);
	m_map[i].g = FROM_PGPLOT(g);
	m_map[i].b = FROM_PGPLOT(b);
//	cout << i << " = " << m_map[i].r << endl;
	m_map[i].refcount = (i < 16 ? 1 : 0);
    }
}
	
// Free a colour index
int ColourMap::freeColour(int index)
{
    if (index >= 16 && index <= m_ci_high)
    {
	if (m_map[index].refcount == 0)
	{
	    gerr << "ColourMap: freeing unallocated colour " << index << endl;
	    return -1;
	}
	if (--m_map[index].refcount == 0)
	{
	    ;
#if 0
// Nice in principle, but likely that somebody will hardcopy a plot that is still on the screen
// and which they have just freed all the colours for! This way, the colour will get copied
// correctly to the new device
	    pgscr(index, 0.0F, 0.0F, 0.0F);	// for debugging purposes
	    m_map[index].r = 0;
	    m_map[index].g = 0;
	    m_map[index].b = 0;
#endif
	}
	return 0;
    }
    else
    {
	gerr << "ColourMap: freeing out of range colour " << index << endl;
	return -1;
    }
}

// Return index for a new colour, -1 if no more colours
// new index will be >= start if a new colour is allocated
int ColourMap::newRGBColour(pg_float r, pg_float g, pg_float b, int in_table, int start)
{
    int index = findIndex(r, g, b, in_table);
    if (index != -1)
    {
	m_map[index].refcount++;
	return index;
    }
    index = nextFreeIndex();
    if (index != -1)
    {
	pgscr(index, r, g, b);
	m_map[index].r = FROM_PGPLOT(r);
	m_map[index].g = FROM_PGPLOT(g);
	m_map[index].b = FROM_PGPLOT(b);
	m_map[index].refcount++;
	m_map[index].in_table = in_table;
    }
    return index;
}

// Return index for a new colour, -1 if no more colours
int ColourMap::newNamedColour(const char* name, int in_table, int start)
{
    String rgb_file_name;
    fstream rgb_file;
    if (getenv("PGPLOT_RGB") != 0)
    {
	rgb_file_name = getenv("PGPLOT_RGB");
	rgb_file.open(rgb_file_name, ios::in);
    }
    else if (getenv("PGPLOT_DIR") != 0)
    {
	cat(getenv("PGPLOT_DIR"), DIRECTORY_SEPARATOR, rgb_file_name);
	rgb_file_name += "rgb.txt";
	rgb_file.open(rgb_file_name, ios::in);
    }
    else
    {
	*g_err << "COLOUR: Cannot locate \"rgb.txt\" - PGPLOT_RGB and/or PGPLOT_DIR not defined" << endl;
	return -1;
    }
    if (!rgb_file)
    {
	*g_err << "COLOUR: No \"rgb.txt\" file available for named colours" << endl;
	ginf << "COLOUR: Are PGPLOT_RGB and/or PGPLOT_DIR defined correctly?" << endl;
	return -1;
    }
// File format is   R  G  B  Name   with R,B,G in range 0->255
    String fields[4];
    int i, found(0);
    char buffer[128];
    while(!found && rgb_file.good())
    {
	i = 0;
        rgb_file.getline(buffer, sizeof(buffer), '\n');
// skip leading spaces or else it thinks we have another field!
	while(buffer[i] == ' ' || buffer[i] == '\t')
	    ++i;
        split(buffer+i, fields, 4, RXwhite);
	if (strcasecmp(name, fields[3]) == 0)
	    found = 1;
    }
    rgb_file.close();
    if (!found)
    {
	*g_err << "COLOUR: No such colour as \"" << name << "\" in \"" << rgb_file_name << "\"" << endl;
	return -1;
    }
    return newRGBColour(atof(fields[0]) / 255.0, atof(fields[1]) / 255.0, atof(fields[2]) / 255.0, in_table, start);
}

int ColourMap::newHLSColour(pg_float h, pg_float l, pg_float s, int in_table, int start)
{
    pg_float r, g, b;
    grxrgb(h, l, s, r, g, b);
    return newRGBColour(r, g, b, in_table, start);
}

// see if we already have a colour - if so return index; if not, return -1
int ColourMap::findIndex(pg_float r, pg_float g, pg_float b, int in_table)
{
    int ir = FROM_PGPLOT(r);
    int ig = FROM_PGPLOT(g);
    int ib = FROM_PGPLOT(b);
    for(int i=m_ci_low; i<=m_ci_high; i++)
    {
	if (in_table == m_map[i].in_table && ir == m_map[i].r && ig == m_map[i].g && ib == m_map[i].b)
	{
	    return i;
	}
    }
    return -1;
}

// Allocate a colourtable (array of colour indexes) for given array of ncol RGB values
// return NULL if unable to allocate enough colour cells
// colourtable is allocated by new and must be delete[]ed by caller
// sets ncol to the number allocated
g_integer* ColourMap::newRGBColourTable(g_real* r, g_real* g, g_real* b, int len, 
					int& ncol, g_real contrast, g_real brightness)
{
    g_integer* table = new g_integer[ncol];
    int i = allocTable(table, ncol, r, g, b, len, contrast, brightness);
    if (i == ncol)
	return table;
    ginf << "ColourMap: insufficient colours available for table of " << ncol << endl;
    while(i > 0 && i != ncol)
    {
        ncol = i;
        freeColourTable(table, ncol);
        i = allocTable(table, ncol, r, g, b, len, contrast, brightness);
    }
    ginf << "ColourMap: created table using only " << ncol << endl;
    return table;
}
	

// Returns number of colours allocated, so if return value != ncol
// there were not enough available
// table[ncol] must be pre-allocated
int ColourMap::allocTable(g_integer* table, int ncol, const g_real* r, const g_real* g, const g_real* b, 
				int len, g_real contrast, g_real brightness)
{
	g_real colalf, col(0.0), dicol, rl, gl, bl; 
      	g_real nc = (g_real)(len - 1);
   	int icol, current_index(-1);
	g_real dcol = 1.0 / (g_real)(ncol);
	for(int i=0; i<ncol; i++)
	{
	    colalf = nc * pow(col, contrast);
	    if (colalf >= nc)
		colalf = nc * 0.99999;
	    icol = (int)colalf;
	    dicol = colalf - (float)icol;
	    rl = r[icol] + dicol * (r[icol+1] - r[icol]);
	    gl = g[icol] + dicol * (g[icol+1] - g[icol]);
	    bl = b[icol] + dicol * (b[icol+1] - b[icol]);
	    current_index = newRGBColour((float)rl, (float)gl, (float)bl, 1, current_index);
	    if (current_index == -1)
		return i;
	    table[i] = current_index;
	    col += dcol;
	}
	return ncol;
}

// Mark cells of a previous allocated colourtable as available
int ColourMap::freeColourTable(g_integer* table, int ncol)
{
    for(int i=0; i<ncol; i++)
    {
	freeColour(table[i]);
    }
    return 0;
}

// Find next free index after "start", return -1 if no index is available
int ColourMap::nextFreeIndex(pg_int start)
{
    if (start < m_ci_low)
    {
	start = m_ci_low;
    }
    for(int i=start; i<=m_ci_high; i++)
    {
	if (m_map[i].refcount == 0)
	{
	    return i;
	}
    }
    return -1;
}

// Return number of available colours
int ColourMap::numFreeColours()
{
    int nfree(0);
    for(int i=m_ci_low; i<=m_ci_high; i++)
    {
	if (m_map[i].refcount == 0)
	{
	    nfree++;
	}
    }
    return nfree;
}

// Copy allocated colours across
void ColourMap::copyColours(const ColourMap& colour_map)
{
    int start = (m_ci_low > colour_map.m_ci_low ? m_ci_low : colour_map.m_ci_low);
    int end = (m_ci_high < colour_map.m_ci_high ? m_ci_high : colour_map.m_ci_high);
    for(int i=start; i<=end; i++)
    {
	m_map[i] = colour_map.m_map[i];
    }
}

ColourMap::ColourMap(const ColourMap* pcm) : m_ci_low(0), m_ci_high(-1), m_map(0)
{
    initialize();
    if (pcm != 0)
    {
        copyColours(*pcm);
    }
}

ColourMap::ColourMap(const ColourMap& cm) : m_ci_low(0), m_ci_high(-1), m_map(0)
{
    initialize();
    copyColours(cm);
}

void ColourMap::operator=(const ColourMap& colour_map)
{
    if (this != &colour_map)
    {
        m_ci_low = colour_map.m_ci_low;
        m_ci_high = colour_map.m_ci_high;
        int n = m_ci_high + 1;
	delete[] m_map;
        m_map = new Colour[n];
        memcpy(m_map, colour_map.m_map, n*sizeof(Colour));
    }
}

void ColourMap::install()
{
//    for(pg_int i=m_ci_low; i<=m_ci_high; i++)
    for(pg_int i=16; i<=m_ci_high; i++)
    {
// Probably should be > 0, but we may have been copied from a freed colourmap that
// somebody is just about to do a hardcopy on ...
	if (m_map[i].refcount >= 0)
	{
	    pgscr(i, TO_PGPLOT(m_map[i].r), TO_PGPLOT(m_map[i].g), 
			TO_PGPLOT(m_map[i].b));
	}
    }
    return;
}

// Print out allocated colours
ostream& ColourMap::print(ostream& os)
{
    os << "ColourMap contains " << m_ci_high - m_ci_low + 1 << " colours, with indices from " 
       				<< m_ci_low << " to " << m_ci_high << endl;
    os << numFreeColours() << " unallocated colours available" << endl;
    for(int i=m_ci_low; i<=m_ci_high; i++)
    {
	if (m_map[i].refcount > 0)
	{
	    os << i << " = RGB(" << m_map[i].r << ", " << m_map[i].g << ", " << m_map[i].b << ")";
	    os << (m_map[i].in_table ? " [table]" : "") << endl;
	}
    }
    return os;
}

ColourMap::~ColourMap()
{
    delete []m_map;
}

// return reference count (>=0) on success, -1 on error
int ColourMap::queryColour(int index, g_real& r, g_real& g, g_real& b)
{
    if (index >= m_ci_low && index <= m_ci_high)
    {
	r = TO_PGPLOT(m_map[index].r);
	g = TO_PGPLOT(m_map[index].g);
	b = TO_PGPLOT(m_map[index].b);
	return m_map[index].refcount;
    }
    else
    {
	*g_err << "COLOURMAP: Invalid colour index specified = " << index << endl;
	ginf << "COLOURMAP: Valid index range is " << m_ci_low << " to " << m_ci_high << endl;
	return -1;
    }
}
