/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <math.h>
#include "g_types.h"
#include "Pictlist.h"
#include "Pginclude.h"
#include "Genie_functions.h"
#include "contour_plot.h"
#include "contour_labels.h"
#include "pg_coord_trans_func.h"

static const char* rcsid = "@(#)$Id: draw_contour.cxx,v 1.15 1997/04/08 16:32:31 cmm Exp $";
USE_RCSID(rcsid);

static ContourPlot *last_contour_plot = NULL;	// last contour plot drawn (used by labels)
static int last_contour_picture = -1;		// picture containing "last_contour_plot"

static void finish_draw_contour(Drawable_objects *PTR, g_objectid objid)
{
        
    objid[ObjectPosition] = PicturePTR[Current_Picture]->disp_add(PTR);
    objid[ObjectPicture] = Current_Picture;
    last_contour_plot = (ContourPlot*)PTR;		// make a note to allow future labelling
    last_contour_picture = Current_Picture;
}

// Return 0 if all OK, -1 on error
int new_contour(g_real* values, 
		int nx_values, 
		int ny_values, 
		g_real* tr, 
		g_real* contours, 
		int ncont,
	      	int line_style, 
		int line_colour, 
		g_real line_thickness,
		g_objectid objid)  
{
    PGCoordinateTransformFunction* ctf = new PGCoordinateTransformFunction(tr); 
    if (ctf == 0)
	return -1;
    ContourPlot *PTR = new ContourPlot(values, nx_values, ny_values, contours, 
					ncont, ctf, line_style, line_colour, 
					(float)line_thickness);
    if (PTR == 0)
	return -1;
    finish_draw_contour(PTR, objid);
    return 0;
}


void draw_contour(g_real* values,
                int nx_values,
                int ny_values,
                g_real* tr,
                g_real* contours,
                int ncont,
                int line_style,
                int line_colour,
                g_real line_thickness)  
{
    g_integer objid[GENIE_OBJECTID_SIZE];
    if (new_contour(values, nx_values, ny_values, tr, contours, ncont, line_style,
		line_colour, line_thickness, objid) == 0)
    {
        display_object(objid);
    }
}

// Return 0 on success, -1 on failure
int new_contour_array(g_real* values, 
			int nx_values, 
			int ny_values, 
			g_real* wx, 
			int* wx_dims, 
			int wx_ndims, 
			g_real* wy, 
			int* wy_dims, 
			int wy_ndims, 
			g_real* contours, 
			int ncont, 
			int line_style, 
			int line_colour, 
			g_real line_thickness,
			g_objectid objid)  
{
    CoordinateTransformMode x_mode, y_mode;
    x_mode = (wx_dims[0] > nx_values ? BinMode : PointMode);
    y_mode = (wy_dims[wy_ndims-1] > ny_values ? BinMode : PointMode);
    
    PGCoordinateTransformFunction* ctf = 
		new PGCoordinateTransformFunction(wx, wx_dims, wx_ndims, x_mode,
						 wy, wy_dims, wy_ndims, y_mode);
    if (ctf == 0)
	return -1;
    ContourPlot *PTR = new ContourPlot(values, nx_values, ny_values, contours, 
					ncont, ctf, line_style, line_colour, (float)line_thickness);
    if (PTR == 0)
	return -1;
    finish_draw_contour(PTR, objid);
    return 0;
}

void draw_contour_array(g_real* values,
                        int nx_values,
                        int ny_values,
                        g_real* wx,
                        int* wx_dims,
                        int wx_ndims,
                        g_real* wy,
                        int* wy_dims,
                        int wy_ndims,
                        g_real* contours,
                        int ncont,
                        int line_style,
                        int line_colour,
                        g_real line_thickness)  
{
    g_integer objid[GENIE_OBJECTID_SIZE];
    if (new_contour_array(values, nx_values, ny_values, wx, wx_dims, wx_ndims, wy, 
		      wy_dims, wy_ndims, contours, ncont, line_style, line_colour,
		      line_thickness, objid) == 0)
    {
        display_object(objid);
    }
}

// return 0 on success, -1 on failure
int new_contour_function(g_real* values, 
			int nx_values, 
			int ny_values, 
			void* func_ptr, 
			g_real* contours, 
			int ncont,
		        int line_style, 
			int line_colour, 
			g_real line_thickness,
			g_objectid objid)  {

    PGCoordinateTransformFunction* ctf = 
		new PGCoordinateTransformFunction((coord_tran_func_t)func_ptr);
    if (ctf == 0)
	return -1;
    
    ContourPlot *PTR = new ContourPlot(values, nx_values, ny_values, contours, 
					ncont, ctf, line_style, line_colour, 
					(float)line_thickness);
    if (PTR == 0)
	return -1;
    finish_draw_contour(PTR, objid);
    return 0;
}

void draw_contour_function(g_real* values,
                        int nx_values,
                        int ny_values,
                        void* func_ptr,
                        g_real* contours,
                        int ncont,
                        int line_style,
                        int line_colour,
                        g_real line_thickness)  {

    g_integer objid[GENIE_OBJECTID_SIZE];
    
    if (new_contour_function(values, nx_values, ny_values, func_ptr, contours,
			 ncont, line_style, line_colour, line_thickness,
			 objid) == 0)
    {
        display_object(objid);
    }
}


void new_label_contour(const char* label, 
			int start, 
			int end, 
			int step, 
			int intval, 
			int minint, 
			int colour, 
			int font, 
			g_real height, 
			g_real line_width,
			g_objectid contourid,
			g_objectid objid)  
{
    ContourPlot *PTR = (ContourPlot*)PicturePTR[contourid[ObjectPicture]]->objectPointer(contourid[ObjectPosition]);    

    ContourLabels *lab = new ContourLabels(PTR, label, start, end, step, intval, minint, colour,
			font, (float)height, (float)line_width);

    objid[ObjectPosition] = PicturePTR[Current_Picture]->disp_add(lab);
    objid[ObjectPicture] = Current_Picture;
}

void draw_label_contour(const char* label,
                        int start,
                        int end,
                        int step,
                        int intval,
                        int minint,
                        int colour,
                        int font,
                        g_real height,
                        g_real line_width,
			g_objectid contourid)  
{
    g_integer objid[GENIE_OBJECTID_SIZE];
    new_label_contour(label, start, end, step, intval, minint, colour,
		      font, height, line_width, contourid, objid);
    display_object(objid);
}
 
void label_contour(const char* label, int start, int end, int step, int intval,
		int minint, int colour, int font, g_real height, g_real line_width)
{
    g_integer contourid[GENIE_OBJECTID_SIZE];
    if (last_contour_plot == 0 || last_contour_picture != Current_Picture)
    {
        error_add("label_contour", "No valid contour plot in current picture", "");
        return;
    }
    contourid[ObjectPicture] = last_contour_picture;
    contourid[ObjectPosition] = PicturePTR[last_contour_picture]->
		objectNumber(last_contour_plot);

    draw_label_contour(label, start, end, step, intval,
		minint, colour, font, height, line_width, contourid);
}
/*
 * itype = 0 -> linear contours
 * itype = 1 -> log contours
 * itype = 2 -> sqrt contours
 */

g_real* get_contour_levels(g_real* values, int nvalues, int ncont, int itype)
{
    static const double scale_log = 65000.0;	// PGPLOT uses the same value for LOG scaling in PGIMAG
    if (ncont <= 0)
    {
	error_add("get_contour_levels", "Invalid number of contours", "");
	return 0;
    }
    g_real val_max = -FLT_MAX, val_min = FLT_MAX;
    for(int i=0; i<nvalues; i++)
    {
	if (!IS_UNDEFINED_REAL(values[i]))
	{
	    if (values[i] > val_max)
	        val_max = values[i];
	    if (values[i] < val_min)
	        val_min = values[i];
	}
    }
    g_real* contours = new g_real[ncont];
// Algorithm: for 3 contours in range 0 - 3, would put them at 0.5 1.5 2.5
    g_real step;
    if (itype == 0)
    {
        step = (val_max - val_min) / (g_real)ncont;
    }
    else if (itype == 1)
    {
	step = log10(1.0 + scale_log * fabs(val_max - val_min)) / (g_real)ncont;
    }
    else if (itype == 2)
    {
	step = sqrt(fabs(val_max - val_min)) / (g_real)ncont;
    }
    g_real offset = step / 2.0;
    g_real temp;
    for(int j=0; j<ncont; j++)
    {
	temp = offset + (g_real)j * step;
	if (itype == 0)
	{
	    contours[j] = val_min + temp;
	}
	else if (itype == 1)		// log contours
	{
	    contours[j] = val_min + (pow(10.0, temp) - 1.0) / scale_log;
	}
	else if (itype == 2)		// sqrt contours
	{
	    contours[j] = val_min + temp * temp;
	}
    }
#if 0
    gerr << "Max and min: " << val_max << " " << val_min << endl;
    for(int k=0; k<ncont; k++)
    {
	gerr << contours[k] << endl;
    }
#endif
    return contours;
}

//
// Calculate a TR matrix so that the (nx,ny) grid is mapped to the
// current world coordinates of the window
//
// wx[i][j] = tr[0] + tr[1] * i + tr[2] * j
// wy[i][j] = tr[3] + tr[4] * i + tr[5] * j
//
// However, i and j start at 1 (FORTRAN)
//
g_real* get_contour_tr_matrix(int nx, int ny)
{
    pg_float xlow, xhigh, ylow, yhigh;
    pgqwin(xlow, xhigh, ylow, yhigh);
    g_real xstep = (xhigh - xlow) / (g_real)(nx-1);
    g_real ystep = (yhigh - ylow) / (g_real)(ny-1);
    g_real* tr = new g_real[6];
    if (tr != 0)
    {
    	tr[0] = xlow - xstep;
    	tr[1] = xstep;
    	tr[2] = 0.0;
    	tr[3] = ylow - ystep;
    	tr[4] = 0.0;
    	tr[5] = ystep;
    }
    return tr;
}
