/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <math.h>
#include "g_types.h"
#include "Pginclude.h"
#include "drawable_values.h"

static const char* rcsid = "@(#)$Id: drawable_values.cxx,v 1.11 1997/05/14 14:07:40 faa Exp $";
USE_RCSID(rcsid);

DrawableValues::DrawableValues(float* values, int dim1, ArrayStyle values_style, ArrayStyle dims_style)
{
    setup(values, &dim1, 1, values_style, dims_style);
}

DrawableValues::DrawableValues(float* values, int dim1, int dim2, ArrayStyle values_style, ArrayStyle dims_style)
{
    int dims_array[2] = { dim1, dim2 };
    setup(values, dims_array, 2, values_style, dims_style);
}

DrawableValues::DrawableValues(float* values, int* dims_array, int ndims, ArrayStyle values_style, ArrayStyle dims_style)
{
    setup(values, dims_array, ndims, values_style, dims_style);
}

DrawableValues::DrawableValues(double* values, int dim1, ArrayStyle values_style, ArrayStyle dims_style)
{
    setup(values, &dim1, 1, values_style, dims_style);
}

DrawableValues::DrawableValues(double* values, int dim1, int dim2, ArrayStyle values_style, ArrayStyle dims_style)
{
    int dims_array[2] = { dim1, dim2 };
    setup(values, dims_array, 2, values_style, dims_style);
}

DrawableValues::DrawableValues(double* values, int* dims_array, int ndims, ArrayStyle values_style, ArrayStyle dims_style)
{
    setup(values, dims_array, ndims, values_style, dims_style);
}

// dims_style is how we want to store m_values internally
int DrawableValues::setupDims(int* dims_array, int ndims, ArrayStyle dims_style)
{
    m_dims_style = dims_style;
    m_ndims = ndims;
    m_npoints = 1;
    int i, j;
    for(i=0; i<ndims; i++)
    {
	m_npoints *= dims_array[i];
    }
    m_dims_array = new int[ndims];
    memcpy(m_dims_array, dims_array, ndims*sizeof(int));
    int dope = 1;
    m_dope_array = new int[ndims];
    for(i=ndims; i>0; i--)
    {
	if (dims_style == C)
	    j = i - 1;
	else // Fortran
	    j = ndims - i;
	m_dope_array[j]  = dope;
	dope *= dims_array[j];
    }
    return 0;
}

int DrawableValues::setup(float* values, int* dims_array, int ndims, ArrayStyle values_style, ArrayStyle dims_style)
{
    setupDims(dims_array, ndims, dims_style);
    m_values = duplicate(values, dims_array, ndims, values_style, dims_style);
    return 0;
}

int DrawableValues::setup(double* values, int* dims_array, int ndims, ArrayStyle values_style, ArrayStyle dims_style)
{
    setupDims(dims_array, ndims);
    m_values = duplicate(values, dims_array, ndims, values_style, dims_style);
    return 0;
}
	
// Return value at arbitrary subscripts e,g, for array[2][3][4] send array {2, 3, 4}
pg_float& DrawableValues::operator()(int* pos, int npos)
{
    int j(0);
    for(int i=0; i<npos; i++)
    {
	j += pos[i] * m_dope_array[i];
    }
    return m_values[j];
}

DrawableValues::~DrawableValues()
{
    delete []m_values;
    delete []m_dims_array;
    delete []m_dope_array;
}

pg_float* DrawableValues::duplicate(const float* old_array, int n)
{
    pg_float* new_array = new pg_float[n];
    if (new_array == 0)
    {
	return 0;
    }
    if (sizeof(pg_float) == sizeof(float))
    {
    	memcpy(new_array, old_array, n*sizeof(pg_float));
    }
    else		// pg_float must be "double"
    {
	for(int i=0; i<n; i++)
	{
	    new_array[i] = (pg_float)old_array[i];
	}
    }
    return new_array;
}

pg_float* DrawableValues::duplicate(const double* old_array, int n)
{
    pg_float* new_array = new pg_float[n];
    if (new_array == 0)
    {
	return 0;
    }
    if (sizeof(pg_float) == sizeof(double))
    {
    	memcpy(new_array, old_array, n*sizeof(pg_float));
    }
    else		// pg_float must be "float"
    {
	for(int i=0; i<n; i++)
	{
	    if (old_array[i] >= FLT_MAX || old_array[i] <= -FLT_MAX)
	    {
		new_array[i] = UNDEFINED_PG_FLOAT;
	    }
//	    else if (old_array[i] < -FLT_MAX)
//	    {
//		new_array[i] = -FLT_MAX;
//	    }
	    else
	    {
	        new_array[i] = (pg_float)old_array[i];
	    }
  	}
    }
    return new_array;
}

pg_float* DrawableValues::duplicate(const float* old_array, int* old_dims_array, int old_ndims, 
					ArrayStyle old_dims_style, ArrayStyle new_dims_style)
{
    pg_float* array = 0;
    int npoints = points(old_dims_array, old_ndims);
    if ((old_ndims == 1) || (old_dims_style == new_dims_style))
    {
	return duplicate(old_array, npoints);
    }
    if (old_ndims == 2) // Swap rows and columns
    {
	array = new pg_float[npoints];
	unsigned count(0);
	for(int i=0; i<old_dims_array[0]; i++)
	{
	    for(int j=0; j<old_dims_array[1]; j++)
	    {
 		if (old_dims_style == Fortran)
		{
		    array[count] = old_array[i + j * old_dims_array[0]];
		}
		else	// Old array is a C array and we want to turn it into a Fortran array
		{
		    array[i + j * old_dims_array[0]] = old_array[count];
		}
		++count;
	    }
	}
    }
    else
    {
	gerr << "Duplicate: cannot handle arrays of dimension " << old_ndims << endl;
    }
    return array;
}

pg_float* DrawableValues::duplicate(const double* old_array, int* old_dims_array, int old_ndims, 
					ArrayStyle old_dims_style, ArrayStyle new_dims_style)
{
    pg_float* array = 0 ;
    double temp;
    int npoints = points(old_dims_array, old_ndims);
    if ((old_ndims == 1) || (old_dims_style == new_dims_style))
    {
	return duplicate(old_array, npoints);
    }
    if (old_ndims == 2) // Swap rows and columns
    {
	array = new pg_float[npoints];
	unsigned count(0);
	for(int i=0; i<old_dims_array[0]; i++)
	{
	    for(int j=0; j<old_dims_array[1]; j++)
	    {
 		if (old_dims_style == Fortran)
		{
		    temp = old_array[i + j * old_dims_array[0]];
// Genie undefinedes come across as DBL_MAX, which is too large for a float - we must trap these
		    if (temp >= FLT_MAX)
			array[count] = UNDEFINED_PG_FLOAT;
		    else
			array[count] = (float)temp;
		}
		else	// Old array is a C array and we want to turn it into a Fortran array
		{
		    temp = old_array[count];
		    if (temp >= FLT_MAX)
			array[i + j * old_dims_array[0]] = UNDEFINED_PG_FLOAT;
		    else
			array[i + j * old_dims_array[0]] = (float)temp;
		}
	   	++count;
	    }
	}
    }
    else
    {
	gerr << "Duplicate: cannot handle arrays of dimension " << old_ndims << endl;
    }
    return array;
}

int DrawableValues::points(int* dims_array, int ndims)
{
    int npoints(1);
    for(int i=0; i<ndims; i++)
    {
	npoints *= dims_array[i];
    }
    return npoints;
}

DrawableValues DrawableValues::slice(int dim, double val) const
{
    if (m_ndims != 2)
    {
	gerr << "Error in DrawableValues::slice" << endl;
	return DrawableValues();
    }
    int size = m_npoints / m_dims_array[dim];
    pg_float* res = new pg_float[size];
    int low = (int)floor(val), high = (int)ceil(val);
    double alpha = val - (double)low;
    if (0 == dim)
    {
	for(int i=0; i<m_dims_array[1]; i++)
	{
	    if ( (*this)(low, i) == UNDEFINED_PG_FLOAT || (*this)(high, i) == UNDEFINED_PG_FLOAT )
		res[i] = UNDEFINED_PG_FLOAT;
	    else
    	        res[i] = (*this)(low, i) + alpha * ((*this)(high, i) - (*this)(low, i));
	}
    }
    else if (1 == dim)
    {
	for(int i=0; i<m_dims_array[0]; i++)
	{
	    if ( (*this)(i, low) == UNDEFINED_PG_FLOAT || (*this)(i, high) == UNDEFINED_PG_FLOAT )
		res[i] = UNDEFINED_PG_FLOAT;
	    else
    	        res[i] = (*this)(i, low) + alpha * ((*this)(i, high) - (*this)(i, low));
	}
    }
    else
	return DrawableValues();
    DrawableValues resy(res, size);
    delete[] res;
    return resy;
}
