/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include "smalltalk.h"
#include "smalltalk_object.h"
#include "g_fileaccess.h"
#include "genievars.h"
#include "fstream_list.h"
#include "read_ascii.h"

static fstream_list open_streams;

static OOP asciiFileOpen(SmalltalkObject& asciiFile);
static OOP asciiFileClose(SmalltalkObject& asciiFile);
static OOP asciiFileSkip(SmalltalkObject& asciiFile, g_integer nskip);
static OOP asciiFileRewind(SmalltalkObject& asciiFile);
static OOP asciiFileLines(SmalltalkObject& asciiFile);
static OOP asciiFileReadFixedFormat(SmalltalkObject& asciiFile, String& fields, String& format, g_integer count);
static OOP asciiFileReadFreeFormat(SmalltalkObject& asciiFile, String& fields, String& separator, g_integer count);
static OOP asciiFileWriteFreeFormat(SmalltalkObject& asciiFile, SmalltalkObject& aVariable, String& separator, String& user_comment);
static OOP asciiFileWriteFixedFormat(SmalltalkObject& asciiFile, SmalltalkObject& aVariable, String& format, String& user_comment);

// Get "handle" and "fs" objects
// '\002' is used to indicate no comment character
#define GET_HANDLE_AND_STREAM \
    GenieInteger handle(asciiFile("handle")); \
    if ((g_integer)handle < 0 || (g_integer)handle >= open_streams.count()) \
    { \
	gerr << "ASCIIFILE: Invalid handle" << endl; \
	return nilOOP; \
    } \
    fstream* fs = open_streams[handle]; \
    if (fs == 0) \
    { \
	gerr << "ASCIIFILE: Stream number " << (g_integer)handle << " is invalid" << endl; \
	return nilOOP; \
    } \
    GenieString comment(asciiFile("comment")); \
    GenieString delimiter(asciiFile("delimiter"));

OOP g_fileAccess(FileAccessCode code, OOP arguments)
{
    BEGINSAVEBLOCK

    OOP result(nilOOP);
    SmalltalkObject asciiFile(arguments, 1);

    switch(code)
    {
	case gxAsciiFileOpen:
	    result = asciiFileOpen(asciiFile);
	    break;

    	case gxAsciiFileRewind:
	    result = asciiFileRewind(asciiFile);
	    break;

    	case gxAsciiFileClose:
	    result = asciiFileClose(asciiFile);
	    break;

    	case gxAsciiFileSkip:
	{
	    GenieInteger nskip(arguments, 2);
	    result = asciiFileSkip(asciiFile, nskip);
	    break;
	}

    	case gxAsciiFileLines:
	    result = asciiFileLines(asciiFile);
	    break;

    	case gxAsciiFileReadFreeFormat:
  	{
	    GenieString fields(arguments, 2);
	    GenieString separator(arguments, 3);
	    GenieInteger count(arguments, 4);
	    result = asciiFileReadFreeFormat(asciiFile, fields, separator, count);
	    break;
	}

    	case gxAsciiFileReadFixedFormat:
  	{
	    GenieString fields(arguments, 2);
	    GenieString format(arguments, 3);
	    GenieInteger count(arguments, 4);
	    result = asciiFileReadFixedFormat(asciiFile, fields, format, count);
	    break;
	}

    	case gxAsciiFileWriteFreeFormat:
        {
	    SmalltalkObject aVariable(arguments, 2);	// Array
	    GenieString separator(arguments, 3);
	    GenieString user_comment(arguments, 4);
	    result = asciiFileWriteFreeFormat(asciiFile, aVariable, separator, user_comment);
	    break;
	}

	default:
	    gerr << "ASCIIFILE: Invalid callout number = " << code << endl;
	    break;
    }
    ENDSAVEBLOCK
    return result;
}

static OOP asciiFileOpen(SmalltalkObject& asciiFile)
{
    GenieString file(asciiFile("file"));
    fstream* fs = new fstream;
    int fd;
// Try for read/write initially
#ifdef __VMS
    fd = open(file, O_RDWR|O_CREAT, 0644, "shr=get");
#else
    fd = open(file, O_RDWR|O_CREAT, 0644);
#endif /* __VMS */
    if (fd == -1)
    {
	ginf << "ASCIIFILE: Cannot open READ/WRITE - trying READONLY" << endl;
#ifdef __VMS
	fd = open(file, O_RDONLY, 0, "shr=put");
#else
	fd = open(file, O_RDONLY, 0);
#endif /* __VMS */
    }
    GenieInteger handle;
    if (fd != -1)
    {
        fs->attach(fd);
	open_streams.push(fs);
	handle = open_streams.count() - 1;
    }
    else
    {
	gerr << "ASCIIFILE: Error opening \"" << file << "\"" << endl;
	gerr << "ASCIIFILE: " << strerror(errno) << endl;
	handle = -1;
    }
    asciiFile("handle", handle);
    return (OOP)asciiFile;
}

static OOP asciiFileClose(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
    fs->close();
    delete fs;
    open_streams[handle] = 0; 
    asciiFile("handle", GenieInteger(-1));
    return asciiFile;
}

static OOP asciiFileRewind(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
    fs->clear();		// set stream state to good
    fs->seekg(0, ios::beg);
    if (!fs->good())
	*g_err << "ASCIIFILE: Error rewinding stream" << endl;
    return asciiFile;
}

static OOP asciiFileLines(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
    GenieInteger nlines;
    nlines = lines_in_file(*fs, comment[0], delimiter[0]);
    if (!fs->good())
	*g_err << "ASCIIFILE: Error computing lines in file" << endl;
    return (OOP)nlines;
}

// Tries to skip "nskip" lines, returns actual number skipped
static OOP asciiFileSkip(SmalltalkObject& asciiFile, g_integer nskip)
{
    GET_HANDLE_AND_STREAM;
    int nlines(0);
    while(nskip > 0 && skip_comments(*fs, comment[0], delimiter[0]))
    {
	fs->ignore(INT_MAX, delimiter[0]);
	nlines++;
	nskip--;
    }
    if (!fs->good())
	*g_err << "ASCIIFILE: Error skipping stream" << endl;
    return (OOP)GenieInteger(nlines);
}


enum ItemType { UnknownItem, GenieIntegerItem, IntegerItem, 
		LongIntegerItem, GenieRealItem, FloatItem, 
		DoubleItem, GenieStringItem, StringItem };

static OOP asciiFileReadFixedFormat(SmalltalkObject& asciiFile, String& field_names,
					String& format, g_integer count)
{
#define MAXFIELDS	10	/* max items in an input line */
    GET_HANDLE_AND_STREAM;
    if (count == -1)
	count = lines_in_file(*fs, comment[0], delimiter[0]);
// "field_names" is a comma separated list of names to call the workspace fields 
    int i, j, nitems;
    String field_labels[MAXFIELDS];
    nitems = split(field_names, field_labels, MAXFIELDS, ',');
    ItemType field_type[MAXFIELDS];
    size_t field_size[MAXFIELDS];
    char buffer[SHORTCHARBUFF];
    void* result_array[MAXFIELDS];
    for(i=0; i<MAXFIELDS; i++)
    {
	result_array[i] = 0;
	field_type[i] = UnknownItem;
	field_size[i] = 0;
    }
// Now parse line to see what type of data we have
// work out field types
// first work out a reduced format string
    const char* fmt_pointer;
    i = 0;
    int in_format = 0;
    char* _format = new char[1+strlen(format)];
    for(fmt_pointer = format; *fmt_pointer != '\0'; fmt_pointer++)
    {
	if (!in_format && *fmt_pointer != '%')
	{
	    continue;
	}
	if (*fmt_pointer == '%' && *(fmt_pointer + 1) == '%')
	{
	    fmt_pointer++;
	    continue;
	}
	if (*fmt_pointer == '%' && *(fmt_pointer+1) != '*' && !in_format)
	{
	    in_format = 1;
	    _format[i++] = *fmt_pointer;
	    continue;
	}
	if (in_format && isalpha(*fmt_pointer))
	{
	    _format[i++] = *fmt_pointer;
	}
	if (strchr("diounxefgsc", *fmt_pointer) != 0)
	{
	    in_format = 0;
	}
    }
    _format[i] = '\0';
    String format_names[MAXFIELDS+1];
    int nfields = split(_format, format_names, MAXFIELDS+1, '%');
// as input string is e.g. %f%d we will get a split into
// three parts with the first part null; hence must decrement nfields
    nfields--;
    if (nfields != nitems)
    {
	*g_err << "ASCIIFILE: Inconsistency - nfields = " << nfields << ", nitems = " << nitems << endl;
        *g_err << "ASCIIFILE: field names = \"" << field_names << "\"" << endl;
        *g_err << "ASCIIFILE: item format = \"" << format << "\"" << endl;
        *g_err << "ASCIIFILE: reduced format = \"" << _format << "\"" << endl;
	ginf << "ASCIIFILE: Your format string must match up with all items on a line" << endl;
	ginf << "ASCIIFILE: If you wish to ignore something, use the %* notation e.g. %*f to ignore a float" << endl;
	return nilOOP;
    }
    delete[] _format;
    Regex integer_item("[diounx]", 1),  float_item("[efg]", 1), 
	string_item("[sc]", 1), long_integer_item("l[diounx]"), 
	double_item("l[efg]");
    for(i=0; i<nfields; i++)
    {
	if (format_names[i+1].matches(long_integer_item))
	{
	    field_type[i] = LongIntegerItem;
	    result_array[i] = new long[count];
	    field_size[i] = sizeof(long) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(integer_item))
	{
	    field_type[i] = IntegerItem;
	    result_array[i] = new int[count];
	    field_size[i] = sizeof(int) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(double_item))
	{
	    field_type[i] = DoubleItem;
	    result_array[i] = new double[count];
	    field_size[i] = sizeof(double) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(float_item))
	{
	    field_type[i] = FloatItem;
	    result_array[i] = new float[count];
	    field_size[i] = sizeof(float) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(string_item))
	{
	    field_type[i] = StringItem;
	    result_array[i] = new char[512*count];
	    field_size[i] = 512;
// This makes sure any string we write is NULL terminated - the %c in
// sscanf() does not ensure this
	    memset(result_array[i], '\0', 512*count);
	    continue;
	}
	*g_err << "ASCIIFILE: unknown format conversion \"" << format_names[i+1] << "\"" << endl;
	return nilOOP;
    }
// Now go back to beginning and read whole file
    for(i=0; i<count; i++)
    {
        if (skip_comments(*fs, comment[0], delimiter[0]))
        {
            fs->getline(buffer, sizeof(buffer), delimiter[0]);
        }
        if (fs->good())
	{
// need to use all of result_array[MAXFIELDS] in call
	    if ((j = sscanf(buffer, format, 
		(char*)result_array[0] + i * field_size[0],
		(char*)result_array[1] + i * field_size[1],
		(char*)result_array[2] + i * field_size[2],
		(char*)result_array[3] + i * field_size[3],
		(char*)result_array[4] + i * field_size[4],
		(char*)result_array[5] + i * field_size[5],
		(char*)result_array[6] + i * field_size[6],
		(char*)result_array[7] + i * field_size[7],
		(char*)result_array[8] + i * field_size[8],
		(char*)result_array[9] + i * field_size[9]
	    )) != nfields)
	    {
	 	*g_err << "ASCIIFILE: sscanf() can only match " << j << " items on line" << endl;
		*g_err << "ASCIIFILE: please check format \"" << format << "\" is correct" << endl;
	        return nilOOP;
	    }
	}
	else
	{
	    *g_err << "ASCIIFILE: Error in input stream when reading data line" << i << endl;
	    ginf << "ASCIIFILE: Probably trying to read more lines than available in file" << endl;
	    return nilOOP;
	}
    }
    GenieWorkspace result(nitems);
    float* fa;
    int* ia;
    long* la;
    double* da;
    for(j=0; j<nitems; j++)
    {
        switch(field_type[j])
        {
            case FloatItem:
		fa = (float*)result_array[j];
    		result(field_labels[j], GenieRealArray(fa, count));
		break;

            case DoubleItem:
		da = (double*)result_array[j];
    		result(field_labels[j], GenieRealArray(da, count));
		break;

            case IntegerItem:
		ia = (int*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(ia, count));
		break;

            case LongIntegerItem:
		la = (long*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(la, count));
		break;

	    case StringItem:
	  	{
		    GenieStringArray gsa(count);
		    for(i=0; i<count; i++)
		    {
			gsa(i, (char*)result_array[j] + i * field_size[j]);
		    }
    		    result(field_labels[j], gsa);
		}
	 	break;

	    default:
	        *g_err << "ASCIIFILE: unknown item type in input file" << endl;
		*g_err << "ASCIIFILE: position  = " << j << ", label = \"" 
			<< field_labels[j] << "\"" << endl;
	        return nilOOP;
    	        break;
	}
	delete []result_array[j];
    }
//    if (nitems == 1)
//	return result(field_labels[0]);
//    else
//	return result;
    return result;
#undef MAXFIELDS
}


// Returns workspace unless only one item on line
static OOP asciiFileReadFreeFormat(SmalltalkObject& asciiFile, String& field_names,
					String& separator, g_integer count)
{
    GET_HANDLE_AND_STREAM;
    if (count == -1)
	count = lines_in_file(*fs, comment[0], delimiter[0]);
// "field_names" is a comma separated list of names to call the workspace fields 
#if defined(__DECCXX) || defined(_WIN32)
    const int MAXFIELDS(10);
#else
    const int MAXFIELDS(field_names.freq(',') + 1);
#endif
    String field_labels[MAXFIELDS];
    int nfields = split(field_names, field_labels, MAXFIELDS, ',');
    ItemType field_type[MAXFIELDS];
    char buffer[SHORTCHARBUFF];
    streampos mark = fs->tellg();
// Read first line
    if (skip_comments(*fs, comment[0], delimiter[0]))
    {
        fs->getline(buffer, sizeof(buffer), delimiter[0]);
    }
    if (!fs->good())
    {
        gerr << "ASCIIFILE: unable to read first line of file" << endl;
        return nilOOP;
    }
// Now parse line to see what type of data we have
    int i, j, nitems;
    Regex item_separator(separator, 1);
    Regex integer_item("[+-]?[0-9]+", 1);
    Regex real_item1("[+-]?[0-9]*\\.[0-9]*", 1);
    Regex real_item2("[+-]?[0-9]*[\\.]?[0-9]*[eE][+-]?[0-9]+", 1);
    String line(buffer), items[MAXFIELDS];
    void* result_array[MAXFIELDS];
    nitems = split(line, items, MAXFIELDS, item_separator);
    for(i=0; i<nitems; i++)
    {
	if (items[i].matches(real_item1) || items[i].matches(real_item2))
	{
	    field_type[i] = GenieRealItem;
	    result_array[i] = new g_real[count];
 	}
	else if (items[i].matches(integer_item))
	{
	    field_type[i] = GenieIntegerItem;
	    result_array[i] = new g_integer[count];
	}
	else
	{
	    field_type[i] = GenieStringItem;
	    result_array[i] = new char*[count];
	}
    }
// Now go back to beginning and read whole file
    fs->seekg(mark);
    g_real* gra;
    g_integer* gia;
    const char **gsa;
    for(i=0; i<count; i++)
    {
        if (skip_comments(*fs, comment[0], delimiter[0]))
        {
            fs->getline(buffer, sizeof(buffer), delimiter[0]);
        }
        if (fs->good())
	{
            nitems = split(buffer, items, MAXFIELDS, item_separator);
            for(j=0; j<nitems; j++)
	    {
	        switch(field_type[j])
	        {
	            case GenieRealItem:
			gra = (g_real*)result_array[j];
		        gra[i] = atof(items[j]);
		        break;

	            case GenieIntegerItem:
			gia = (g_integer*)result_array[j];
		        gia[i] = atol(items[j]);
		        break;
		    case GenieStringItem:
			gsa = (const char**)result_array[j];
		   	gsa[i] = strcpy((char*)malloc(1+strlen(items[j])), items[j]);
			break;

		    default:
		        gerr << "ASCIIFILE: unknown item type in input file at position " << j << endl;
		        return nilOOP;
    		        break;
  	        }
	    }
        }
	else
	{
	    *g_err << "ASCIIFILE: error reading input file at line " << i << endl;
	    ginf << "ASCIIFILE: Probably trying to read more lines than available in file" << endl;
	    return nilOOP;
	}
    }
    GenieWorkspace result(nfields+1);
    for(j=0; j<(nfields == 0 ? 1 : nfields); j++)
    {
        switch(field_type[j])
        {
            case GenieRealItem:
		gra = (g_real*)result_array[j];
    		result(field_labels[j], GenieRealArray(gra, count));
		break;

            case GenieIntegerItem:
		gia = (g_integer*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(gia, count));
		break;

	    case GenieStringItem:
		gsa = (const char**)result_array[j];
		result(field_labels[j], GenieStringArray(gsa, count));
		for(i=0; i<count; i++)
		    free(gsa[i]);
		break;

	    default:
	        gerr << "ASCIIFILE: unknown item type in input file" << endl;
	        return nilOOP;
    	        break;
	}
	delete []result_array[j];
    }
    if (nfields == 0)
	return result(field_labels[0]);
    return result;
}

static OOP asciiFileWriteFreeFormat(SmalltalkObject& asciiFile, SmalltalkObject& aVariable, String& separator, String& user_comment)
{
    GET_HANDLE_AND_STREAM;
    int gv_len = aVariable.arrayLength();
    fs->flags(ios::showpoint);	// show decimal point
    int len, i, j, k, ncolumns(0), nlines(0);
// Determine number of lines - use longest array length found
    GenieVariable** gv_list = new GenieVariable*[gv_len+1];
    int* gv_list_len = new int[gv_len+1];
    GenieVariable* pgv;
    for(j=1; j<=gv_len; j++)
    {
	if (aVariable.arrayAt(j) != nilOOP)
	{
	    ++ncolumns;
	    pgv = new GenieVariable(aVariable.arrayAt(j));
	    switch(pgv->type())
	    {
		case GenieIntegerArrayType:
		    len = ((GenieIntegerArray*)*pgv)->length();
		    gv_list_len[j] = len;
		    break;

		case GenieRealArrayType:
		    len = ((GenieRealArray*)*pgv)->length();
		    break;

		case GenieStringArrayType:
		    len = ((GenieStringArray*)*pgv)->length();
		    break;

		default:
		    len = 1;
		    break;
	    }
	    if (len > nlines)
		nlines = len;
	    gv_list[j] = pgv;
	    gv_list_len[j] = len;
	}
 	else
	{
	    gv_list[j] = 0;
	}
    }
    if (nlines == 0 || ncolumns == 0)
    {
	*g_err << "WRITEASCIIFREE: No lines" << endl;
	return nilOOP;
    }
    ginf << "ASCIIFILE: Number of lines written = " << nlines << endl;
    if (comment[0] != '\002')
    {  
	char buffer[128];
        time_t clock;
	time(&clock);
	struct tm* tm_struct = localtime(&clock);
	if (strftime(buffer, sizeof(buffer), "%a %d %h at %H:%M:%S %Z", tm_struct) == 0)
        {
            strcpy(buffer, "<unknown>");
        }
	*fs << comment[0] << delimiter[0];
        *fs << comment[0] << " Data written by Open GENIE on " << buffer << delimiter[0];
	*fs << comment[0] << " " << nlines << " data line(s) and " << ncolumns << " column(s)" << delimiter[0];
	*fs << comment[0] << delimiter[0];
	if (strlen(user_comment) > 0)
	{
	    *fs << comment[0];
	    for(j=0; j<strlen(user_comment); j++)
	    {
	        if (user_comment[j] == '\n')
	 	    *fs << delimiter[0] << comment[0];
	        else
	            fs->put(user_comment[j]);
	    }
	    *fs << delimiter[0] << comment[0] << delimiter[0];
	}
    }
// Now write out data
    for(i=0; i<nlines; i++)
    {
	k = 0;
	for(j=1; j<=gv_len; j++)
	{
	    pgv = gv_list[j];
	    len = gv_list_len[j];
	    if (pgv != 0)
	    {
		k++;
	        switch(pgv->type())
	        {
		    case GenieIntegerType:
			*fs << (g_integer)*pgv;
		        break;

		    case GenieRealType:
			*fs << (g_real)*pgv;
		        break;

		    case GenieStringType:
			*fs << (const char*)*pgv;
		        break;

		    case GenieIntegerArrayType:
			*fs << (*(GenieIntegerArray*)*pgv)[i%len];
		        break;

		    case GenieRealArrayType:
			*fs << (*(GenieRealArray*)*pgv)[i%len];
		        break;

		    case GenieStringArrayType:
			*fs << (*(GenieStringArray*)*pgv)[i%len];
		        break;

		    default:
			*g_err << "ASCIIFILE: Invalid type" << endl;
			*fs << "Invalid";
		        break;	
	        }
		if (k != ncolumns)
		    *fs << separator;
	    }
	}
	*fs << delimiter[0];
    }
    for(j=1; j<=gv_len; j++)
    {
	delete gv_list[j];
    }
    delete[] gv_list;
    delete[] gv_list_len;
    return 0;
}


static OOP asciiFileWriteFixedFormat(SmalltalkObject& asciiFile, GenieVariable& aVariable, String& format)
{
    GET_HANDLE_AND_STREAM;
    switch(aVariable.type())
    {
	default:
	    *g_err << "ASCIIFILE: Attempt to write unsupported type" << endl;
	    break;
    }
    return nilOOP;
}

