/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         g_fortextern
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        External callout for I/O routines (many originally written in FORTRAN, hence the name)
 *
 */
#include "geniedefs.h"
#include <sys/types.h>	/* for lseek and u_long */
#include <sys/stat.h>	/* for open */
#include <fcntl.h>	/* for open */

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#ifdef _WIN32
#	include <io.h>
#endif

#include <stdio.h>
#include <math.h>		/* needed for sqrt */
#include <stdlib.h>		/* for atoi */
#ifdef _WIN32
	char _itoa_buff[18];	/* _itoa can return 17 characters */
#	define dec(x) itoa(x, _itoa_buff, 10)	/* Seems to do the same thing !! */
#else
#	include <stream.h>
#endif /* _WIN32 */
#include "genievars.h"
#include "genie_types.h"
#include "fortran_types.h"
#include "g_fortextern.h"
#include "transform_utils.h"
#include "units_p.h"
#include "spectrum.h"
#include "g3io.h"
#include "Genie_functions.h"
#include "fstream_list.h"
#include "genie_intermediate_file.h"
#if HAVE_LIBMFHDF
#include "hdf_intermediate_file.h"
#endif /* HAVE_LIBMFHDF */
#include "read_ascii.h"
#include "smalltalk_object.h"
#include "callout_list.h"
#include "isis_raw_file.h"
#include "ranges.h"

static const char* rcsid = "@(#)$Id: g_fortextern.cxx,v 1.70 1997/06/29 15:15:07 faa Exp $";
USE_RCSID(rcsid);

static fstream_list open_streams;

#define	arrayDopeOneD(__gv,__n) \
		arrayGetDims(__gv)[0]=(__n); \
		arrayGetDims(__gv)[1]=0; \
		arrayGetDope(__gv)[0]=1; \
		arrayGetDope(__gv)[1]=0

extern "C" {
void for_rtl_init_ (int *, char **);
int for_rtl_finish_( void );
	}

static int directoryGenieIntermediate(GenieWorkspace& info, GenieString& file, GenieInteger full_listing);
static int directoryGenieASCIIIntermediate(GenieWorkspace& info, GenieString& file, GenieInteger full_listing);
static int check_and_open(String& runid, fort_int& ntc, fort_int& ndet, fort_int& nuse);
int readOldIntermediateFile(String& translated_file, int block, GenieWorkspace& result);
int focusSpectrumRange(g_integer iverbose, GenieWorkspace& wksp, GenieWorkspace& specpars, 
		       	GenieWorkspace& detpars, g_integer iunit, String& file, g_integer* iarray, int len);
int overlap_range(g_real* x1, int n1, g_real* x2, int n2, g_real* xnew);
int detector_efficiency(g_real* x, g_real* y, g_real* e, int npt, 
	double gp, double gsig0, double gt, double lam0, double wt, double rho, double ct, double wsig0);

OOP g_fortextern( fortran_function_code function, OOP parameters ) 
{

    BEGINSAVEBLOCK

    GenieCalloutList gcl(parameters);

    switch(function)
    {
    	case FreddieFortranTest:
	{
            gerr << "g_fortextern(test function called)\n";
	}
        break;

    	case FortranGetSNSSpectrum:
	{
  	    GenieInteger ispec(gcl, "SPECTRUM_NO");
	    GenieString runid(gcl, "FILE");
	    String run_no;
	    fort_real l1, l2, tthe, delt, duration, fi;
	    GenieWorkspace spectrum(gcl, "SPECTRUM");		/* for result */
    	    translatePath(runid, 'f');
	    fort_int ntc, ndet, nuse;
	    if (check_and_open(runid, ntc, ndet, nuse) != 0)
	    {
		break;
	    }
    	    fort_int ndetmax(ndet+1), ntcmax(ntc+1);
	    fort_int quick(0), errcode(0);
	    fort_int* spec_work = new fort_int[ndetmax];
	    fort_int* idat = new fort_int[ntcmax];		// for counts
	    fort_real* delt_work = new fort_real[ndetmax];
	    fort_real* tthe_work = new fort_real[ndetmax];
	    fort_real* l2_work = new fort_real[ndetmax];
	    fort_real* tcb = new fort_real[ntcmax];
	    double denom;
	    g_real yval, eval;

	    fort_real* user_tables = new fort_real[nuse];
	    String run_title, comb_time, user_name, inst_name, spec;

	    READ_DATA(runid,errcode,ispec,delt_work,spec_work,
		      tthe_work,l2_work,ndetmax,tcb,idat,ntcmax,
		      l1,l2,tthe,delt,fi,run_title,duration,
	  	      comb_time,user_name,inst_name,run_no, 
		      user_tables, nuse, quick);
	    if (errcode == 0)
	    {
	        spectrum("NTC", ntc);
	        spectrum("FI", fi);
	        spectrum("NUT", nuse);
	        spectrum("L1", l1);
	        spectrum("L2", l2);
	        spectrum("YMAX", 0.0);

 	        GenieRealArray ydata(ntc);
	        GenieRealArray edata(ntc);

	        int ii;
	        for(int i=0; i<ntc; i++)
	        {
/* first Y point is dummy in get routines */
	            ii = i + 1;
	            denom = tcb[i+1] - tcb[i];
	            yval = (double)idat[ii] / denom;
	            eval = sqrt((double)idat[ii]) / denom;
	            ydata(i) = yval;
	            edata(i) = eval;
	        }

/* 
 *  assign units on x and y axes --- these will be TOF (microseconds) 
 *  and neutron counts per microsecond respectively
 */
	        spectrum("XLABEL", unit_name(2, 'x'));
	        spectrum("YLABEL", unit_name(2, 'y'));

	        spectrum("X", GenieRealArray(tcb, ntc+1));
	        spectrum("Y", ydata);
	        spectrum("E", edata);
	        spectrum("UT", GenieRealArray(user_tables, nuse));

	        spectrum("TITLE", run_title);
	        spectrum("TIME", comb_time);
	        spectrum("HISTORY", " ");
	        spectrum("USER_NAME", user_name);
	        spectrum("INST_NAME", inst_name);
	        spectrum("RUN_NO", run_no);
	        spectrum("FILE", runid);
	        spectrum("DELTA", delt);
	        spectrum("RUN_DURATION", duration);
	        spectrum("SPEC_NO", dec(ispec));
	        spectrum("EFIXED", 0.0);
	        spectrum("TWOTHETA", tthe);
	        spectrum("EMODE", 0);
	        spectrum("BIN_GROUPING", 0);
	    }
	
	    delete []tcb;
	    delete []delt_work;
	    delete []tthe_work;
	    delete []l2_work;
	    delete []spec_work;
	    delete []idat;
	    delete []user_tables;
	}
	break;

        case FortranGetSNSItem:
	{
	    fort_int ntc, ndet, nuse;
	    GenieString item(gcl, "ITEM"), runid(gcl, "FILE");
	    GenieVariable item_value(gcl, "ITEM_VALUE");
    	    translatePath(runid, 'f');
	    if (check_and_open(runid, ntc, ndet, nuse) != 0)
	    {
		break;
	    }
	    getSNSItem(runid, item, item_value);
	}
	break;

	case GetSNSSpectrumArray:
	{
	    GenieIntegerArray iarray(gcl, "IARRAY");
	    GenieString runid(gcl, "FILE");
	    GenieVariable spectrum_array(gcl, "SPECTRUM");
	    GenieWorkspaceArray wa(iarray.length());
	    ISISRawFile raw_file;
	    raw_file.open(runid);
	    for(int i=0; i<iarray.length(); i++)
	    {
	        GenieWorkspace gw;
	        raw_file.readSpectrum(gw, iarray[i]);
	        wa(i, gw);
	    }
	    spectrum_array.become(wa.value());
	}
	break;

	case GetSNSMultiSpectrum:
	{
	    GenieIntegerArray iarray(gcl, "IARRAY");
	    GenieString runid(gcl, "FILE");
	    GenieVariable multi_spectrum(gcl, "SPECTRUM");
	    GenieWorkspace gw;
	    ISISRawFile raw_file;
	    raw_file.open(runid);
	    raw_file.readSpectra(gw, iarray, iarray.length());
	    multi_spectrum.become(gw.value());
	}
	break;

        case ReadLine:
	{
	    String message;
	    GenieInteger stream_id(gcl, "stream_id"); 
	    GenieString delim(gcl, "delim"), line(gcl, "line"), comment(gcl, "comment");
	    char buffer[SHORTCHARBUFF];
	    fstream* fs = open_streams[stream_id];
	    if (!fs->good())
	    {
		cat("Cannot connect input stream to file descriptor " , dec(stream_id), message);
	 	error_add("ReadLine", message, "Check descriptor points to a valid file");
		break;
	    }
	    if (skip_comments(*fs, comment[0], delim[0]))
 	    {
	        fs->getline(buffer, sizeof(buffer), delim[0]);
	    }
	    if (fs->good())
	    {
	        line = buffer;
	    }
	    else
	    {
		line = '\0';
	    }
	}
	break;
	    
	case LinesInFile:
	{
	    GenieString file(gcl, "file"), delim(gcl, "delim"), comment(gcl, "comment");
	    GenieInteger nlines(gcl, "nlines");
	    nlines = lines_in_file(file, comment[0], delim[0]);
	}
	break;

	case FocusSpectrumRange:
	{
	    GenieInteger iunit(gcl, "IUNIT"), iverbose(gcl, "IVERBOSE");
	    GenieIntegerArray iarray(gcl, "IARRAY");
	    GenieString file(gcl, "FILE");
	    GenieWorkspace specpars(gcl, "SPECPARS"), detpars(gcl, "DETPARS"), wksp(gcl, "WKSP");
	    if (detpars.defined())
	    {
	        ginf << "Performing detector efficiency correction for " << (const char*)detpars("type") << endl;
	    }
	    else
	    {
	        ginf << "No detector efficiency correction performed" << endl;
	    }
	    focusSpectrumRange(iverbose, wksp, specpars, detpars, iunit, file, iarray, iarray.length());
	}
	break;

// We are passed some genie array s
	case ReadAscii:
	{
	    GenieWorkspace res(gcl, "res");
	    GenieString delim(gcl, "delim"), comment(gcl, "comment"), file(gcl, "file"), field_names(gcl, "field_names");
	    read_ascii(file, field_names, res);  //  , comment[0], delim[0]);
	}
	break;

	case WriteAscii:
	{
	    GenieInteger mode(gcl, "MODE");		// not used at present
	    GenieString file(gcl, "FILE"), comment(gcl, "COMMENT"), format(gcl, "FORMAT");
// Horrible hack - SmalltalkObject does not have an associate() method,
// So we read an integer and then steal the OOP
	    GenieInteger gvv(gcl, "GV");
	    SmalltalkObject gv(gvv.value());
	    write_ascii(file, comment, format, gv);
	}
	break;

	case WriteAsciiStream:
	{
	    GenieInteger mode(gcl, "MODE");		// not used at present
	    GenieVariable gv(gcl, "GV");
	    GenieInteger stream_id(gcl, "STREAM");
	    if ( ((int)stream_id < 0) || ((int)stream_id >= open_streams.count()) )
	    {
		error_add("WRITE", "Attempt to write to unopened or invalid stream number", "");
		break;
	    }
	    fstream* fs = open_streams[stream_id];
	    if (fs != 0)
	    {
		gv.writeAscii(*fs);
	    }
	    else
	    {
		error_add("WRITE", "Attempt to write to previously open but now closed closed stream", 
				  "Remove the offending CLOSE/STREAM command");
	    }
	}
	break;
	
	case ReadAsciiStream:
	{
	    String message;
	    fstream* fs;
	    GenieVariable res(gcl, "RES");
	    GenieInteger stream_id(gcl, "STREAM");
	    GenieString comment(gcl, "COMMENT");
	    if ( ((int)stream_id < 0) || ((int)stream_id >= open_streams.count()) )
	    {
		error_add("READ", "Attempt to read from unopened or invalid stream number", "");
		break;
	    }
  	    fs = open_streams[stream_id];
	    if (fs != 0)
	    {
		res.readAscii(*fs);
//	    	read_ascii_stream(*fs, res);
	    }
	    else
	    {
		error_add("READ", "Attempt to read from previously open but now closed closed stream", 
				  "Remove the offending CLOSE/STREAM command");
	    }
	}
	break;

	case CloseStream:
	{
	    String message;
	    GenieInteger stream_id(gcl, "stream_id");
	    if ( ((int)stream_id >= 0) && ((int)stream_id < open_streams.count()) )
	    {
	        fstream* fs = open_streams[stream_id];
		if (fs != 0)
		{
	            fs->close();
	            delete fs;
	            open_streams[stream_id] = 0;
		}
	    }
	    else
	    {
		cat("Attempt to close invalid stream number = ", dec(stream_id), message);
		error_add("CLOSE", message, "");
	    }
	}
	break;

	case OpenStream:
	{
	    String message;
	    GenieString file(gcl, "file");
	    GenieInteger stream_id(gcl, "stream_id"), mode(gcl, "mode");
	    fstream* fs = new fstream;
	    int flags(0);
	    if (mode & 1) 
		flags |= ios::in;
	    if (mode & 2) 
		flags |= (ios::out|ios::trunc);
	    if (mode & 4) 
		flags |= ios::app;
	    fs->open(file, flags);
	    if (*fs)
	    {
			open_streams.push(fs);
			stream_id = open_streams.count() - 1;
	    }
	    else
	    {
			cat("Cannot open ", file, message);
			error_add("OpenStream", message, "");
	   		stream_id = -1;
	    }
 	}
	break;

	case LengthIntermediateFile:
	{
            GenieString file(gcl, "FILE");
	    GenieInteger nblocks(gcl, "NBLOCKS");
            IntermediateFile* int_file;
	    g_integer nb;
	    FileType type = getFileType(file);
// Special case: Genie II intermediate files
	    if (type == FileTypeG2Intermediate)
	    {
		g2NSpectrum(file, &nb);
		nblocks = nb;
		break;
	    }
	    else if (type == FileTypeG3Intermediate)
		int_file = new GenieIntermediateFile;
#if HAVE_LIBMFHDF
	    else if (type == FileTypeHDF)
		int_file = new HDFIntermediateFile;
#endif /* HAVE_LIBMFHDF */
	    else
	    {
		gerr << "ERROR: Not and intermediate file" << endl;
		break;
	    }
            if (!int_file->open(file, IntermediateFile::Append))
            {
                break;
            }
            nblocks = int_file->numberOfBlocks();
	    int_file->close();
	    delete int_file;
	    break;
	}

	case DirectoryIntermediateFile:
	{
	    GenieWorkspace info(gcl, "info");
	    GenieString file(gcl, "file");
   	    GenieInteger full_listing(gcl, "full_listing");
	    FileType type = getFileType(file);
	    if (type == FileTypeG3Intermediate)
                directoryGenieIntermediate(info, file, full_listing);
	    else if (type == FileTypeOpenGenieASCII)
                directoryGenieASCIIIntermediate(info, file, full_listing);
	    else
		*g_err << "LIST: \"" << file << "\" is not a supported type" << endl;
	}
	break;

   	case WriteIntermediateFile:
	{
// type should be "FileTypeG3Intermediate" (3) or "FileTypeHDF" (5)
	    GenieInteger type(gcl, "type"), mode(gcl, "mode"), nblocks(gcl, "nblocks");
	    GenieString file(gcl, "file"), label(gcl, "label"), comment(gcl, "comment");
	    GenieVariable variable(gcl, "variable");
	    IntermediateFile* int_file;
	    if ((g_integer)type == FileTypeG3Intermediate)
		int_file = new GenieIntermediateFile;
#if HAVE_LIBMFHDF
	    else if ((g_integer)type == FileTypeHDF)
		int_file = new HDFIntermediateFile;
#endif /* HAVE_LIBMFHDF */
	    else
	    {
		gerr << "Invalid intermedaite file type = " << (int)type << endl;
		break;
	    }
 	    int_file->open(file, (GenieIntermediateFile::ModeFlags)(g_integer)mode) && 
			int_file->appendBlock(variable, label, comment);
	    nblocks = int_file->numberOfBlocks();
	    int_file->close();
	    delete int_file;
	}
	break;

   	case ReadIntermediateFile:
	{
	    GenieString file(gcl, "FILE"); 
	    GenieVariable item(gcl, "ITEM"), res(gcl, "RES");
	    GenieInteger nblocks(gcl, "NBLOCKS");
	    GenieIntermediateFile int_file;
	    nblocks = -1;		// what is returned on an error
	    if (!int_file.open(file, GenieIntermediateFile::Input))
		break;
	    
	    nblocks = int_file.numberOfBlocks();
	    if (item.type() == GenieStringType)
	    {
		int_file.seek((c_g_string)item);
	        int_file.readBlock(res);
	    }
	    else if (item.type() == GenieIntegerType)
	    {
		int_file.seek((g_integer)item - 1);
	        int_file.readBlock(res);
	    }
	    else
	    {
		error_add("ReadIntermediateFile", "invalid type" , "");
	    }
	}
	break;

   	case ReadOldGenieIntermediateFile:
	{
	    GenieString file(gcl, "FILE");
	    GenieInteger nblocks(gcl, "NBLOCKS"), block(gcl, "BLOCK");
	    GenieWorkspace result(gcl, "RES");
	    String translated_file(file);
	    translatePath(translated_file, 'f');
// nblocks is -1 on error
	    nblocks = readOldIntermediateFile(translated_file, block, result);
	}
	break;
#if 0

    case FortranWriteSpectrum:
	type = (fortran_write_mode) getInteger(GEBF("TYPE"));
	istyle = (fortran_write_style) getInteger(GEBF("ISTYLE"));
	imode = (fortran_open_mode) getInteger(GEBF("IMODE"));
	file = getString(GEBF("INT_FILE"));
	tgv = GEBF("WKSP");
	blocks = fortranWriteSpectrum(type, istyle, imode, file, tgv);
	putInteger(tgv, blocks);
	PEBF("BLOCKS", tgv);
	break;

    case FortranMultiGetSNSSpectrum:
	istart = getInteger(GEBF("ISTART"));
	iend = getInteger(GEBF("IEND"));
	iopt = getInteger(GEBF("OPT"));
	if ((error_code = multiGetSNSSpectrum(runid, istart, iend, iopt, ntcmax, ntc, ndetmax, nuse, &spectrum)) == 0)
	{
	    PEBF("WKSP", spectrum);
	}
	break;


/*
	if ( iopt == 0 )
	{
	    spectrum = GEBF("WKSP");
	    putElementByField(spectrum, "X", x_data);
	    putElementByField(spectrum, "Y", y_data);
	    putElementByField(spectrum, "E", e_data);
	    putElementByField(spectrum, "YMAX", tgv);
	    PEBF("WKSP", spectrum);
	}
	else if ( iopt == 1 )
	{
	    PEBF("XARRAY", x_data);
	    PEBF("YARRAY", y_data);
	    PEBF("EARRAY", e_data);
	    PEBF("YMAX", tgv);
	}
	else
	{
	     error_code = 1;
	     sprintf(buffer, "Invalid value of IOPT (%i) specified\n", iopt);
	     Error_message = buffer;
	}
*/

    case FortranImportSpectrum:
	iopt = getInteger(GEBF("IOPT"));
	spectrum = GEBF("TWKSP"); 	
/*
 * spectrum is already assigned with the old workspace 
 */
	ntcmax = 33000;
	if (undefined(spectrum))
	{
	    fs = initFSpectrum(ntcmax, 30);	/* load command */
	}
	else
	{
	    fs = extractFSpectrum(spectrum, ntcmax, 30);	/* function and transform */
	}
	ispec = 0;
	if ( iopt == 1 )
	{
	    TRAN_IN(fs->n, ntcmax, fs->x, fs->y, fs->e, fs->xl, fs->yl, fs->tl, fs->d, 
		     inst_code, fs->in, xcode, ycode, fs->l1, fs->l2, 
		     fs->tt, ispec, irun_no, fs->un, fs->dn, fs->tm,
		     fs->fi, fs->ut, st_var, fs->h, ferr);
	}
	else if ( iopt == 2 )
	{
	    FUNC_IN(fs->n, fs->y, fs->e, fs->xl, fs->yl, fs->tl, xcode, ycode,
		    fs->ut, st_var, fs->h, ferr);
	}
	else if ( iopt == 3 )
	{
	    LOAD_IN(fs->n, fs->x, fs->y, fs->e, fs->xl, fs->yl, fs->tl, fs->d,
		    inst_code, fs->in, xcode, ycode, fs->l1, fs->l2, fs->tt,
		    ispec, irun_no, fs->un, fs->dn, fs->tm, fs->fi, fs-> ut,
		    st_var, fs->h, ferr);
	}
	error_code = ferr;
	fs->xl = unit_name(xcode, 'x');
	fs->yl = unit_name(ycode, 'y');
	fs->rn = crunid(irun_no);
	fs->sn = dec(ispec);
	if ( iopt == 3 )
	{
	    spectrum = createFSpectrum(fs);
	}
	else
	{
	    replaceFSpectrum(&spectrum, fs);
	}
	deleteFSpectrum(fs);

	e_data = newGarrayReal(20);
	arrayDopeOneD(e_data ,20);
	for(i=0; i<20; i++)
	{
	    denom = (g_real) st_var[i];
	    arrayPutReal(e_data, i, denom);
	}
	PEBF("VARRAY", e_data);
	PEBF("TWKSP", spectrum);
	break;

    case FortranLoadOut:
	file = getString(GEBF("LOAD_FILE"));
	LOAD_OUT(file, ferr);
	error_code = ferr;
	break;

    case FortranExportSpectrum:
	iopt = getInteger(GEBF("IOPT"));
	e_data = GEBF("VARRAY");	/* global array V of user parameters */
	for(i=0; i<20; i++)
	{
	    denom = arrayGetReal(e_data, i);
	    st_var[i] = (fort_real) denom;
	}
	spectrum = GEBF("WKSP");
	fs = extractFSpectrum(spectrum, 0, 30);
	ntcmax = fs->n + 1;
	inst_code = 0;
	xcode = unit_code(fs->xl, 'x');
	ycode = unit_code(fs->yl, 'y');
	ispec = (fort_int) atol(fs->sn);
	irun_no = (fort_int) atol(fs->rn);
	if ( iopt == 1 )
	{
	    TRAN_OUT(fs->n, ntcmax, fs->x, fs->y, fs->e, fs->xl, fs->yl, fs->tl, fs->d, 
		     inst_code, fs->in, xcode, ycode, fs->l1, fs->l2, 
		     fs->tt, ispec, irun_no, fs->un, fs->dn, fs->tm,
		     fs->fi, fs->ut, st_var, fs->h, ferr);
	}
	else if ( iopt == 2 )
	{
	    FUNC_OUT(fs->n, fs->x, fs->y, fs->e, fs->xl, fs->yl, fs->tl, fs->l1, fs->l2,
		     fs->tt, fs->fi, fs->d, fs->ut, st_var, xcode, ycode, fs->h, ferr);
	}
	error_code = ferr;
	deleteFSpectrum(fs);
	break;

    case FortranReadSpectrum:
	file = getString(GEBF("INT_FILE"));
	iopt = getInteger(GEBF("READ_MODE"));
	istyle = (fortran_write_style) getInteger(GEBF("ISTYLE"));
	if ( istyle == G2Style )
	{
	    translate_file = translatePath(file, 'f');
	    spectrum = newGworkspace(27);
	    block = getInteger(GEBF("ITEM"));
	    error_code = g2ReadSpectrum(translate_file, &spectrum, block, &blocks);
	    putInteger(tgv, blocks);
	    PEBF("BLOCKS", tgv);
	    PEBF("VARIABLE_READ", spectrum);
	}
	else if ( istyle == G3Style && iopt == 1 )
	{
	    translate_file = translatePath(file, 'c');
	    block = 0;
	    marker = getString(GEBF("ITEM"));
	    g3ReadBlock(translate_file, block, marker, iopt, &tgv1, &blocks);
	    PEBF("VARIABLE_READ", tgv1);
	    putInteger(tgv2, blocks);
	    PEBF("BLOCKS", tgv2);
	}
	else if ( istyle == G3Style && iopt == 2 )
	{
	    translate_file = translatePath(file, 'c');
	    block = getInteger(GEBF("ITEM"));
	    marker = "";
	    g3ReadBlock(translate_file, block, marker, iopt, &tgv1, &blocks);
	    PEBF("VARIABLE_READ", tgv1);
	    putInteger(tgv2, blocks);
	    PEBF("BLOCKS", tgv2);
	}
	break;
	    
    case FortranNSpectrum:
	file = getString(GEBF("INT_FILE"));
	iopt = getInteger(GEBF("MODE"));
	switch(iopt)
	{
	    case 2:
	        translate_file = translatePath(file, 'f');
	        error_code = g2NSpectrum(translate_file, &blocks);
		break;
	    case 3:
	        translate_file = translatePath(file, 'c');
	        error_code = g3NSpectrum(translate_file, &blocks);
		break;
	    default:
		blocks = 0;
		error_code = 1;
		break;
	}
	putInteger(tgv, blocks);
	PEBF("NSPEC", tgv);
	break;

    case FortranDirIntFile:
	translate_file = translatePath(getString(GEBF("INT_FILE")), 'c');
	iopt = getInteger(GEBF("MODE"));
	dirG3IntFile(translate_file, iopt, &tgv1, &tgv2, &tgv3);
	PEBF("DIRECTORY", tgv1);
	PEBF("SIZES", tgv2);
	PEBF("NSPEC", tgv3);
	break;
	
    case FortranWriteG3File:
	file = getString(GEBF("INT_FILE"));
	iopt = getInteger(GEBF("MODE"));	/* append=1, new=2 --- see g3io.h */
	marker = getString(GEBF("MARKER"));
	tgv1 = GEBF("VAR_TO_WRITE");
	undefineGvar(tgv2);
	undefineGvar(tgv3);
	istyle = (fortran_write_style) getInteger(GEBF("ISTYLE"));
	if ( istyle == G2Style )
	{
	    translate_file = translatePath(file, 'f');
	    error_code = fortranWriteG2File(translate_file, iopt, tgv1, &tgv2, &tgv3);
	}
	else if ( istyle == G3Style )
	{
	    translate_file = translatePath(file, 'c');
	    error_code = fortranWriteG3File(translate_file, iopt, tgv1, marker, &tgv2, &tgv3);
	}
	else
	{
	    error_add("Write", "Invalid Style", "");
	    error_code = 1;
	}
	PEBF("NBYTES", tgv2);
	PEBF("BLOCKS", tgv3);
	break;

    case FortranRunUserProg:
	translate_file = translatePath(getString(GEBF("PROG")), 'c');
	error_code = system(translate_file);
	putInteger(tgv, error_code);
	PEBF("STATUS", tgv);
	break;

    case WriteIntermediateFile:
    {
	GenieVariable gv;
	GenieInteger mode;
	GenieString label, comment, file;
	associateGenieVariables(parameters, "VARIABLE", &gv, "MODE", &mode, NULL);
	label.associate(parameters, "LABEL");
  	comment.associate(parameters, "COMMENT");
 	file.associate(parameters, "FILE");
	GenieIntermediateFile gif;
	gif.open(file);

#endif

// Determine the type of FILENAME, returning the string TYPE
    case FortranFileType:
    {
	FileType fileType;
	GenieString file(gcl, "FILE"), file_type(gcl, "TYPE");
    	translatePath(file, 'c');
	if ((fileType = getFileType(file)) == FileTypeError)
	{
	    file_type = "error";
	}
	else
	{
	    file_type = fileTypeNames[fileType];
	}
    }
    break;

    case WriteGenieASCII:
    {
	    GenieInteger type(gcl, "type"), mode(gcl, "mode"), nblocks(gcl, "nblocks");
	    GenieString file(gcl, "file"), label(gcl, "label"), comment(gcl, "comment");
	    static char buffer[128];
	    GenieVariable gv, variable(gcl, "variable");
	    fstream fs;
	    int nb = 0;
	    int flags(0);
	    if (mode & 1) // read
		flags |= ios::in;
	    if (mode & 2) // new file
	    {
		flags |= (ios::out|ios::trunc);
	    }
	    if (mode & 4) // append
		flags |= (ios::in|ios::out);
	    fs.open(file, flags);
	    if (!fs)
	    {
		*g_err << "PUT: Cannot open \"" << file << "\"" << endl;
		break;
	    }
	    if (mode & 2)	// new
		fs << G3_ASCII_FILE_HEADER_BASE << "\n";
	    else if (mode & 4)	// append
	    {
		fs.getline(buffer, sizeof(buffer), '\n');	// skip header
		while(fs.good())
		{
		    fs.getline(buffer, sizeof(buffer), '\n');	// label
		    gv.readAscii(fs);				// variable
		    while(fs.good() && fs.get() != '#')
			;
		    nb++;
		}
	        fs.clear();
	    }
	    fs.seekg(0, ios::end);
	    fs << "# " << (const char*)label << "\n";
	    variable.writeAscii(fs);
	    nb++;
	    fs.close();
	    nblocks=nb;
    }
    break;

    case ReadGenieASCII:
    {
	    GenieString file(gcl, "FILE"); 
	    GenieVariable gv, item(gcl, "ITEM"), res(gcl, "RES");
	    GenieInteger nblocks(gcl, "NBLOCKS");
	    int nb(1), block_no(-1), i;
	    static char buffer[256];
	    const char* block_name = 0;
	    fstream fs;
	    fs.open(file, ios::in);
	    if (!fs)
	    {
		*g_err << "GET: Cannot open \"" << file << "\"" << endl;
		break;
	    }
	    if (item.type() == GenieIntegerType)
		block_no = (g_integer)item;
	    else if (item.type() == GenieStringType)
		block_name = (const char*)item; 
	    fs.getline(buffer, sizeof(buffer), '\n');	// skip header
	    while(fs.good())
	    {
		    fs.getline(buffer, sizeof(buffer), '\n');	// label
		    i = (buffer[0] == '#' ? 2 : 1);		// may have already skipped #
		    if (block_no == nb || (block_name != 0 && strcmp(buffer+i, block_name) == 0)) // +i to skip leading "# "
		    	res.readAscii(fs);				// read variable
		    else
		        gv.readAscii(fs);				// skip variable
		    while(fs.good() && fs.get() != '#')			// skip to next label
			;
		    nb++;
	    }
	    fs.close();
	    nblocks=nb-1;
    }
    break;

    default:
	break;
    }
#if 0
    for_rtl_status = for_rtl_finish_();
//    printf("Fortran RTL status = %i\n", for_rtl_status);
#endif
    ENDSAVEBLOCK

    return parameters;
}


#if 0
int fortranWriteG2File(String file, g_integer mode, g_variable gv_to_write, 
		       g_variable *g_nbytes, g_variable *g_blocks)
{
    fort_int imode, len, lend, iunit=39, len_stream=33000, irun=0, nspec=0, 
	xcode=-1, ycode=-1, ispec=0, ierror=0, minimum=1;
    f_spectrum_struct *fss;
    fss = extractFSpectrum(gv_to_write, 0, 30);
    len = (fort_int) file.length();
    lend = (fort_int) fss->f.length();
    xcode = unit_code(fss->xl, 'x');
    ycode = unit_code(fss->yl, 'y');
    if (ycode == 0) { ycode = -1; }	// Genie II used -1 for dimensionless rather than 0
    irun = atol(fss->rn);
    ispec = atol(fss->sn);
    imode = (fort_int) mode;
    WRITE_INT(iunit, file, len, len_stream , fss->x, fss->y, fss->e, nspec, ispec, fss->n, fss->d, irun,
	     xcode, ycode, fss->em, fss->fi, fss->l1, fss->l2, fss->ef, fss->tt, fss->dn,
	     fss->in, fss->tl, fss->un, fss->tm, fss->h, fss->xl, fss->yl, imode, 
	     minimum, fss->f, lend, fss->ut, ierror); 
    deleteFSpectrum(fss);
    putInteger(*g_blocks, nspec);
    undefineGvar(*g_nbytes);
    return ierror;
}

int fortranWriteG3File(String file, g_integer mode, g_variable gv_to_write, 
		       const char* marker, g_variable *g_nbytes, g_variable *g_blocks)
{
    int fd, blocks=0;
    size_t nbytes=0;
    file_open_mode o_mode;

    o_mode = (file_open_mode) mode;
    if ((fd = openG3File(file, o_mode, &blocks)) == -1)
    {
	return -1;
    }
    nbytes = writeGvar(fd, gv_to_write, marker);
    blocks = updateG3FileHeader(fd, 1);
    closeG3File(fd);
    putInteger(*g_nbytes, nbytes);
    putInteger(*g_blocks, blocks);
    return 0;
}

#endif

// return -1 on error, else number of blocks in file
int readOldIntermediateFile(String& translated_file, int block, GenieWorkspace& result)
{
    static char error_mess[128];
    fort_int imode, len, iunit=39, irun=0, nspec=0, xcode=-1, ycode=-1, block_no, ispec=0, ierror=0;
    f_spectrum_struct *fss, *fstmp;
    len = (fort_int) translated_file.length();
    block_no = (fort_int)block;
    fstmp = initFSpectrum(1, 30);
    imode = 1; /* read_int will just set npt */
    READ_INT(iunit, translated_file, len, block_no, fstmp->x, fstmp->y, fstmp->e, nspec, ispec, fstmp->n, fstmp->d, irun,
	     xcode, ycode, fstmp->em, fstmp->fi, fstmp->l1, fstmp->l2, fstmp->ef, fstmp->tt, fstmp->dn,
	     fstmp->in, fstmp->tl, fstmp->un, fstmp->tm, fstmp->h, fstmp->xl, fstmp->yl, imode, fstmp->ut, ierror); 
    if (ierror != 0)
    {
	sprintf(error_mess, "Error code %d", ierror);
	error_add("readOldIntermediateFile", error_mess, "");
        deleteFSpectrum(fstmp);
	return -1;
    }
    imode = 0; /* proper read */
    fss = initFSpectrum(fstmp->n, 30);
    READ_INT(iunit, translated_file, len, block_no, fss->x, fss->y, fss->e, nspec, ispec, fss->n, fss->d, irun,
	     xcode, ycode, fss->em, fss->fi, fss->l1, fss->l2, fss->ef, fss->tt, fss->dn,
	     fss->in, fss->tl, fss->un, fss->tm, fss->h, fss->xl, fss->yl, imode, fss->ut, ierror); 
    deleteFSpectrum(fstmp);
    if (ierror != 0)
    {
	sprintf(error_mess, "Error code %d", ierror);
	error_add("readOldIntermediateFile", error_mess, "");
        deleteFSpectrum(fss);
	return -1;
    }
    fss->rn = crunid(irun);
    fss->sn = dec(ispec);
    fss->f = translated_file;
    fss->xl = unit_name(xcode, 'x');
    if (ycode == -1) { ycode = 0; }	// Genie II used -1 for dimensionless rather than 0
    fss->yl = unit_name(ycode, 'y');
    fss->ym = 0.0F;
//    for(i=0; i<fss->n; i++)
//    {
//	fss->ym = max(fss->ym, fss->y[i]);
//    }
//    fss->nut = 30;
    result("NTC", fss->n);
    result("X", GenieRealArray(fss->x, fss->n + 1));
    result("Y", GenieRealArray(fss->y, fss->n));
    result("E", GenieRealArray(fss->e, fss->n));
    result("NUT", fss->nut);
    result("UT", GenieRealArray(fss->ut, fss->nut));
    result("TITLE", fss->tl);
    result("TIME", fss->tm);
    result("HISTORY", fss->h);
    result("FI", fss->fi);
    result("USER_NAME", fss->un);
    result("INST_NAME", fss->in);
    result("RUN_NO", fss->rn);
    result("FILE", fss->f);
    result("DELTA", fss->d);
    result("RUN_DURATION", fss->dn);
    result("SPEC_NO", fss->sn);
    result("XLABEL", fss->xl);
    result("YLABEL", fss->yl);
    result("BIN_GROUPING", fss->bg);
    result("YMAX", fss->ym);
    result("L1", fss->l1);
    result("L2", fss->l2);
    result("TWOTHETA", fss->tt);
    result("EMODE", fss->em);
    result("EFIXED", fss->ef);
    deleteFSpectrum(fss);
    return(nspec);
}

#if 0
int g2ReadSpectrum(String file, g_variable* spectrum, g_integer block, g_integer* blocks)
{
    int i;
    fort_int imode, len, iunit=39, irun=0, nspec=0, xcode=-1, ycode=-1, block_no, ispec=0, ierror=0;
    f_spectrum_struct *fss, *fstmp;
    len = (fort_int) file.length();
    block_no = (fort_int)block;
    fstmp = initFSpectrum(1, 30);
    imode = 1; /* read_int will just set npt */
    READ_INT(iunit, file, len, block_no, fstmp->x, fstmp->y, fstmp->e, nspec, ispec, fstmp->n, fstmp->d, irun,
	     xcode, ycode, fstmp->em, fstmp->fi, fstmp->l1, fstmp->l2, fstmp->ef, fstmp->tt, fstmp->dn,
	     fstmp->in, fstmp->tl, fstmp->un, fstmp->tm, fstmp->h, fstmp->xl, fstmp->yl, imode, fstmp->ut, ierror); 
    if (ierror != 0)
    {
	error_add("g2ReadSpectrum", " ", "");
	undefineGvar(*spectrum);
	*blocks = (g_integer) LONG_MAX;
        deleteFSpectrum(fstmp);
	return ierror;
    }
    imode = 0; /* proper read */
    fss = initFSpectrum(fstmp->n, 30);
    READ_INT(iunit, file, len, block_no, fss->x, fss->y, fss->e, nspec, ispec, fss->n, fss->d, irun,
	     xcode, ycode, fss->em, fss->fi, fss->l1, fss->l2, fss->ef, fss->tt, fss->dn,
	     fss->in, fss->tl, fss->un, fss->tm, fss->h, fss->xl, fss->yl, imode, fss->ut, ierror); 
    deleteFSpectrum(fstmp);
    if (ierror != 0)
    {
	error_add("g2ReadSpectrum", " ", "");
	undefineGvar(*spectrum);
	*blocks = (g_integer) LONG_MAX;
        deleteFSpectrum(fss);
	return ierror;
    }
    fss->nut = 30;
    fss->rn = crunid(irun);
    fss->sn = dec(ispec);
    fss->f = file;
    fss->xl = unit_name(xcode, 'x');
    if (ycode == -1) { ycode = 0; }	// Genie II used -1 for dimensionless rather than 0
    fss->yl = unit_name(ycode, 'y');
    fss->ym = 0.0;
    for(i=0; i<fss->n; i++)
    {
	fss->ym = max(fss->ym, fss->y[i]);
    }
    *spectrum = createFSpectrum(fss);
    *blocks = (g_integer)nspec;
    deleteFSpectrum(fss);
    return((int)ierror);
}

#endif
/*
 * determine number of spectra in an intermenidate file
 */
int g2NSpectrum(String file, g_integer *nspectrum)
{
    static char error_mess[128];
    fort_int imode, len, iunit=39, irun=0, nspec=0, xcode=-1, ycode=-1, block_no=1, ispec=1, ierror=0;
    f_spectrum_struct *fstmp;
    String translated_file(file);	
    translatePath(translated_file, 'f');
    len = (fort_int)translated_file.length();
    fstmp = initFSpectrum(1, 1);
    imode = 1; /* read_int will just set npt */
    READ_INT(iunit, translated_file, len, block_no, fstmp->x, fstmp->y, fstmp->e, nspec, ispec, fstmp->n, fstmp->d, irun,
	     xcode, ycode, fstmp->em, fstmp->fi, fstmp->l1, fstmp->l2, fstmp->ef, fstmp->tt, fstmp->dn,
	     fstmp->in, fstmp->tl, fstmp->un, fstmp->tm, fstmp->h, fstmp->xl, fstmp->yl, imode, fstmp->ut, ierror); 
    deleteFSpectrum(fstmp);
    *nspectrum = (g_integer)nspec;
    if (ierror != 0)
    {
        sprintf(error_mess, "Error code %d", ierror);
        error_add("g2NSpectrum", error_mess, "");
    }
    return ((int)ierror);
} 

#if 0
/*
 * return 0 on success, 1 on error
 */
int g3NSpectrum(String file, g_integer *nspectrum)
{
    int fd, blocks;
    file_open_mode o_mode;
    o_mode = OpenRead;

    fd = openG3File(file, o_mode, &blocks);
    if ( fd > 0 )
    {
        closeG3File(fd);
        *nspectrum = blocks;
        return 0;
    }
    else
    {
	*nspectrum = LONG_MAX;	/* so returned undefined to Genie */
	return -1;
    }
}

int dirG3IntFile(String file, g_integer iopt, g_variable *directory, g_variable *sizes, g_variable *nspec)
{
    int fd, blocks, i;
    char **direc;
    size_t *sizes_array;
    file_open_mode o_mode;
    g_integer gi;

    o_mode = OpenRead;
    if ((fd = openG3File(file, o_mode, &blocks)) == -1)
    {
	return -1;
    }
    dirG3File(fd, &direc, &sizes_array); 
    closeG3File(fd);
    *directory = newGarrayString(blocks);
    arrayDopeOneD(*directory, blocks);
    *sizes = newGarrayInteger(blocks);
    arrayDopeOneD(*sizes, blocks);
    putInteger(*nspec,  blocks);
    for(i=0; i<blocks; i++)
    {
	arrayPutString(*directory, i, direc[i], strlen(direc[i]));
	gi = (g_integer) sizes_array[i];
	arrayPutInteger(*sizes, i, gi);
    }
    return 0;
}

int g3ReadBlock(const char* file, g_integer block, const char *marker, int imode, g_variable *gv, g_integer *blocks)
{
    size_t nbytes;
    int fd, nblocks;
    file_open_mode o_mode;
    o_mode = OpenRead;
    if ((fd = openG3File(file, o_mode, &nblocks)) == -1)
    {
	return -1;
    }
    *blocks = nblocks;
    if ( imode == 1 ) 
    {
	seekG3FileMarker(fd, marker);
    } 
    else if ( imode == 2 )
    {
	seekG3FileBlock(fd, block);
    } 
    nbytes = readGvar(fd, gv, &marker);
    closeG3File(fd);
    return 0;
}

int multiGetSNSSpectrum(String file, g_integer istart, g_integer iend, g_integer iopt, 
	fort_int ntcmax, fort_int ntc, fort_int ndetmax, fort_int nuse, g_variable *spectrum)
{
    char ebuff[100];
    long i, j, ii;
    fort_real *delt_work, *l2_work, *tthe_work, user_tables[30]; 
    fort_int *idat, *idat_t, *spec_work, nuser[30], ispec;
    fort_real duration, l1, l2, tthe, delt, fi, yval, yval_max, denom;
    f_spectrum_struct *fss;
    fort_int quick, errcode(0), nspec;

    nspec = iend - istart + 1;
    for(i=0; i<30; i++) 
    {
	nuser[i] = 0;
    }
    fss = initFSpectrum(ntc, 30);
    delt_work = (fort_real*) M_alloc(ndetmax*sizeof(fort_real));
    l2_work = (fort_real*) M_alloc(ndetmax*sizeof(fort_real));
    spec_work = (fort_int*) M_alloc(ndetmax*sizeof(fort_int));
    tthe_work = (fort_real*) M_alloc(ndetmax*sizeof(fort_real));
    idat = (fort_int*) M_alloc(ntcmax*sizeof(fort_int));
    idat_t = (fort_int*) M_alloc(ntcmax*sizeof(fort_int));
    quick = 0;
    ispec = istart;
    READ_DATA(file, errcode, ispec, delt_work, spec_work, tthe_work, l2_work,
	      ndetmax, fss->x, idat, ntcmax, fss->l1, fss->l2, fss->tt, fss->d, fss->fi, fss->tl, fss->dn,
	      fss->tm, fss->un, fss->in, fss->rn, fss->ut, nuse, quick);
    if ( errcode != 0 ) 
    {
	sprintf(ebuff, "error code %i from %i", errcode, ispec);
	error_add("READ_DATA in multiget", ebuff, "");
	return errcode;
    }
    fss->nut = nuse;
    cat(dec(istart), "-", dec(iend), fss->sn);
    fss->f = file;
    fss->xl = unit_name(2, 'x');
    fss->yl = unit_name(2, 'y');
    quick = 1;
    for(i=istart+1; i<=iend; i++)
    {
 	ispec = i;
    	READ_DATA(file, errcode, ispec, delt_work, spec_work, tthe_work, l2_work,
	          ndetmax, fss->x, idat_t, ntcmax, l1, l2, tthe, delt, fi, fss->tl, duration,
	          fss->tm, fss->un, fss->in, fss->rn, user_tables, nuse, quick);
        if ( errcode != 0 ) 
        {
	    sprintf(ebuff, "error code %i from %i", errcode, ispec);
	    error_add("READ_DATA in multiget", ebuff, "");
	    return errcode;
        }
	fss->dn = fss->dn + duration;
	fss->d = fss->d + delt;
	fss->fi = fss->fi + fi;
	fss->l1 = fss->l1 + l1;
	fss->l2 = fss->l2 + l2;
	fss->tt = fss->tt + tthe;
	for(j=0; j<ntc; j++)
	{
	    idat[j] += idat_t[j];
	}
	for(j=0; j<nuse; j++)
	{
	    (nuser[j])++;
	    fss->ut[j] = fss->ut[j] + user_tables[j];
	}
	fss->nut = max(fss->nut, nuse);
    }
    fss->dn = fss->dn / (fort_real)nspec;
    fss->d = fss->d / (fort_real)nspec;
    fss->l1 = fss->l1 / (fort_real)nspec;
    fss->l2 = fss->l2 / (fort_real)nspec;
    fss->tt = fss->tt / (fort_real)nspec;
    fss->fi = fss->fi / (fort_real)nspec;
    yval_max = 0.0;
    for(i=0; i<ntc; i++)
    {
	ii = i + 1;
	denom = fss->x[i+1] - fss->x[i];
/* first Y point is dummy in get routines, whereas we will have last Y point a dummy */
	yval = (fort_real)idat[ii] / denom;
	fss->y[i] = yval;
	fss->e[i] = sqrt((double)idat[ii]) / denom;
	yval_max = ( yval > yval_max ? yval : yval_max );
    }
    fss->ym = yval_max;
/* average user parameters */
    for(j=0; nuser[j] > 0; j++)
    {
	fss->ut[j] = fss->ut[j] / (fort_real)nuser[j];
    }
    fss->ym = 0.0;
    for(i=0; i<fss->n; i++)
    {
	fss->ym = max(fss->ym, fss->y[i]);
    }
    *spectrum = createFSpectrum(fss);
    deleteFSpectrum(fss);
    return 0;
}

#endif

/* now call getpari etc. */
/* they return codes of: 0 = found, 3 = not found, anything else = error */
int getSNSItem(String& runid, GenieString& item, GenieVariable& item_value)
{
    static void *buffer = new fort_real[20000];    
    fort_int length_in = 20000, length_out, errcode(0);
    enum {Getpari, Getparr, Getparc} parType;
    String citem, message;
    int i;
    parType = Getpari;
    item.upcase();		// all CRPT names are in upper case
    PGDESCRIPTOR(frunid, runid);
    PGDESCRIPTOR(fitem, item);
    PGDESCRIPTOR(fcitem, citem);
    GETPARI(frunid, fitem, (fort_int*)buffer, length_in, length_out, errcode);
    if (errcode == 3)
    {
	clear_errors();		// Get rid of "cannot find parameter" error message from GETPARI
    	parType = Getparr;
	GETPARR(frunid, fitem, (fort_real*)buffer, length_in, length_out, errcode);
    }
    if (errcode == 3)
    {
	clear_errors();		// Get rid of "cannot find parameter" error message from GETPARR
    	parType = Getparc;
    	GETPARC(frunid, fitem, fcitem, length_in, errcode);
        citem = fcitem.spot;
    }
    if (errcode != 0)
    {
        if (errcode == 3)
        {
            clear_errors();     // Get rid of "cannot find parameter" error message from GETPARC
        }
	cat("Cannot find item ", item, message);
	error_add("GETITEM", message, "");
	return errcode;
    }
    switch(parType)
    {
    	case Getparc:
	{
	    item_value.become(GenieString::toOOP(citem));
	}
	break;

	case Getpari:
  	{
	    if (length_out == 1)
	    {
		item_value = GenieInteger(*(fort_int*)buffer);
		break;
	    }
	    GenieIntegerArray gia(length_out);
	    fort_int* itemp = (fort_int*)buffer;
	    for(i=0; i<length_out; i++)
	    {
		gia(i) = (g_integer)itemp[i];
	    }
	    item_value = gia;
	}
	break;

	case Getparr:
 	{
	    if (length_out == 1)
	    {
		item_value = GenieReal(*(fort_real*)buffer);
		break;
	    }
	    GenieRealArray gra(length_out);
	    fort_real* rtemp = (fort_real*)buffer;
	    for(i=0; i<length_out; i++)
	    {
		gra(i) = (g_real)rtemp[i];
	    }
	    item_value = gra;
	}
	break;

	default:
	{
	    error_add("GETITEM", "invalid item type", "");
	}
	break;
    }
    return errcode;
}

void TRUNCATE_FILE(STRING& file)
{
    int flags = O_RDWR | O_TRUNC;
    int fd = open(translatePath(file.spot, 'c'), flags, 0644);
    close(fd);
    return;
}

static int check_and_open(String& runid, fort_int& ntc, fort_int& ndet, fort_int& nuse)
{
    fort_int errcode(0);
    OPEN_DATA_FILE(runid, ntc, ndet, nuse, errcode);
    if (errcode != 0)
    {
	String mess = String("Cannon open file ") + runid;
        error_add("OPEN_DATA_FILE", mess, "Check file exists and is readable");
	return 1;
    }
    else
    {
        return 0;
    }
}

void GINFO_C(STRING& imessage)
{
//    FILE* file = fdopen(fd, "w");
//    FILE* file = stderr;
//    fwrite(imessage, sizeof(char), ilen, file);
//    putc('\n', file);
    gerr << "GINFO: " << imessage.spot << endl;
    return;
}

// iarray contains all the spectrum numbers
int focusSpectrumRange(g_integer iverbose, GenieWorkspace& wksp, GenieWorkspace& specpars, 
		       	GenieWorkspace& detpars, g_integer iunit, String& file, g_integer* iarray, int len)
{
    ISISRawFile raw_file;
    if (raw_file.open(file) != 0)
    {
	gerr << "focusSpectrumRange: cannot open \"" << file << "\"" << endl;
	return 1;
    }
    raw_file.readSpectrum(wksp, iarray[0]);
    translatePath(file, 'f');
    fort_int ntcold, ndet, nuse;
    if (check_and_open(file, ntcold, ndet, nuse) != 0)
    {
	gerr << "focusSpectrumRange: error opening \"" << file << "\"" << endl;
	return 1;
    }
    fort_int ndetmax(ndet+1), ntcoldmax(ntcold+1);
    fort_real l1, l2, tthe, delt, fi, duration;
    fort_int quick(0), errcode(0);
    fort_int* spec_work = new fort_int[ndetmax];
    fort_int* idat = new fort_int[ntcoldmax];		// for counts
    fort_real* delt_work = new fort_real[ndetmax];
    fort_real* tthe_work = new fort_real[ndetmax];
    fort_real* l2_work = new fort_real[ndetmax];
    fort_real* tcb = new fort_real[ntcoldmax];
    const char *xunit_out, *yunit_out;
    double denom, ymax, emin(0.0), emax(0.0);
    g_real efixed(0.0);
    g_integer emode(0), ntc;
    
    fort_real* user_tables = new fort_real[nuse];
    String run_title, comb_time, user_name, inst_name, spec, run_no;

    g_integer ntcnew = ntcold;
    g_integer ntcfocus(0);
    GenieRealArray gv_twotheta, gv_l1, gv_l2, gv_efixed; 
    GenieIntegerArray gv_emode;
    g_real* xold = new g_real[ntcold+1];
    g_real* yold = new g_real[ntcold];
    g_real* eold = new g_real[ntcold];
    g_real* xnew = new g_real[ntcnew+1];
    g_real* ynew = new g_real[ntcnew];
    g_real* enew = new g_real[ntcnew];
    g_real* yfocus = new g_real[ntcnew];
    g_real* efocus = new g_real[ntcnew];
    if (specpars.defined()) 
    {
        gv_twotheta = specpars("TWOTHETA");
	gv_emode = specpars("EMODE");
	gv_efixed = specpars("EFIXED");
	gv_l1 = specpars("L1");
	gv_l2 = specpars("L2");
    }
//
// Set up parameters for detector
//
    double gp(0.0), gsig0(0.0), gt(0.0), lam0(0.0), wt(0.0), 
		rho(0.0), ct(0.0), wsig0(0.0);
    if (detpars.defined())
    {
        gp = detpars("PRESSURE");                   // Gas pressure (atms)
        gsig0 = detpars("GAS_SIGMA0");              // Gas cross section at LAM0 (cm-2)
        gt = detpars("PATHLENGTH");                 // Gas path length (cm)
        lam0 = detpars("WAVELENGTH");               // Characteristic wavelength (cm)
        wt = detpars("WEIGHT");                     // Molecular weight of container atoms (g mol-1)
        rho = detpars("DENSITY");                   // Density of container (g cm-3)
        ct = detpars("THICKNESS");                  // Container wall thickness (cm)
        wsig0 = detpars("WALL_SIGMA0");             // Wall cross section at LAM0 (cm-2)
    }
    int i, j, jj;
    for(j=0; j<ntcnew; j++)
    {
	yfocus[j] = 0.0;
	efocus[j] = 0.0;
    }
    g_real* xfocus = 0;

    for(i=0; i<len; i++)
    {
// Read in each spectrum to xold, yold, eold
	quick = ( i != 0 ? 1 : 0 );
        READ_DATA(file,errcode,iarray[i],delt_work,spec_work,
	          tthe_work,l2_work,ndetmax,tcb,idat,ntcoldmax,
	          l1,l2,tthe,delt,fi,run_title,duration,
  	          comb_time,user_name,inst_name,run_no, 
	          user_tables, nuse, quick);
	if (errcode != 0)
	{
	    gerr << "focusSpectrumRange: error returned from READ_DATA" << endl;
	    return 1;
 	}
	for(j=0; j<ntcold; j++)
	{
/* first Y point is dummy in get routines */
	    jj = j + 1;
	    denom = tcb[j+1] - tcb[j];
	    xold[j] = tcb[j];
	    yold[j] = (double)idat[jj] / denom;
	    eold[j] = sqrt((double)idat[jj]) / denom;
	}
   	xold[ntcold] = tcb[ntcold];
 	if (gv_twotheta.defined())
	   	tthe = (float)gv_twotheta[i];
 	if (gv_l1.defined())
		l1 = (float)gv_l1[i];
 	if (gv_l2.defined())
		l2 = (float)gv_l2[i];
 	if (gv_emode.defined())
		emode = gv_emode[i];
 	if (gv_efixed.defined())
		efixed = gv_efixed[i];
        ntc = ntcold;
   	if (iverbose)
	{
	    gout << "FOCUS: Using l1, l2, twotheta = " << l1 << ", " << l2 << ", " << tthe 
		 << " for spectrum " << iarray[i] << endl;
	}
// First convert to wavelength (unit code "2" -> "3") and apply detector correction
	if (convert_units(xold, yold, eold, &ntc, unit_name(2, 'x'), &xunit_out, unit_name(2, 'y'),
		&yunit_out, 3, l1, l2, emode, efixed, tthe, &ymax, emin, emax, delt) != 0)
	{
	    gerr << "focusSpectrumRange: units error" << endl;
	    return 1;
	}
	if (detpars.defined())
	{
	    detector_efficiency(xold, yold, eold, ntc, gp, gsig0, gt, lam0, wt, rho, ct, wsig0);
	}
// Now convert to whatever we asked for (unit code "3" -> "iunit")
	if (convert_units(xold, yold, eold, &ntc, unit_name(3, 'x'), &xunit_out, unit_name(3, 'y'),
		&yunit_out, iunit, l1, l2, emode, efixed, tthe, &ymax, emin, emax, delt) != 0)
	{
	    gerr << "focusSpectrumRange: units error" << endl;
	    return 1;
	}
// Now generate a reference range to rebin each to 
        if (xfocus == 0) 
	{
	    ntcfocus = ntcold;
	    xfocus = new g_real[ntcfocus+1];
	    for(j=0; j<ntcold+1; j++)
	    {
		xfocus[j] = xold[j];
	    }
	}
	ntcnew = overlap_range(xold, ntcold+1, xfocus, ntcfocus+1, xnew) - 1;
// Now rebin each party to the reference
// First current focussed range
    	if (rebin(xfocus, yfocus, efocus, ntcfocus, xnew, ynew, enew, ntcnew, ymax) != 0)
	{
	    gerr << "focusSpectrumRange: rebin error" << endl;
	    return 1;
	}
   	ntcfocus = ntcnew;
	memcpy(xfocus, xnew, (ntcnew+1)*sizeof(g_real));
	memcpy(yfocus, ynew, ntcnew*sizeof(g_real));
	memcpy(efocus, enew, ntcnew*sizeof(g_real));
// Now read in range
    	if (rebin(xold, yold, eold, (g_integer)ntcold, xnew, ynew, enew, ntcnew, ymax) != 0)
	{
	    gerr << "focusSpectrumRange: rebin error" << endl;
	    return 1;
	}
// Now add up result
	for(j=0; j<ntcfocus; j++)
	{
	    yfocus[j] += ynew[j];
	    efocus[j] = sqrt(efocus[j] * efocus[j] + enew[j] * enew[j]);
	}
    }
    if (ntcfocus == 0)
        *g_err << "FOCUS: No overlap between spectra specified" << endl;
    wksp("X", GenieRealArray(xfocus, ntcfocus+1));
    wksp("Y", GenieRealArray(yfocus, ntcfocus));
    wksp("E", GenieRealArray(efocus, ntcfocus));
    wksp("XLABEL", unit_name(iunit, 'x'));
    wksp("YLABEL", unit_name(iunit, 'y'));
    char spec_no[128];
    sprintf(spec_no, "%ld - %ld", iarray[0], iarray[len-1]);
    wksp("SPEC_NO", spec_no);
    delete[] xold; delete[] yold; delete[] eold;
    delete[] xnew; delete[] ynew; delete[] enew;
    delete[] yfocus; delete[] efocus; delete[] xfocus;
    delete[] spec_work; delete[] idat; delete[] delt_work; delete[] tthe_work; 
    delete[] l2_work; delete[] tcb; delete[] user_tables;
    return 0;
}

// ya is pre-allocated to length npoints
// points read from RAW file, grouped into bins of size NGROUP
// Returns grouped number of points, -1 on error
int get_spectrum(String& runid, fort_int spec, fort_int* ya, fort_int npoints, int ngroup)
{
    fort_int *raw_ya, ntc, ndet, nuse, errcode(0), nos(1);
    STRING runid_i;
    if (check_and_open(runid, ntc, ndet, nuse) != 0)
    {
	return -1;
    }
    raw_ya = new fort_int[ntc];
    runid_i.length = runid.length();
    strcpy(runid_i.spot, runid);
    GETDAT(runid_i, spec, nos, raw_ya, ntc, errcode);
    if (errcode == 4)
    {
	*g_err << "GETDAT: Invalid spectrum number (" << spec << ") or number of spectra (" << nos << ") requested" << endl;
	return -1;
    }
    if (errcode != 0)
    {
	*g_err << "GETDAT: Error code " << errcode << " while reading " << nos << " spectra beginning with " << spec << endl;
	return -1;
    }
// Now group the bins
// first Y point is dummy, hence +1 in sum
    int sum, igroup(ngroup), iold(0), inew(0), i;
    while((igroup == ngroup) && (iold < ntc) && (inew < npoints))
    {
	igroup = ( iold + ngroup > ntc ? ntc - iold : ngroup );
	sum = 0;
	for(i=iold; i<iold+igroup; i++)
	{
	    sum += raw_ya[i+1];
	}
	ya[inew] = sum;
	iold += igroup;
	inew++;
    }
    return inew;
}

/*
 * determine an X range containing the intersection of x1 and x2
 * xnew should be of size max(n1,n2)
 * returns: new size
 */
int overlap_range(g_real* x1, int n1, g_real* x2, int n2, g_real* xnew)
{
    double xmax, *xcurr;
    int i, nmax;
    if (n1 <= 0 || n2 <= 0)
    {
	gerr << "overlap_range: invalid input parameters" << endl;
    }
    if (x1[0] < x2[0])
    {
	xcurr = x2;
	nmax = n2;
	xmax = x1[n1-1];
    } 
    else
    {
	xcurr = x1;
	nmax = n1;
	xmax = x2[n2-1];
    }
// copy over common region
    for(i=0; (i < nmax) && (xcurr[i] < xmax); i++)
    {
	xnew[i] = xcurr[i];
    }
// add in final boundary if we didn't copy everything in xcurr[]
    if (i < nmax)
    {
	xnew[i] = xmax;
	i++;
    }
    return i;
}

// Input in wavelength units
int detector_efficiency(g_real* x, g_real* y, g_real* e, int npt, 
	double gp, double gsig0, double gt, double lam0, double wt, double rho, double ct, double wsig0)
{
        static const double vm = 2.24136e4;                // Molar volume of gas cm3 mol-1
        static const double ac = 6.022169e23;              // Avogadro's constant mol-1

        double gn = ac*gp/vm;                              // Number density of gas
        double wn = ac*rho/wt;                             // Number density of wall
        double sgn = gn*gsig0*gt/lam0;                     // Exponential term for gas
        double swn = wn*wsig0*ct/lam0;                     // Exponential term for wall    

    double corr, sg, sw;
    for(int i=0; i<npt; i++)
    {
	  sg = sgn*x[i];                // multiply exponent by wavelength
          sw = swn*x[i];                // multiply exponent by wavelength
          corr = (1.0-exp(-sg))*exp(-sw);    // Efficiency calculation
          y[i] /= corr;    // Efficiency calculation
          e[i] /= corr;    // Efficiency calculation
    }
    return 0;
}

static int directoryGenieIntermediate(GenieWorkspace& info, GenieString& file, GenieInteger full_listing)
{
	    GenieVariableType item_type;
	    char *label, *comment, *user, *date;
	    GenieIntermediateFile int_file;
	    if (!int_file.open(file, GenieIntermediateFile::Input))
	    {
		return 1;
	    }
	    int len = int_file.numberOfBlocks();
	    GenieString header = int_file.header();
	    GenieStringArray labels(len), comments(len), users(len), dates(len), types(len);
	    for(int i=0; i<len; i++)
	    {
	        int_file.seek(i);
		int_file.readLabel(label);
		labels(i, label);
	 	item_type = int_file.skipItem(comment, user, date);
		if ((int)full_listing)
		{
		    comments(i, comment);
		    users(i, user);
		    dates(i, date);
		}
		types(i, GXVariableNames[item_type]);
		int_file.freeString(comment);
		int_file.freeString(date);
		int_file.freeString(user);
		int_file.freeString(label);
	    }
	    info("label", labels);
	    info("comment", comments);
	    info("user", users);
	    info("date", dates);
	    info("header", header);
	    info("type", types);
	return 0;
}

static int directoryGenieASCIIIntermediate(GenieWorkspace& info, GenieString& file, GenieInteger full_listing)
{
	    GenieVariable gv;
	    int j, nb(0);
	    static char buffer[256];
	    fstream fs;
	    fs.open(file, ios::in);
	    if (!fs)
	    {
		*g_err << "LIST: Cannot open \"" << file << "\"" << endl;
		return 1;
	    }
	    fs.getline(buffer, sizeof(buffer), '\n');	// skip header
// get block count
	    while(fs.getline(buffer, sizeof(buffer), '\n') && fs.good())
	    {
		    gv.readAscii(fs);				// skip variable
		    while(fs.good() && fs.get() != '#')
			;
		    nb++;
	    }
	    fs.clear();
	    fs.seekg(0, ios::beg);
	    GenieStringArray labels(nb), types(nb), comments(nb), users(nb), dates(nb);
	    fs.getline(buffer, sizeof(buffer), '\n');
	    info("header", buffer);
	    for(int i=0; i<nb; i++)
	    {
		fs.getline(buffer, sizeof(buffer), '\n');
		j = (buffer[0] == '#' ? 2 : 1);			// have we skipped '#'
		labels(i, buffer+j);
		gv.readAscii(fs);				// skip variable
	  	types(i, getClassString(gv.value()));
		while(fs.good() && fs.get() != '#')
		    ;
	    }
	    fs.close();
	    info("label", labels);
	    info("comment", comments);
	    info("user", users);
	    info("date", dates);
	    info("type", types);
    return 0;
}
