/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         g_module
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Callout for MODULE interface to calling user programs
 *
 */
#include "geniedefs.h"
#include <limits.h>

#include <iostream.h>
#ifdef _WIN32
#	include <io.h>
	char _itoa_buff[18];	/* _itoa can return 17 characters */
#	define dec(x) itoa(x, _itoa_buff, 10)	/* Seems to do the same thing !! */
#else
#	include <stream.h>
#endif /* _WIN32 */
#include <signal.h>
#include <setjmp.h>
#include "gsignals.h"
#include <gnustring.h>
#include "smalltalk.h"
#include "genievars.h"
#include "module.h"
#include "fortran_types.h"
#include "g_module.h"
#include "Genie_functions.h"
#include "callout_list.h"
#include "genie_cmodule.h"


//extern "C" 
//{
//    void for_rtl_init_(int*, char**);
//    int for_rtl_finish_();
//}
static const char* rcsid = "@(#)$Id: g_module.cxx,v 1.39.2.3 1997/09/03 17:01:16 faa Exp $";
USE_RCSID(rcsid);

typedef void (*module_func_t)(GenieWorkspace& pars, GenieWorkspace& res);

Module genie_modules;

static int execute_module(module_func_t funcptr, GenieWorkspace& pars, GenieWorkspace& res);

extern "C" {
int map_stdio();
int unmap_stdio();
}

static void map_fortran_stdio();
static void unmap_fortran_stdio();

// return nilOOP on error, arguments on success
OOP g_module(module_function_code function, OOP arguments)
{
    GenieCalloutList callout_list(arguments);
    OOP result(arguments);		// success by default

    BEGINSAVEBLOCK

    switch(function)
    {
	case Module_test:
	    break;

	case Module_load:
	{
	    GenieString file(callout_list, "FILE"), desc(callout_list, "DESC");
	    if (genie_modules.load(file, desc) == -1)
		result = nilOOP;
	}
	break;

	case Module_execute:
	{
	    ModuleFunctionType func_type;
	    GenieString func(callout_list, "FUNC"), type(callout_list, "TYPE");
	    GenieWorkspace pars(callout_list, "PARS"), res(callout_list, "RES");
	    if (type == "FORTRAN")
	    {
		func_type = FortranModuleFunction;
	    }
	    else
	    {
		func_type = CModuleFunction;
	    }
#if HAVE_DLFCN_H || defined(__VMS) || defined(__hpux)
	    module_func_t funcptr = (module_func_t)genie_modules.getFunctionPointer(func, func_type);
#else
// The system does not support dynamic libraries, so we can only return this function
// Modules can be tested by changing their function name to "module_test_function" and linking them into genie
	    module_func_t funcptr = module_test_function;
#endif
	    if (funcptr != 0)
	    {
// No need to call map_fortran_stdio() and unmap_fortran_stdio() now as we are one process
//		int argc = 1;
//	 	char* argv[] = { "gmain", 0 };
//		if (func_type == FortranModuleFunction)
//	  	    for_rtl_init_(&argc, argv);
		if (execute_module(funcptr, pars, res) != 0)
		    gerr << "MODULE: An error occurred while executing \"" << (const char*)func << "\"" << endl;
//		if (func_type == FortranModuleFunction)
//	  	    for_rtl_finish_();
	    }
	    else
	    {
		result = nilOOP;
	    }
	}
	break;

        case Module_list:
        {
            genie_modules.list(gout);
	    gout.flush();
        }
        break;

	case Module_unload:
	{
	    GenieString file(callout_list, "FILE");
	    if (genie_modules.unload(file) == -1)
		result = nilOOP;
	}
	break;

	default:
            PERROR("Module routines called with invalid function(g_module)");
            break;
    }

    ENDSAVEBLOCK

    return arguments;
}

static sigjmp_buf module_signal_context;

static void module_signal_handler(int sig)
{
    siglongjmp(module_signal_context, sig);
}

#ifdef _WIN32
#	define SIGBUS SIGUSR1
#endif

static int module_signals_to_handle[] = { SIGILL, SIGSEGV, SIGBUS, SIGFPE };
#define NMODULESIGNALS (sizeof(module_signals_to_handle) / sizeof(int))

#ifdef __VMS

#include <ssdef.h>
#include <stsdef.h>
#include <lib$routines.h>

static int vms_execute_module(module_func_t funcptr, GenieWorkspace& pars, GenieWorkspace& res)
{
    lib$establish(lib$sig_to_ret);
    (*funcptr)(pars, res);
    return SS$_NORMAL;
}

// 0 on success, otherwise signal causing problem
static int execute_module(module_func_t funcptr, GenieWorkspace& pars, GenieWorkspace& res)
{
    volatile signal_handler old_sig_handler[32];
    int i;
    unsigned long status;
// Set signals to default action
    for(i=1; i<32; i++)
    {
	if ( (old_sig_handler[i] = genie_signal(i, SIG_DFL)) == SIG_ERR )
	    old_sig_handler[i] = SIG_DFL;
    }
    status = vms_execute_module(funcptr, pars, res);
    if ( (status & 07) != STS$K_SUCCESS )
    {
	*g_err << "MODULE: exception raised during module execution" << endl;
// Modify from Sever error to error, otherwise program will exit in call to lib$signal()
	if ( (status & 07) == STS$K_SEVERE )	
	{
	    status &= ~STS$K_SEVERE;	// clear severe error
	    status |= STS$K_ERROR;	// signal error instead
	}
	lib$signal(status);
    }
// Restore signals
    for(i=1; i<32; i++)
    {
	genie_signal(i, old_sig_handler[i]);
    }
    return status == SS$_NORMAL ? 0 : 1;
}

#else

// 0 on success, otherwise signal causing problem
static int execute_module(module_func_t funcptr, GenieWorkspace& pars, GenieWorkspace& res)
{
    unsigned i;
    volatile int sigret;
    volatile signal_handler old_sig_handler[NMODULESIGNALS];
// Attempt to catch likely signals
    for(i=0; i<NMODULESIGNALS; i++)
    {
	old_sig_handler[i] = genie_signal(module_signals_to_handle[i], module_signal_handler);
    }
    if ( (sigret = sigsetjmp(module_signal_context, 1)) == 0 )
    {
    	(*funcptr)(pars, res);
    }
    else // signalled due to some error
    {
  	gerr << "MODULE: Signal raised in user code during execution" << endl;
	gerr << printsig(sigret) << endl;
    }
// restore signals
    for(i=0; i<NMODULESIGNALS; i++)
    {
	genie_signal(module_signals_to_handle[i], old_sig_handler[i]);
    }
    return sigret;
}

#endif /* __VMS */

void g3_module_get_string(GenieWorkspace& pars, const STRING& name, STRING& val) 
{ 
    String message;
    GenieString gs = pars(name.spot);
    if (gs.defined())
    {
        strcpy(val.spot, (const char*)gs);
    }
    else
    {
	cat("No string parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
        strcpy(val.spot, "<Undefined>");
    }
    val.length = strlen(val.spot);
}

void g3_module_put_string(GenieWorkspace& res, const STRING& name, const STRING& val) { res(name.spot, val.spot); }

void g3_module_string_array_alloc(const fort_int& length, void*& handle)
{
    GenieStringArray* gsa = new GenieStringArray(length);
    handle = gsa;
}

void g3_module_string_array_add(void*& handle, const fort_int& element, const STRING& val)
{
    GenieStringArray* gsa = (GenieStringArray*)handle;
    (*gsa)(element - 1, val.spot);			// FORTRAN arrays start at 1, C arrays at 0
}

void g3_module_string_array_free(GenieWorkspace& res, const STRING& name, void*& handle)
{
    GenieStringArray* gsa = (GenieStringArray*)handle;
    res(name.spot, *gsa);
    delete gsa;
}

void g3_module_get_real(GenieWorkspace& pars, const STRING& name, fort_real& val) 
{ 
    String message;
    GenieReal gr = pars(name.spot);
    if (gr.defined())
    {
	val = (float)gr;
    }
    else
    {
	cat("No real parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = FLT_MAX;
    }
}

void g3_module_get_double(GenieWorkspace& pars, const STRING& name, fort_double& val) 
{ 
    String message;
    GenieReal gr = pars(name.spot);
    if (gr.defined())
    {
	val = (g_real)gr;
    }
    else
    {
	cat("No real parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = DBL_MAX;
    }
}

void g3_module_put_real(GenieWorkspace& res, const STRING& name, const fort_real& val) { res(name.spot, (g_real)val); }

void g3_module_put_double(GenieWorkspace& res, const STRING& name, const fort_double& val) { res(name.spot, (g_real)val); }

void g3_module_get_int(GenieWorkspace& pars, const STRING& name, fort_int& val) 
{ 
    String message;
    GenieInteger gi = pars(name.spot);
    if (gi.defined())
    {
	val = (g_integer)gi;
    }
    else
    {
	cat("No integer parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = INT_MAX;
    }
}

void g3_module_put_int(GenieWorkspace& res, const STRING& name, const fort_int& val) { res(name.spot, (g_integer)val); }

void g3_module_get_real_array(GenieWorkspace& pars, const STRING& name, fort_real* val, fort_int& len)
{
    g3_module_get_real_array_alloc(pars, name, &val, len);
}

// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_real_array_alloc(GenieWorkspace& pars, const STRING& name, fort_real** val, fort_int& len)
{
    String message;
    GenieRealArray gra;
    int l_len;
    gra.become(pars(name.spot));
    if (!gra.defined())
    {
	cat("No real array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gra.length();
	*val = (fort_real*)malloc(l_len*sizeof(fort_real));
    }
    else
    {
        l_len = (len > gra.length() ? gra.length() : len);
    }
    for(int i=0; i<l_len; i++)
    {
	(*val)[i] = (float)gra[i];
    }
    len = gra.length();
}

void g3_module_get_double_array(GenieWorkspace& pars, const STRING& name, fort_double* val, fort_int& len)
{
    g3_module_get_double_array_alloc(pars, name, &val, len);
}
// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_double_array_alloc(GenieWorkspace& pars, const STRING& name, fort_double** val, fort_int& len)
{
    String message;
    int l_len;
    GenieRealArray gra;
    gra.become(pars(name.spot));
    if (!gra.defined())
    {
	cat("No real array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gra.length();
	*val = (fort_double*)malloc(l_len*sizeof(fort_double));
    }
    else
    {
        l_len = (len > gra.length() ? gra.length() : len);
    }
    for(int i=0; i<l_len; i++)
    {
	(*val)[i] = gra[i];
    }
    len = gra.length();
}

void g3_module_get_int_array(GenieWorkspace& pars, const STRING& name, fort_int* val, fort_int& len)
{
    g3_module_get_int_array_alloc(pars, name, &val, len);
}

// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_int_array_alloc(GenieWorkspace& pars, const STRING& name, fort_int** val, fort_int& len)
{
    String message;
    GenieIntegerArray gia;
    int l_len;
    gia.become(pars(name.spot));
    if (!gia.defined())
    {
	cat("No integer array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gia.length();
	*val = (fort_int*)malloc(l_len*sizeof(fort_int));
    }
    else
    {
        l_len = (len > gia.length() ? gia.length() : len);
    }
    for(int i=0; i<l_len; i++)
    {
        (*val)[i] = gia[i];
    }
    len = gia.length();
}

void g3_module_put_double_array(GenieWorkspace& pars, const STRING& name, const fort_double* val, const fort_int& len)
{
    GenieRealArray gra(val, len);
    pars(name.spot, gra);
}

void g3_module_put_real_array(GenieWorkspace& pars, const STRING& name, const fort_real* val, const fort_int& len)
{
    GenieRealArray gra(val, len);
    pars(name.spot, gra);
}

fort_int g3_module_put_nd_real_array(GenieWorkspace& pars, const STRING& name, const fort_real* val, 
					const fort_int* dims_array, const fort_int& ndims)
{
    GenieRealArray gra(val, dims_array, ndims);
    pars(name.spot, gra);
    return 0;
}

void g3_module_put_int_array(GenieWorkspace& pars, const STRING& name, const fort_int* val, const fort_int& len)
{
    GenieIntegerArray gia(val, len);
    pars(name.spot, gia);
}

// See genie_cmodule.h for options flag
void g3_module_print(const STRING& s, const fort_int& option)
{
    if (option == CMODULE_PRINT_NORMAL)
        gout.write(s.spot, strlen(s.spot)) << endl;
    else if (option == CMODULE_PRINT_INFORMATION)
        ginf.write(s.spot, strlen(s.spot)) << endl;
    else if (option == CMODULE_PRINT_ERROR)
        g_err->write(s.spot, strlen(s.spot)) << endl;
    return;
}

// Get the major and minor version ID of the interface we are using
fort_int g3_module_get_version(fort_int& major, fort_int& minor)
{
    major = GENIE_INTERFACE_MAJOR_VERSION;
    minor = GENIE_INTERFACE_MINOR_VERSION;
    return 0;
}

// Check to see if "major" & "minor" are compatable with the interface we are currently using
// Return 1 if OK, 0 if not
fort_int g3_module_version_ok(const fort_int& major, const fort_int& minor)
{
    if ( (major == GENIE_INTERFACE_MAJOR_VERSION) && (minor <= GENIE_INTERFACE_MINOR_VERSION) )
    {
	return 1;		// All OK
    }
    else
    {
	*g_err  << "Interface version mismatch: this MODULE requires (at least) V" << major << "." << minor << endl
		<< "while GENIE is providing only V" << GENIE_INTERFACE_MAJOR_VERSION << "." << GENIE_INTERFACE_MINOR_VERSION << endl;
	ginf << "To use this MODULE in GENIE you could try to rebuild it by" << endl
             << "executing MODULE/COMPILE on the original program source;"  << endl
	     << "however, you may find you are missing some library functions" << endl;
	return 0;		// Not OK
    }
}
    
static void map_fortran_stdio()
{
    static const char* local_tty = OOPToString(typeNameToOOP("gXUUlocalUtty"));
    static fort_int len_local_tty = strlen(local_tty);
#ifndef HAVE_LIBF2C
    f_map_fortran_stdio(local_tty, len_local_tty);	// doesn't work for f2c
#endif
}

static void unmap_fortran_stdio()
{
#ifndef HAVE_LIBF2C
    f_unmap_fortran_stdio();				// doesn't work for f2c
#endif
}

// On input, tty_len is the number of chacters in "itty"
// On output, it should be set to the length of itty
void g3_module_open_stdio(char* itty, fort_int& tty_len)
{
    static const char* local_tty = OOPToString(typeNameToOOP("gXUUlocalUtty"));
    static int len_local_tty = strlen(local_tty);
    strncpy(itty, local_tty, tty_len);		// provided tty_len is big enough, we get NULL termination
    tty_len = (len_local_tty + 1 > tty_len ? tty_len : len_local_tty);
    return;
}

void g3_module_close_stdio(fort_int& stdio_unit)
{
    return;
}

/*
 * Re-assign I/O streams so FORTRAN can freely use units 5 and 6
 * Unfortunatelf, it doesn't work :-(
 */

static int old_stdin(-1), old_stdout(-1), old_stderr(-1); 	// initialise as an invalid descriptor

int map_stdio()
{
// Save current I/O channels
    if ((old_stdin = dup(0)) == -1)
	perror("map_stdio: dup()");
    if ((old_stdout = dup(1)) == -1)
	perror("map_stdio: dup()");
    if ((old_stderr = dup(2)) == -1)
	perror("map_stdio: dup()");
    return 0;
}

int unmap_stdio()
{
// Restore old I/O channels
    if (dup2(0, old_stdin) == -1)
        perror("unmap_stdio: dup2()");
    if (dup2(1, old_stdout) == -1)
        perror("unmap_stdio: dup2()");
    if (dup2(2, old_stderr) == -1)
        perror("unmap_stdio: dup2()");
// Close temporary channels so we don't run out of descriptors later on
    if (close(old_stdin) == -1)
	perror("unmap_stdio: close()");
    if (close(old_stdout) == -1)
	perror("unmap_stdio: close()");
    if (close(old_stderr) == -1)
	perror("unmap_stdio: close()");
    old_stdin = old_stdout = old_stderr = -1;		// Invalidate descriptors
    return 0;
}
