/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdlib.h>

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#ifdef _WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include <stdio.h>		// Only for sprintf()
#include <string.h>
// For getcputimes
//#include <sys/times.h>
//#include <unistd.h>
//#include <sys/resource.h>

#ifndef _WIN32
#	include <sys/wait.h>
#endif

#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#if defined(__VMS) && !defined(__VMS_POSIX)
#    define fcntl(a,b,c) 0
#    include <unixio.h>
#endif

#if	( defined(__VMS) && __VMS_VER < 70000000 ) || defined(_WIN32)
#	define WEXITSTATUS(__arg)	__arg
#	define WIFEXITED(__arg)	(__arg != 1)
#endif /* ( defined(__VMS) && __VMS_VER < 70000000 ) || _WIN32*/

#include "gsignals.h"

static const char* rcsid = "@(#)$Id: g_system.cxx,v 1.46.2.1 1997/08/24 23:13:34 faa Exp $";
USE_RCSID(rcsid);

#undef MIN		// lose an unecessary extra definition of this !!

#include "smalltalk.h"
#include "genie_types.h"
#include "g_system.h"
#include "genievars.h"
#include "genielog.h"
#include "genie_tmpfile.h"

#include "callout_list.h"

static void print_parameters( g_variable arguments );
static void run_test( g_variable arguments );
static void clear_pipe();

//ostream Gout;
//ostream Gerr;
//istream Gin;

extern char *symToCString( OOP oopSym );

bool gcTortureInCallout = false;
int gcTortureFacility = -1;

// Return oopParameters on success, nilOOP on failure
OOP g_system( system_function_code function, OOP oopParameters )
{

    OOP result(oopParameters);		// default: success
    GenieCalloutList callout_list(oopParameters);

    BEGINSAVEBLOCK

    char buffer[LONGCHARBUFF];

    switch(function){

    case Test:
    {
        cout << "Called Test: entrypoint in g_system.cc" << endl;
        break;
    }

    case SystemCommand:	     //  Send a command to the system for execution
    {
// Set res to 0 on success, 1 on error
        int status;
        GenieString command(callout_list, "command");
	GenieInteger res(callout_list, "res");

// -- the system() call doesn't like signals so restore to defaults
        Restore_signals();   

        // System return wait(2) code
        status = system(command);
	if (status == -1)
	{
	    PERROR("g_system(SystemCommand)");
	    result = nilOOP;
	}
	else if ( (status != 0) && WIFEXITED(status) )
	{
            *g_err << "SYSTEM: system() returned unusual status = " << WEXITSTATUS(status) << endl;
	    *g_err << "SYSTEM: executed command was \"" << command << "\"" << endl;
	    result = nilOOP;
	}
	res = (status != 0 ? 1 : 0);
        Set_signals();	       // -- Return signal state to normal for GENIE
        break;
    }

    case GetEnv:	        // Read an environment variable
    {
        GenieString name(callout_list, "name");
        GenieString value(callout_list, "value");
        char* cp(0);
	if ((const char*)name != NULL && (cp = getenv(name)) != NULL)
	{
	    value = cp;
	}
        break;
    }

    case GCTorture:
    {
	GenieInteger flag(callout_list, "flag"), facility(callout_list, "facility"); 
#if defined(GC_TORTURE)
	if ((int)flag == 0)
	{
	    gcTortureInCallout = false;
	    gcTortureFacility = -1;
	}
 	else if ((int)flag == 1)
	{
	    gcTortureInCallout = true;
	    gcTortureFacility = facility;
	}
#else
	gerr << "GCTORTURE unavaliable - Smalltalk was not configured with the \"--enable-gc-torture\" option" << endl;
	result = nilOOP;
#endif /* GC_TORTURE */
	break;
    }

    case GetGenieVersion:
    {
	    GenieString info(callout_list, "VERSION_INFO");
	    info = (const char*)version;
	    break;
    }

    case GenieLog:
    {
	    GenieString message(callout_list, "MESSAGE");
	    if (genielog(message) == -1)
	    {
	  	result = nilOOP;
		*g_err << "GENIELOG: error sending message" << endl;
	    }
	    break;
    }

    case MakeTempFile:
    {
	GenieString prefix(callout_list, "PREFIX"), directory(callout_list, "DIRECTORY"), suffix(callout_list, "SUFFIX"), tempfile(callout_list, "TEMPFILE");
	const char* tmpfile = genie_tmpfile(prefix, suffix, directory);
	if (tmpfile != NULL)
  	{
	    tempfile = tmpfile;
	}
	else
	{
	    *g_err << "TMPFILE: Unable to generate unique temporary filename" << endl;
	    result = nilOOP;
	}
	break;
    }

    case GetCPUTimes:
    {
#if 0
	GenieWorkspace structtms(callout_list, "STRUCTTMS");
	struct tms struct_tms;
	times(&struct_tms);
	structtms("user", struct_tms.tms_utime);
	structtms("system", struct_tms.tms_stime);
	structtms("user_child", struct_tms.tms_cutime);
	structtms("system_child", struct_tms.tms_cstime);
	struct rusage usage;
	if (getrusage(RUSAGE_SELF, &usage) == 0)
	{
	    structtms("maxrss", usage.ru_maxrss);
	    structtms("pagefaults", usage.ru_majflt);
	    structtms("pagereclaims", usage.ru_majflt);
	    structtms("nswap", usage.ru_nswap);
	}
	else
	{
	    perror("getrusage");
	    result = nilOOP;
	}
#endif
	break;
    }

    case RaiseSignal:
    {
	GenieInteger sig(callout_list, "SIG");
	if (raise(sig) != 0)
	{
	    *g_err << "RAISE: error raising signal " << (g_integer)sig << endl;
	    result = nilOOP;
	}
	break;
    }

    default:
	PERROR("unknown system callout");
	result = nilOOP;
        break;
    }
 
    ENDSAVEBLOCK

    return result;
}

