/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef G_TYPES_H
#define G_TYPES_H

#if HAVE_LIMITS_H
#include <limits.h>
#endif
#if HAVE_FLOAT_H
#include <float.h>
#endif
#if HAVE_VALUES_H
#include <values.h>
#endif

#if HAVE_NAN_H
#include <nan.h>
#endif

#ifndef DBL_MAX
#define DBL_MAX MAXDOUBLE
#endif
#ifndef FLT_MAX
#define FLT_MAX MAXFLOAT
#endif
#ifndef LONG_MAX
#define LONG_MAX MAXLONG
#endif
#ifndef INT_MAX
#define INT_MAX MAXINT
#endif

// How to handle Undefined in PGPLOT
#define UNDEFINED_PG_FLOAT      FLT_MAX
#define UNDEFINED_PG_INT        INT_MAX

class GenieWorkspace;
class GenieIntegerArray;

typedef long g_integer;
typedef double g_real;
typedef char* g_string;		// some routines take (const g_string)
typedef const char* c_g_string;
typedef char g_char;
typedef GenieWorkspace g_work;

typedef GenieIntegerArray GenieObjectID;
#define GENIE_OBJECTID_SIZE		2		/* number of array elements */
typedef g_integer* g_objectid;
enum GenieObjectIDInfo { ObjectPosition = 0, ObjectPicture = 1 };

// Macros to test for undefined values
static const g_integer UndefinedInteger = LONG_MAX;
#define IS_UNDEFINED_INTEGER(__val)	((g_integer)(__val) == UndefinedInteger)
#define UNDEFINE_INTEGER(__val)		((__val) = UndefinedInteger)	/* must return undefined value */
#define GINTEGER_EQUAL(__val1, __v1l2)	\
	( (__val1 != UndefinedInteger) && (__val2 != UndefinedInteger) && (__val1 == __val2) )

static const g_string UndefinedString = (const g_string)0;
#define IS_UNDEFINED_STRING(__val)	((const g_string)(__val) == UndefinedString)
#define UNDEFINE_STRING(__val)		((__val) = UndefinedString)

#if 0
static double UndefinedRealDummy =	DBL_MAX; 
static const g_real UndefinedReal = 	NaN(UndefinedRealDummy);		// Just Create a NaN
#define IS_UNDEFINED_REAL(__val)	IsNANorINF(__val)
#define UNDEFINE_REAL(__val)		NaN(__val)
#else
static const g_real UndefinedReal = 	DBL_MAX;
#define IS_UNDEFINED_REAL(__val)	((g_real)(__val) == UndefinedReal)
#define UNDEFINE_REAL(__val)		((__val) = UndefinedReal)	/* must return undefined value */
//#define GREAL_EQUAL(__val1, __val2)	\
//	( (__val1 != UndefinedReal) && (__val2 != UndefinedReal) && (fabs(__val1 - __val2) < 1e-6) )
#define GREAL_EQUAL(__val1, __val2)	\
	( (__val1 != UndefinedReal) && (__val2 != UndefinedReal) && (__val1 == __val2) )
#endif

// -1 is used as an error code for routines returning a "GenieVariableType"
// Start genie types at 1 so they are treated as "true"
// Always add new types to end so as not to break intermediate file reads
enum GenieVariableType
{
    GenieTypeError = 0,
    GenieIntegerType = 1, GenieIntegerArrayType = 2, GenieRealType = 3, GenieRealArrayType = 4, 
    GenieStringType = 5, GenieStringArrayType = 6, GenieWorkspaceType = 7, GenieWorkspaceArrayType = 8,
    GenieUndefinedType = 9, GenieUndefinedArrayType = 10, GenieBaseType = 11, GenieIntervalType = 12
};

// Use GenieVariableType enum as an index into this array
// This string is the name that the type is referred to by in Smalltalk
static const char* GXVariableNames[] = 
{
    "<ERROR in type>", "GXInteger", "GXIntegerarray", "GXReal", "GXRealarray", 
    "GXString", "GXStringarray", "GXWorkspace", "GXWorkspacearray",
    "<undefined>", "<undefined array>", "<genie variable base type>", "Interval"
};

#define GenieFlag(__genie_variable_type)	(2 << __genie_variable_type)

#define GenieIntegerTypeFlag			GenieFlag(GenieIntegerType)
#define GenieIntegerArrayTypeFlag		GenieFlag(GenieIntegerArrayType)
#define GenieRealTypeFlag			GenieFlag(GenieRealType)
#define GenieRealArrayTypeFlag			GenieFlag(GenieRealArrayType)
#define GenieStringTypeFlag			GenieFlag(GenieStringType)
#define GenieStringArrayTypeFlag		GenieFlag(GenieStringArrayType)
#define GenieWorkspaceTypeFlag			GenieFlag(GenieWorkspaceType)
#define GenieWorkspaceArrayTypeFlag		GenieFlag(GenieWorkspaceArrayType)

// Combinations of flags - used to check for a particular group of objects
#define GenieFlatArrayTypeFlag      		(GenieIntegerArrayTypeFlag|GenieRealArrayTypeFlag)
#define GenieArrayTypeFlag          		(GenieStringArrayTypeFlag|GenieWorkspaceArrayTypeFlag)
#define GenieArrayBaseTypeFlag      		(GenieFlatArrayTypeFlag|GenieArrayTypeFlag)

#endif /* G_TYPES_H */
