/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

//
//    Module:         genie_data_access
//    Author:         Chris Moreton-Smith, ISIS
//
//
#include "geniedefs.h"
#include <time.h>
#include "genie_data_access.h"
#include "parsedefs.h"
#include "parse_funcs.h"
#include "smalltalk.h"
#include "genievars.h"

static const char* rcsid = "@(#)$Id: genie_data_access.cxx,v 1.26 1997/07/02 20:15:45 faa Exp $";
USE_RCSID(rcsid);

gdai *This_session;

// Routines to allow simple but powerful data access to Freddie's
// wonder intermediate files and HDF and ISIS raw files and ...
//   who knows what else ?

// All these routines are based on a very simple assumption that it
// can all be done in Open GENIE anyway. Unfortunately Open GENIE looks so
// much like an interative thingy that no one will believe the power of the
// interface it makes possible. So we write a simple but powerful API and
// the point is made !!


// Constructor for an Open GDAI session

// Activate a closet Open GENIE session to do the work
int GX_activate_session( STRING default_format = "GENIE", int inf = 1, int debug = 0)
{
	if (!This_session)
		This_session = new gdai(default_format, inf, debug);  // Constructs a session
	else
	{
		gerr << "GX_activate_session: error, Open GENIE is already active" << endl;
		ginf << "  (in a module, you do not need to activate)" << endl;
	}    
	return 0;
}        
    

// Deactivate the session
int GX_deactivate_session( void ) 
{
    delete This_session;  // Closes the session

    return 0;
}        


// for file association and contents inspection
int GX_select_source( STRING filename )
{
    String Command = "set/file/input ";
    return This_session->gcl(Command + "\"" + filename + "\"");
}


// for file association and contents inspection
int GX_select_destination( STRING filename, STRING file_format = NULL )
{
    if (file_format)
        This_session->file_format(file_format);                       // Make this a checking member

    String Command = "set/file/output ";
    return This_session->gcl(Command + "\"" + filename + "\"");
}


int GX_directory( STRING filename )
{
    String Command = "dir ";
    return This_session->gcl(Command + "\"" + filename + "\"");
}


// Data reading/writing
int GX_get( STRING handle, STRING tag, int object_id )
{
    char locator[SHORTCHARBUFF];
    String Handle = handle;
    ostrstream Locator(locator,SHORTCHARBUFF);

    if (tag && strlen(tag) > 0)
        Locator << "\"" << tag << "\"" << ends;
    else
        Locator << object_id << ends;
    
    return This_session->gcl(Handle + "=get(" + locator + ")");
}


int GX_put( STRING handle, STRING tag, int object_id, STRING flag, STRING comment )
{
    // Can be called for "New" or "Overwrite"
    
    String Command = "put ";
    String Flag = flag;

    if (capitalize(Flag) == "Overwrite")
        *g_err << "***(GX_put) Overwrite flag not yet implemented ***" << endl;
    else if (capitalize(Flag) == "New")
        Command = Command + "/New ";
    else if (capitalize(Flag) == "Append")
        Command = Command + "/Append ";

    Command = Command + This_session->file_format() + " ";

    return This_session->gcl(Command + handle + " \"" + upcase(String(tag)) + "\" " + "\"" + comment + "\"");
}


// Routines for manipulating handles
int GX_assign_handle( STRING handle_lval, STRING handle_rval )
{
    String Handle = handle_lval;
    return This_session->gcl(Handle + "=" + handle_rval);
}

int GX_release_handle( STRING handle )
{
    String Command = "free ";
    return This_session->gcl(Command + "\"" + handle + "\"");
}

// Old name for GX_transfer - kept for backward compatability of VMS SYMBOL_VECTOR (but marked as PRIVATE_PROCEDURE)
extern "C" int GX_associate_handle( STRING handle, STRING direction, STRING type, void* data, int length, int dims[] );

int GX_associate_handle( STRING handle, STRING direction, STRING type, void* data, int length, int dims[] )
{
    return GX_transfer(handle, direction, type, data, length, dims);
}

int GX_transfer( STRING handle, STRING direction, STRING type, void* data, int length, int dims[] )
{
    // This allows data to go both ways, in or out of GENIE

    static OOP gXUUth = registerOOP(internString("gXUUth"));

    BEGINSAVEBLOCK;

    int status = 0;

    String Handle = handle;
    String Direction = direction;
    String Type = type;
    Type.upcase();

    // Get the SystemDictionary
    GenieWorkspace Smalltalk(typeNameToOOP("Smalltalk"));
    
    // Create a dummy handle to use for all the IO and which we know the name of
    // this means that we know what to do with a non identifier hanldle (eg a[4,5]) which
    // We could not parse sufficiently to refer to in smalltalk.
    This_session->gcl("__th=_");

    if ( Direction == "-->" )
    {
        if (Type == "STRING")
        {
            GenieString Data((g_string)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "INT64") 
        {
            GenieInteger Data(*(g_integer*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "INT32" || Type == "INTEGER"))
        {
            GenieInteger Data(*(int*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT64" || Type == "DOUBLE"))
        {
            GenieReal Data(*(g_real*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT32" || Type == "REAL"))
        {
            GenieReal Data(*(float*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "STRING*") 
        {
            gerr << "(GX_transfer)*** Error: \"STRING*\"not yet implemented" << endl;
//            GenieStringArray Data((g_string*)data);
//            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "INT64*") 
        {
            GenieIntegerArray Data((g_integer*)data, length);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "INT32*" || Type == "INTEGER*"))
        {
            GenieIntegerArray Data((int*)data, length);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT64*" || Type == "DOUBLE*"))
        {
            GenieRealArray Data((g_real*)data, length);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT32*" || Type == "REAL*"))
        {
            GenieRealArray Data((float*)data, length);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else
            *g_err << "(GX_transfer)*** Error: \"" << Type << "\" is not a recognized data type" << endl;

        // now assign the value back to the given handle (without a copy)
        status = This_session->gcl(Handle + " <~ __th");
    
    }
    else if ( Direction == "<--" )
    {
        // assign the value out of GENIE (without a copy)
        String Command = "__th";
        status = This_session->gcl(Command + " <~ " + Handle);

        if (Type == "STRING")
        {
            GenieString Data(Smalltalk.at(gXUUth));
            strncpy((char*)data, (c_g_string)Data, length);
            ((char *)data)[length-1] = '\0';
        }
        else if (Type == "INT64") 
        {
            GenieInteger Data(Smalltalk.at(gXUUth));
            *(long*)data = (g_integer)Data;
        }

        else if ( (Type == "INT32" || Type == "INTEGER"))
        {
            GenieInteger Data(Smalltalk.at(gXUUth));
            *(int*)data = (int)Data;
        }
        else if ( (Type == "FLOAT64" || Type == "DOUBLE"))
        {
            GenieReal Data(Smalltalk.at(gXUUth));
            *(g_real*)data = (g_real)Data;
        }
        else if ( (Type == "FLOAT32" || Type == "REAL"))
        {
            GenieReal Data(Smalltalk.at(gXUUth));
            *(float*)data = (float)Data;
        }

        else if (Type == "STRING*") 
        {
            gerr << "(GX_transfer)*** Error: \"STRING*\"not yet implemented" << endl;
//            GenieStringArray Data(Smalltalk.at(gXUUth));
//            strncpy((char*)data, (c_g_string)Data, length);
//            ((char *)data)[length-1] = '\0';
        }
        else if (Type == "INT64*") 
        {
            GenieIntegerArray Data(Smalltalk.at(gXUUth));
            memcpy(data, (g_integer*)Data, length * sizeof(g_integer));
//            ((long *)data)[0] = *((g_integer*)Data);
//            *((long *)data) = *((g_integer*)Data);
        }   
/*        else if ( (Type == "INT32*" || Type == "INTEGER*"))
        {
            GenieIntegerArray Data((g_integer*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT64*" || Type == "DOUBLE*"))
        {
            GenieRealArray Data((g_real*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if ( (Type == "FLOAT32*" || Type == "REAL*"))
        {
            GenieRealArray Data((g_real*)data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else
            *g_err << "(GX_transfer)*** Error: \"" << Type << "\" is not a recognized data type" << endl;

        // now assign the value back to the given handle (without a copy)
        status = This_session->gcl(Handle + " <~ __th");
    
    }
*/  }
    else
    {
        *g_err << "(GX_transfer)*** Error: Direction must be specified as \"IN\" or \"OUT\"" << endl;
        status = 1;
    }

    ENDSAVEBLOCK;

    return status;
}

/*

// Some other possibilities
// friend int GX_compress_file( const char* filename );
// Other realistic possibilities are HDF, or old ISIS raw files

*/
int gdai::gcl(const char* code)
{
    gdbg << code << endl;
    return evalGCLCommand((char *)code);
}

int gdai::smalltalk(const char* code)
{
    evalSmalltalkCommand((char *)code);

    return 0;
}

int gdai::file_format(const char* file_format)
{
    String Fmt = file_format;
    Fmt.upcase();

    if ( Fmt == "GENIE" || Fmt == "" )
        Format = "";
    else if ( Fmt == "HDF" )
        Format = "/HDF";
    else if ( Fmt == "G2" )
        Format = "/G2";
    else
        gerr << "(file_format)*** Error: \"" << file_format << "\" is an unknown file format" << endl;

    return 0;
}


// Destructor for the session
gdai::~gdai()
{

    // This (gracefully) terminates the session
    char *argv[2];
    char *arge[1];
    int   argc  = 1;

    // Select End captive session
    argv[1] = "-E";
    argv[2] = NULL;
    arge[1] = NULL;
    parser_main(argc, argv, arge, in, out, inf, err, dbg);
}


// This constructor is only used inside GENIE and simply sets up a gdai
// object just as if it had been initialized externally.  Once this has
// been done, all the GX_ routines can be used within modules for any
// other form of file access.
gdai::gdai(const char *default_format, istream& instr, genie_ostream& ostr, genie_ostream& oinf, genie_ostream& oerr, genie_ostream& odbg) :
    in(instr.rdbuf()),
    out(ostr),
    inf(oinf),
    err(oerr),
    dbg(odbg)
{
    file_format(default_format);     // use member function to set this
}

	
#undef cout // Got to undefine these to use the real ones !
#undef cin

// This is the constructor called whenever the GX routines are used
// outside Open GENIE.  It also ensures the starting up of a complete
// Open GENIE Session.
gdai::gdai(const char *default_format, int information, int debugging) :
    in(cin.rdbuf()),
    out(cout, "", ""),
    inf(cout, BLUE_ON, BLUE_OFF, information),
    err(cout, RED_ON, RED_OFF),
    dbg(cout, PINK_ON, PINK_OFF, debugging)
{

    char *argv[2];
    char *arge[1];
    int   argc  = 1;

    file_format(default_format);            // use member function to set this

    // Set behaviour of Open GENIE session
    argv[1] = "-B";
    argv[2] = NULL;
    arge[1] = NULL;
    parser_main(argc, argv, arge, in, out, inf, err, dbg);
}
