/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         genie_flat_array
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Optimised "flat" arrays
 *
 */
#include "geniedefs.h"
#include <stdlib.h>
#include "genievars.h"
#include "genie_flat_array.h"

static const char* rcsid = "@(#)$Id: genie_flat_array.cxx,v 1.21 1997/04/13 10:52:42 faa Exp $";
USE_RCSID(rcsid);

#define Array GenieFlatArray<GenieType, CDataType, GXType>

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(OOP oop) :
                GenieArrayBase(nilOOP), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    become(oop);
}

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(const GenieCalloutList& gcl, const char* name) :
                GenieArrayBase(gcl, name), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    if (m_value != nilOOP)
    {
        m_byteArray = byteArrayOOP(m_value);
        m_data = (CDataType*)mstGetCdata(byteArray());
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(const Array& ga, CDataType value) : 
		GenieArrayBase(nilOOP), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    become(allocOOP(ga.dims()));
    fill(value);
}

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(const Array& ga) : 
		GenieArrayBase(nilOOP), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    shallowCopy(ga);
}

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(int dim1, int dim2) : 
	GenieArrayBase(nilOOP), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    become(allocOOP(dim1, dim2));
}

template <class GenieType, class CDataType, GenieVariableType GXType>
GenieFlatArray<GenieType, CDataType, GXType>::GenieFlatArray(const int* dims_array, int ndims) : 
	GenieArrayBase(nilOOP), m_data(0), m_flags(0), m_byteArray(nilOOP)
{
    become(allocOOP(dims_array, ndims));
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::operator=(const Array& ga)
{
    if (this != &ga)
    {
	shallowCopy(ga);
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::operator+=(CDataType ct)
{
    for(int i=0; i<length(); i++)
    {
	(*this)[i] += ct;
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::operator/=(CDataType ct)
{
    for(int i=0; i<length(); i++)
    {
	if (GenieType::isDefined((*this)[i]))
	{
	    if (GenieType::isDefined(ct) && (ct != 0))
	    {
        	(*this)[i] /= ct;
	    }
	    else
	    {
		GenieType::undefine((*this)[i]);
	    }
	}
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::operator()(const CDataType* c_array, int dim1, int dim2)
{
    become(allocOOP(dim1, dim2));
    memcpy(FLAT_ARRAY_DATA, c_array, length()*sizeof(CDataType));
}

template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::become(OOP oop)
{
    GenieArrayBase::become(oop);
    m_flags = 0;
    if (oop != nilOOP)
    {
	m_byteArray = byteArrayOOP(oop);
        m_data = (CDataType*)mstGetCdata(byteArray());
    }
    else
    {
	m_byteArray = nilOOP;
	m_data = 0;
    }
    updateDimsAndDope();
    return oop;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::init(OOP oop, OOP params, int location)
{
    GenieArrayBase::init(oop, params, location);
    m_flags = 0;
    if (oop != nilOOP)
    {
	m_byteArray = byteArrayOOP(oop);
        m_data = (CDataType*)mstGetCdata(byteArray());
    }
    else
    {
	m_byteArray = nilOOP;
	m_data = 0;
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::allocOOP(const int* dims_array, int ndims)
{
    return allocOOP(makeDimsOOP(dims_array, ndims));
}

template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::allocOOP(int dim1, int dim2)
{
    return allocOOP(makeDimsOOP(dim1, dim2));
}

// Return a new genie Array oop of given dimensions
template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::allocOOP(OOP a_dimsOOP)
{
    static OOP gxnew = registerOOP(internString("new:"));
    static OOP gxarray = registerOOP(typeNameToOOP(GXVariableNames[GXType]));
    static const char* gshow_alloc = getenv("GSHOW_ALLOC");
    static unsigned show_alloc_count = 0;
    OOP result;
    if (gshow_alloc != 0)
    {
	show_alloc_count++;
	gout << "GSHOW: GenieFlatArray::allocOOP() called for " << GXVariableNames[GXType] 
	     << ": size = " << arraySize(a_dimsOOP) << ", count = " << show_alloc_count <<"\n";
    }
#if 0
    result = saveOOP(msgSend(gxarray, gxnew, a_dimsOOP, 0));		// Works, but is _far_ too slow
#else
//    result = saveOOP(::allocOOP(instantiate(gxarray)));
    result = saveOOP(instantiateOOPWith(gxarray, 1));		// 1 must agree with realarray.gcl
    instVarAtPut(result, MY_DIMS, a_dimsOOP);
    instVarAtPut(result, MY_DOPE, makeDopeOOP(a_dimsOOP));
    instVarAtPut(result, MY_DATA, makeByteArrayOOP(a_dimsOOP));
#endif
    return result;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::makeByteArrayOOP(OOP a_dimsOOP)
{
    ByteArray byte_array;
    OOP a_byteArrayOOP;
    int len = arraySize(a_dimsOOP) * sizeof(CDataType);
    if( (byte_array = (ByteArray)newInstanceWith(byteArrayClass, len)) != (ByteArray)&nilObject)
    {
        a_byteArrayOOP = saveOOP(::allocOOP((mst_Object)byte_array));
	initEmptyBytes(a_byteArrayOOP, len);
	fillByteArray(a_byteArrayOOP, 0);
    } 
    else
    {
	a_byteArrayOOP = nilOOP;
    }
    return a_byteArrayOOP;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::fillByteArray(OOP a_byteArrayOOP, CDataType value)
{
    CDataType* data = (CDataType*)mstGetCdata(a_byteArrayOOP);
    if (value == 0)
    {
	memset(data, 0, oopNumFields(a_byteArrayOOP));
    }
    else
    { 
        for(int i=0; i<oopNumFields(a_byteArrayOOP)/(int)sizeof(CDataType); i++)
        {
	    data[i] = value;
        }
    }
}

// Make the current array a copy of "ga"
template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::shallowCopy(const Array& ga)
{
    become(shallowCopyOOP(ga.value()));
}

// Return a shallow copy of a genie array OOP
template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::shallowCopyOOP(OOP oop)
{
    OOP copy = allocOOP(dimsOOP(oop));
    memcpy(bytesOOP(copy), bytesOOP(oop), lengthOOP(oop)*sizeof(CDataType));
    return copy;
}
    
// Fill all array elements with value
// mode: 0 =arithmetic, 1 = geometric
template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::fill(CDataType value, CDataType step, int mode)
{
    for(int i=0; i<length(); i++)
    {
	(*this)[i] = value;
	if (mode == 0)
	{
	    value += step;
	}
	else if (mode == 1)
	{
	    value *= step;
	}
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
OOP GenieFlatArray<GenieType, CDataType, GXType>::deepCopyOOP(OOP oop)
{
    return shallowCopyOOP(oop);
}

template <class GenieType, class CDataType, GenieVariableType GXType>
Array GenieFlatArray<GenieType, CDataType, GXType>::operator+(const Array& ga)
{
    Array result(*this);
    for(int i=0; i<length(); i++)
    {
	result[i] += ga[i];
    }
    return result;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
Array GenieFlatArray<GenieType, CDataType, GXType>::operator-(const Array& ga)
{
    Array result(*this);
    for(int i=0; i<length(); i++)
    {
	result[i] -= ga[i];
    }
    return result;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
Array GenieFlatArray<GenieType, CDataType, GXType>::operator/(const Array& ga)
{
    Array result(*this);
    for(int i=0; i<length(); i++)
    {
	if (GenieType::isDefined(result[i]))
	{
	    if ((GenieType::isDefined(ga[i])) && (ga[i] != 0))
	    {
	        result[i] /= ga[i];
	    }
	    else
	    {
	        GenieType::undefine(result[i]);
	    }
	}
    }
    return result;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
Array GenieFlatArray<GenieType, CDataType, GXType>::operator*(const Array& ga)
{
    Array result(*this);
    for(int i=0; i<length(); i++)
    {
	result[i] *= ga[i];
    }
    return result;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::collect(CDataType (*func)(CDataType&))
{
    Array& ga = *this;
    for(int i=0; i<length(); i++)
    {
    	if (GenieType::isDefined(ga[i]))
 	{
	    ga[i] = func(ga[i]);
	}
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::collect(CDataType (*func)(CDataType))
{
    Array& ga = *this;
    for(int i=0; i<length(); i++)
    {
    	if (GenieType::isDefined(ga[i]))
 	{
	    ga[i] = func(ga[i]);
	}
    }
}

// nil_flag: 0=pass nils, 1=skip nils
template <class GenieType, class CDataType, GenieVariableType GXType>
void GenieFlatArray<GenieType, CDataType, GXType>::withCollect(CDataType (*func)(CDataType, CDataType), const Array& ga2, int nil_flag)
{
    Array& ga1 = *this;
    for(int i=0; i<length(); i++)
    {
    	if ((!GenieType::isDefined(ga1[i]) || !GenieType::isDefined(ga2[i])) && nil_flag)
 	{
	    GenieType::undefine(ga1[i]);
	}
	else
	{
	    ga1[i] = func(ga1[i], ga2[i]);
	}
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
CDataType GenieFlatArray<GenieType, CDataType, GXType>::min(bool skip_undefined)
{
    int i;
    CDataType val(0), temp_val;
    GenieType::undefine(val);
// Get a defined starting value
    for(i=0; i<length(); i++)
    {
   	temp_val = (*this)[i];
	if (GenieType::isDefined(temp_val))
	{
	    val = temp_val;
	    break;
	}
    }
    if (!GenieType::isDefined(val))
	return val;

    for(i=0; i<length(); i++)
    {
   	temp_val = (*this)[i];
	if (skip_undefined)
     	{
	    if (GenieType::isDefined(temp_val) && (val > temp_val))
	        val = temp_val;
	}
	else
	{
	    if (val > temp_val)
	        val = temp_val;
	}
    }
    return val;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
CDataType GenieFlatArray<GenieType, CDataType, GXType>::max(bool skip_undefined)
{
    int i;
    CDataType val(0), temp_val;
    GenieType::undefine(val);
// Get a defined starting value
    for(i=0; i<length(); i++)
    {
   	temp_val = (*this)[i];
	if (GenieType::isDefined(temp_val))
	{
	    val = temp_val;
	    break;
	}
    }
    if (!GenieType::isDefined(val))
	return val;

    for(i=0; i<length(); i++)
    {
   	temp_val = (*this)[i];
	if (skip_undefined)
	{
	    if (GenieType::isDefined(temp_val) && (val < temp_val))
	        val = temp_val;
	}
	else
	{
	    if (val < temp_val)
	        val = temp_val;
	}
    }
    return val;
}

template <class GenieType, class CDataType, GenieVariableType GXType>
CDataType GenieFlatArray<GenieType, CDataType, GXType>::sum()
{
    CDataType val(0), temp_val;
    for(int i=0; i<length(); i++)
    {
   	temp_val = (*this)[i];
	if (GenieType::isDefined(temp_val))
	{
	    val += temp_val;
	}
	else
	{
	    return GenieType::undefine(val);
	}
    }
    return val;
}

// This does a byte by byte comparison - if this is too strict, it should be overloaded in the derived class
template <class GenieType, class CDataType, GenieVariableType GXType>
bool GenieFlatArray<GenieType, CDataType, GXType>::operator==(const Array& ga)
{
    if (length() != ga.length())
        return false;

    if (memcmp(this->data(), ga.data(), length()*sizeof(CDataType)) == 0)
    {
	return true;
    }
    else
    {
	return false;
    }
}

template <class GenieType, class CDataType, GenieVariableType GXType>
ostream& GenieFlatArray<GenieType, CDataType, GXType>::print(ostream& os, const char* note) const
{
    GenieArrayBase::print(os, note);
    long old_state = os.setf(ios::showpoint);
    os << GXVariableNames[GXType] << " with " << length() << " elements";
    if (dimsSize() > 1)
        os << " and " << dimsSize() << " dimensions" << endl;
    else
        os << endl;
//    for(int j=1; j<=dimsSize(); j++)
//    {
        os << "\t";
        for(int i=0; i<length(); i++)
        {
            os << (*this)[i] << " ";
            if (i > 6)
                break;
        }
        os << "... (" << length() << " elements)" << endl;
//    }
    os.setf(old_state);
    return os;
}
        
#ifdef __GNUG__
template class GenieFlatArray<GenieReal, g_real, GenieRealArrayType>;
template class GenieFlatArray<GenieInteger, g_integer, GenieIntegerArrayType>;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template GenieFlatArray<GenieReal, g_real, GenieRealArrayType>
#pragma define_template GenieFlatArray<GenieInteger, g_integer, GenieIntegerArrayType>
#endif /* __DECCXX */
