/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         genie_integer_array
 *      Author:         Freddie Akeroyd, ISIS
 *
 */
#include "geniedefs.h"
const int INST_VAR_OFFSET = 2;

#include "genievars.h"
#include "transform_utils.h"
#include "intermediate_file.h"

static const char* rcsid = "@(#)$Id: genie_integer_array.cxx,v 1.18 1997/06/29 15:12:31 faa Exp $";
USE_RCSID(rcsid);

void GenieIntegerArray::init(OOP oop, OOP params, int location)
{
    removeInternalArrays();
    BaseClass::init(oop, params, location);
}
    
// If we re-assign to an array, we need to change any internal rep
void GenieIntegerArray::removeInternalArrays()
{
    if ((m_int != 0) && (m_int != (int*)m_data))
    {
	delete []m_int;
	m_int = 0;
    }
    if ((m_long != 0) && (m_long != (long*)m_data))
    {
	delete []m_long;
	m_long = 0;
    }
    return;
}

GenieIntegerArray::GenieIntegerArray(const int* ip, int dim1, int dim2) :
	GenieFlatArray<GenieInteger, g_integer, GenieIntegerArrayType>(nilOOP), m_int(0), m_long(0)
{
    (*this)(ip, dim1, dim2);
}

GenieIntegerArray::GenieIntegerArray(const long* lp, int dim1, int dim2) :
	GenieFlatArray<GenieInteger, g_integer, GenieIntegerArrayType>(nilOOP), m_int(0), m_long(0)
{
    (*this)(lp, dim1, dim2);
}

void GenieIntegerArray::operator()(const int* ip, int dim1, int dim2)
{
    alloc(dim1, dim2);
    for(int i=0; i<length(); i++)
    {
	(*this)[i] = ip[i];
    }
    removeInternalArrays();	// we remove afer the copy in case we have been re-assign out own array
    return;
}

void GenieIntegerArray::operator()(const long* lp, int dim1, int dim2)
{
    alloc(dim1, dim2);
    for(int i=0; i<length(); i++)
    {
	(*this)[i] = lp[i];
    }
    removeInternalArrays();
    return;
}

int* GenieIntegerArray::setupInt()
{
    if (sizeof(g_integer) == sizeof(int))
    {
	return (int*)FLAT_ARRAY_POINTER;
//	return (int*)m_data;
    }
    g_integer g_temp;
    if (m_int == 0)   
    { 
	m_int = new int[length()];
	for(int i=0; i<length(); i++)
	{
	    g_temp = (*this)[i];
	    if (-g_temp < -INT_MAX) // I mean this, gcc 2.7.2 can't do the +ve version on DU 4.0!!
	    {
		m_int[i] = INT_MAX;
	    }
	    else if (g_temp < -INT_MAX)
	    {
		m_int[i] = -INT_MAX;
	    }
	    else
	    {
		m_int[i] = (int)g_temp;
	    }
	}
    }
    return m_int;
}

long* GenieIntegerArray::setupLong()
{
    if (sizeof(g_integer) == sizeof(long))
    {
	return (long*)FLAT_ARRAY_POINTER;
//	return (long*)m_data;
    }
    if (m_long == 0)   
    { 
	m_long = new long[length()];
	for(int i=0; i<length(); i++)
	{
	    m_long[i] = (long)(*this)[i];
	}
    }
    return m_long;
}

GenieIntegerArray::~GenieIntegerArray()
{
    if (m_flags & WriteBackInternalArrays)
    {
	if (m_int != 0)
	{
	    (*this)(m_int, length());
	}
	if (m_long != 0)
	{
	    (*this)(m_long, length());
	}
    }
    removeInternalArrays();
}

OOP GenieIntegerArray::become(OOP oop)
{
    removeInternalArrays();
    return GenieFlatArray<GenieInteger, g_integer, GenieIntegerArrayType>::become(oop);
}

OOP GenieIntegerArray::become(const GenieVariable& gv)
{
//    return become((GenieIntegerArray)gv);
// HACK - do it properly some time
    return become(gv.value());
}

GenieVariableType GenieIntegerArray::write(IntermediateFile* gif)
{
//    return gif->writeIntegerArray((long*)(*this), dimsArray(), dimsSize()) != 0 ? GenieIntegerArrayType : GenieTypeError;
    return gif->writeIntegerArray(*this) != 0 ? GenieIntegerArrayType : GenieTypeError;
}

GenieVariableType GenieIntegerArray::skip(IntermediateFile* gif)
{
    return gif->skipIntegerArray() != 0 ? GenieIntegerArrayType : GenieTypeError;
}

GenieVariableType GenieIntegerArray::read(IntermediateFile* gif)
{
    return gif->readIntegerArray(*this) != 0 ? GenieIntegerArrayType : GenieTypeError;
}

GenieVariableType GenieIntegerArray::writeAscii(fstream& fs, int indent)
{
    writeAsciiDims(fs, indent);
    fs << setindent(indent);
    for(int i=1; i<=length(); i++)
    {
        fs << (*this)[i-1] << " ";

        if (i%10 == 0)
            fs << "\n" << setindent(indent);
    }
    fs << "\n";
    return GenieIntegerArrayType;
}

GenieVariableType GenieIntegerArray::readAscii(fstream& fs)
{
    OOP a_dimsOOP = readAsciiDims(fs);
    alloc(a_dimsOOP);
    for(int i=0; i<length(); i++)
    {
        fs >> (*this)[i];
    }
    return GenieIntegerArrayType;
}
