/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef GENIE_INTERMEDIATE_FILE
#define GENIE_INTERMEDIATE_FILE

#include "intermediate_file.h"
#include "splash.h"

struct block_item
{
    u_int offset;			// block position in file from xdr_getpos()
    char* label;			// label assigned to the block
    GenieVariableType type;		// The superclass type used to control read/write/skip operations
};					// This may not be the _real_type_ of the variable

// Just here to satisfy SPList template, but it would allow us to sort a
// BlockList if we ever desired to 
inline int operator<(const block_item& a, const block_item& b)
{
    return a.offset < b.offset;
}

typedef SPList<block_item> BlockList;

class GenieIntermediateFile : public IntermediateFile
{
protected:
    FILE* m_file_ptr;			// file pointer for XDR routines
    XDR* m_xdrs_encode;			// XDR for writing
    XDR* m_xdrs_decode;			// XDR for reading
    BlockList block_list;		// list of file contents for fast indexing

public:
    ~GenieIntermediateFile() { if (m_block != -1) { close(); } }
    GenieIntermediateFile() : IntermediateFile(), m_file_ptr(0), m_xdrs_encode(0), m_xdrs_decode(0) {}
    GenieIntermediateFile(const char* file_name, ModeFlags mode = Append) : IntermediateFile(), 
	m_file_ptr(0), m_xdrs_encode(0), m_xdrs_decode(0) { open(file_name, mode); }
    open(const char* file_name, ModeFlags mode = Append);

    int readLabel(char*& label);
    int skipLabel();
    int writeLabel(const char* label);
    int seek(int block = -1); 			// -1 means "end of file", i.e. get ready for an "append"
						// return 1 on success, 0 on error
    int seek(const char* label);		// returns block number, -1 on error	
    int appendBlock(GenieVariableBase& gvb, const char* label = "<no label>", const char* comment = "<no comment>");
    int appendBlock(GenieVariable& gv, const char* label = "<no label>", const char* comment = "<no comment>");
    int writeBlock(GenieVariableBase& gvb, const char* label = "<no label>", const char* comment = "<no comment>");
    int writeBlock(GenieVariable& gv, const char* label = "<no label>", const char* comment = "<no comment>");
    int readBlock(char*& label, GenieVariable& gv);
    int readBlock(GenieVariable& gv);
    u_int getFileOffset();
    u_int setFileOffset(u_int offset);
    int saveOffset(u_int& saved_offset);	// allocate space for an offset field and make a not of its position
    int writeOffset(u_int saved_offset);	// write the current offset to a previosly allocated offset field
    int skipOffset();
    int readOffset(u_int& offset);
    GenieVariableType skipBlock();
    GenieVariableType skipItem();
    GenieVariableType skipItem(char*& comment, char*& user, char*& date);
    int skipRecord();
    int readItem(GenieVariable& gv, char*& comment, char*& user, char*& date);
    int readItem(GenieVariable& gv);
    GenieVariableType readItemType();
    int skipItemType();
    int writeItemType(GenieVariableType gvt);
    int readItemSubclassType(char*& subclass_name);
    int writeItemSubclassType(const char* subclass_name);
    int skipItemSubclassType();
    int writeItem(GenieVariableBase& gvb, const char* comment = "<no comment>");
    int writeItem(GenieVariable& gv, const char* comment = "<no comment>");
    int writeHeader(bool restore_position);
    int readHeader(bool restore_position);
    char* header();
    int close();
    int readInfo(char*& comment, char*& user, char*& date);
    int skipInfo();
    int writeInfo(const char* comment = "<no comment>");
    int skipString();
    int freeString(char*& s);
    int readString(char*& s);
    int writeString(const char* s);
    int skipInteger();
    int readInteger(g3io_int_t& i);
    int writeInteger(g3io_int_t i);
    int skipIntegerArray();
    int readIntegerArray(GenieIntegerArray& ia);
    int writeIntegerArray(GenieIntegerArray& ia);
    int readRealArray(GenieRealArray& gra);
    int writeRealArray(GenieRealArray& gra);
    int skipRealArray();
    int readStringArray(GenieStringArray& gsa);
    int writeStringArray(GenieStringArray& gsa);
    int skipStringArray();
    int readArrayDims(GenieArrayBase& gab);
    int writeArrayDims(const GenieArrayBase& gab);
    int skipReal();
    int readReal(g3io_real_t& r);
    int writeReal(g3io_real_t r);
//    bool validFile();
    int block() { return m_block; }
    int numberOfBlocks() const;
    int endRecord();
    int buildBlockList();
    int freeBlockList();
    int updateBlockList(const char* label, GenieVariableType type = GenieUndefinedType);
    int skipHeader();
    ostream& print(ostream& os) const;
};

#endif /* GENIE_INTERMEDIATE_FILE */

