/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         genie_real_array
 *      Author:         Freddie Akeroyd, ISIS
 *
 */
#include "geniedefs.h"

const int INST_VAR_OFFSET = 2;

#include "genievars.h"
#include "transform_utils.h"
#include "intermediate_file.h"

static const char* rcsid = "@(#)$Id: genie_real_array.cxx,v 1.26 1997/07/08 21:53:21 faa Exp $";
USE_RCSID(rcsid);

GenieRealArray::GenieRealArray(const GenieIntegerArray& gia) :
	GenieFlatArray<GenieReal, g_real, GenieRealArrayType>(nilOOP), m_float(0), m_double(0)
{
    become(allocOOP(gia.dims()));
    for(int i=0; i<gia.length(); i++)
    {
	(*this)[i] = gia[i];
    }
}

// If we re-assign to an array, we need to change any internal rep
void GenieRealArray::removeInternalArrays()
{
    if ((m_float != 0) && (m_float != (float*)m_data))
    {
	delete []m_float;
	m_float = 0;
    }
    if ((m_double != 0) && (m_double != (double*)m_data))
    {
	delete []m_double;
	m_double = 0;
    }
    return;
}

GenieRealArray::GenieRealArray(const float* fp, int dim1, int dim2) :
	GenieFlatArray<GenieReal, g_real, GenieRealArrayType>(nilOOP), m_float(0), m_double(0)
{
    (*this)(fp, dim1, dim2);
}

GenieRealArray::GenieRealArray(const float* fp, const int* dims_array, int ndims) :
	GenieFlatArray<GenieReal, g_real, GenieRealArrayType>(nilOOP), m_float(0), m_double(0)
{
    (*this)(fp, dims_array, ndims);
}

GenieRealArray::GenieRealArray(const double* dp, int dim1, int dim2) :
	GenieFlatArray<GenieReal, g_real, GenieRealArrayType>(nilOOP), m_float(0), m_double(0)
{
    (*this)(dp, dim1, dim2);
}
    
void GenieRealArray::operator()(const float* fp, const int* dims_array, int ndims)
{
    alloc(dims_array, ndims);
    for(int i=0; i<length(); i++)
    {
	if (fp[i] == FLT_MAX)
	    (*this)[i] = UndefinedReal;
 	else
	    (*this)[i] = fp[i];
    }
    removeInternalArrays();
    return;
}

void GenieRealArray::operator()(const float* fp, int dim1, int dim2)
{
    alloc(dim1, dim2);
    for(int i=0; i<length(); i++)
    {
	if (fp[i] == FLT_MAX)
	    (*this)[i] = UndefinedReal;
 	else
	    (*this)[i] = fp[i];
    }
    removeInternalArrays();
    return;
}

void GenieRealArray::operator()(const double* dp, int dim1, int dim2)
{
    alloc(dim1, dim2);
    for(int i=0; i<length(); i++)
    {
	(*this)[i] = dp[i];
    }
    removeInternalArrays();
    return;
}

float* GenieRealArray::setupFloat()
{
    if (!defined())
	return 0;	// return NULL if we don't contain anything
    if (sizeof(g_real) == sizeof(float))
    {
	return (float*)FLAT_ARRAY_POINTER;
//	return (float*)m_data;
    }
    g_real g_temp;
    if (m_float == 0)   
    { 
	m_float = new float[length()];
	for(int i=0; i<length(); i++)
	{
	    g_temp = (*this)[i];
	    if (g_temp > FLT_MAX)
	    {
		m_float[i] = FLT_MAX;
	    }
	    else if (g_temp < -FLT_MAX)
	    {
		m_float[i] = -FLT_MAX;
	    }
	    else
	    {
		m_float[i] = (float)g_temp;
	    }
	}
    }
    return m_float;
}

double* GenieRealArray::setupDouble()
{
    if (!defined())
	return 0;	// return NULL if we don't contain anything
    if (sizeof(g_real) == sizeof(double))
    {
	return (double*)FLAT_ARRAY_POINTER;
//	return (double*)m_data;
    }
    if (m_double == 0)   
    { 
	m_double = new double[length()];
	for(int i=0; i<length(); i++)
	{
	    m_double[i] = (double)(*this)[i];
	}
    }
    return m_double;
}

GenieRealArray::~GenieRealArray()
{
    if (m_flags & WriteBackInternalArrays)
    {
        if (m_float != 0)
        {
	    (*this)(m_float, length());
        }
        if (m_double != 0)
        {
	    (*this)(m_double, length());
        }
    }
    removeInternalArrays();
}

OOP GenieRealArray::become(OOP oop)
{
    removeInternalArrays();
    return GenieFlatArray<GenieReal, g_real, GenieRealArrayType>::become(oop);
}

OOP GenieRealArray::become(const GenieVariable& gv)
{
//    return become((GenieRealArray)gv);
// HACK - do it properly some time
    return become(gv.value());
}

GenieVariableType GenieRealArray::read(IntermediateFile* gif)
{
    return gif->readRealArray(*this) != 0 ? GenieRealArrayType : GenieTypeError;
}

GenieVariableType GenieRealArray::write(IntermediateFile* gif)
{
    return gif->writeRealArray(*this) != 0 ? GenieRealArrayType : GenieTypeError;
}

GenieVariableType GenieRealArray::skip(IntermediateFile* gif)
{
    return gif->skipRealArray() != 0 ? GenieRealArrayType : GenieTypeError;
}

bool GenieRealArray::operator==(const GenieRealArray& gra)
{
    if (length() != gra.length())
        return false;
                                   
    for(int i=0; i<length(); i++)
    {
      if (!GREAL_EQUAL((*this)[i], gra[i]))
            return false;
    }
    return true;
}

GenieVariableType GenieRealArray::writeAscii(fstream& fs, int indent)
{
    writeAsciiDims(fs, indent);
    fs << setindent(indent);
    for(int i=1; i<=length(); i++)
    {
        fs << (*this)[i-1] << " ";

	if (i%10 == 0)
	    fs << "\n" << setindent(indent);
    }
    fs << "\n";
    return GenieRealArrayType;
}

GenieVariableType GenieRealArray::readAscii(fstream& fs)
{
    OOP a_dimsOOP = readAsciiDims(fs);
    alloc(a_dimsOOP);
    for(int i=0; i<length(); i++)
    {
        fs >> (*this)[i];
    }
    return GenieRealArrayType;
}

GenieRealArray GenieRealArray::binsToPoints()
{
    int i, j, *da = dimsArray(), ds = dimsSize();
    for(i=0; i<ds; i++)
        da[i] -= 1;
    GenieRealArray result(da, ds);
    if (ds == 1)
    {
        for(i=0; i<da[0]; i++)
        {
            result[i] = ((*this)[i] + (*this)[i+1]) / 2.0;
        }
    }
    else if (ds == 2)
    {
        for(i=0; i<da[0]; i++)
	{
            for(j=0; j<da[1]; j++)
	    {
                result(i,j) = ((*this)(i,j) + (*this)(i+1,j) + (*this)(i,j+1) + (*this)(i+1,j+1)) / 4.0;
	    }
        }
    }
    return result;
}

// dimension is 1 for first, 2 for second 
// fix dimension at value, return array slice
// value is 1.0 for first array element
GenieRealArray GenieRealArray::slice(int dimension, g_real value)
{
    GenieRealArray temp(1);
    int i, j, *da = dimsArray(), ds = dimsSize();
    int low = (int)value, high = low + 1;
    double alpha;
    if (dimension > ds || dimension <= 0)
    {
	gerr << "CUT: illegal dimension for array = " << dimension << endl;
	return temp;
    }
    if (low < 1)
	low = 1;
    if (high > da[dimension-1])
	high = da[dimension-1];
    if (low > high)
    {
	gerr << "CUT: invalid value (" << value << ") for dimension number " << dimension << endl;
	return temp;
    }
    if (low == high)
	alpha = 0.5;
    else
	alpha = value - (double)low;
// Decrement these to index C arrays (which are zero based)
    --low; --high;
    if (ds == 1)
    {
	temp[0] = (1.0 - alpha) * (*this)[low] + alpha * (*this)[high];
	return temp;
    }
// The returned array is one dimension less than we are
    for(i=1, j=0; i<=ds; i++)
    {
	if (i != dimension)
	    da[j++] = da[i-1];
    }
    ds -= 1;	// new array is one dimension smaller
    GenieRealArray result(da, ds);
    if (ds != 1)
    {
	gerr << "CUT: can only do 1D and 2D at moment" << endl;
	return result;
    }
    if (1 == dimension)
    {
	for(i=0; i<dims(2); i++)
	{
	    if ( (*this)(low, i) == UndefinedReal || (*this)(high, i) == UndefinedReal )
		result[i] = UndefinedReal;
	    else
    	        result[i] = (*this)(low, i) * (1.0 - alpha) + (*this)(high, i) * alpha;
	}
    }
    else if (2 == dimension)
    {
	for(i=0; i<dims(1); i++)
	{
	    if ( (*this)(i, low) == UndefinedReal || (*this)(i, high) == UndefinedReal )
		result[i] = UndefinedReal;
	    else
    	        result[i] = (*this)(i, low) * (1.0 - alpha) + (*this)(i, high) * alpha;
	}
    }
    else
    {
	*g_err << "CUT: Illegal dimension value = " << dimension << endl;
    }
    return result;
}
