/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"

#include <iostream.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include "smalltalk.h"

/*********************************************************************************
 * Basic genie includes
 *********************************************************************************/

#include "parse_funcs.h"
#include "genie_types.h"
#include "genievars.h"

/* External module includes */
#include "g_system.h"
#include "g_objgraphics.h"
#include "g_fortextern.h"
#include "g_transform.h"
#include "g_operations.h"
#include "g_extensions.h"
#include "g_testing.h"
#include "g_peak.h"
#include "g_module.h"
#include "g_newgraphics.h"
#include "g_fileaccess.h"

/*********************************************************************************
 * Facilities currently known about by GENIECALL
 *********************************************************************************/
static const char* rcsid = "@(#)$Id: geniecall.cxx,v 1.29.2.1 1997/08/15 15:42:34 faa Exp $";
USE_RCSID(rcsid);

// These must correspond to the numbers in "setup.gcl"

// Start codes at 0 so tie up with facility_names[]
typedef enum fac_code {
    Debugging   = 0,
    System      = 1,
    Operations  = 2,
    Extensions  = 3,
    Objgraphics = 4,
    Fortextern  = 5,
    Transform   = 6,
    Peak        = 7,
    Module     	= 8,
    Newgraphics = 9,
    FileAccess = 10
} facility_code;

// should tie up with facility_code enum
static const char* facility_names[] =
{
    "Debugging",
    "System",   
    "Operations",
    "Extensions",
    "Objgraphics",
    "Fortextern",
    "Transform",
    "Peak",     
    "Module",    
    "Newgraphics",
    "FileAccess"
};

/*********************************************************************************

geniecall( v facility, v function, v ^ parameters)
=========================================

Acts as a one stop decoder for any external call from the genie smalltalk
interpreter.

Parameters:

	facility    -	An integer code specifying the external facility.

	function    -	A code to select the function within the facility.

	parameters  -	A smalltalk OOP to an array of parameters and names.

**********************************************************************************/

// Indicate failure by nilOOP, success by anything else
extern "C"
OOP geniecall(long fac, long function, OOP oopParameters)
{

    OOP result(nilOOP);
    static int debug_callout = (getenv("GENIE_DEBUG_CALLOUT") != 0 ? 1 : 0);
    facility_code facility = (facility_code)fac;

#if defined(GC_TORTURE)
    if (gcTortureInCallout && (facility == gcTortureFacility))
    {
	startGCTorture();
    }
#endif /* GC_TORTURE */

// Quick longjmp() exception  method 
    GenieExceptionType exception_type;
    if (debug_callout)
	gdbg << "CALLOUT: calling facility \"" << facility_names[facility] << "\"\n";

    if ( (exception_type = (GenieExceptionType)sigsetjmp(genie_exception_context, 1)) == ExceptionSetupOK )
    {
        switch(facility)
        {

            case Debugging:		/*** Debugging of callout ***/
	        if (function == 0)
                {
                    gerr << "Free OOPs = " << numFreeOOPs << endl;
                    result = oopParameters;
                }
    	        else
                {
                    result = g_testing( (testing_function_code) function, oopParameters);
    	        }
	        break;

                case Operations:
        	    result = g_operations( (operation_function_code) function, oopParameters);
    	   	    break;
    
                case Extensions:
    	           result = g_extensions( (extension_function_code) function, oopParameters);
    	           break;
    
                case System:
        	    result = g_system( (system_function_code) function, oopParameters);
    	            break;
    
                case Objgraphics:	/*** Genie Object oriented graphic library interface GOOGLI ***/
    	            result = g_objgraphics( (graphics_function_code) function, oopParameters );
    	            break;
    
                case Transform:	/*** Module for linking out to external workspace transformations ***/
        	    result = g_transform( (transform_function_code) function, oopParameters );
    	            break;
    
                case Fortextern:	/*** Module for linking out to external FORTRAN routines ***/
                    result = g_fortextern( (fortran_function_code) function, oopParameters );
    	            break;
    
                case Peak:
                    result = g_peak( (peak_function_code) function, oopParameters);
    	            break;
    
                case Module:
    	            result = g_module( (module_function_code) function, oopParameters);
    	            break;
    
                case Newgraphics:
    	            result = g_newgraphics( (NewGraphicsCode) function, oopParameters);
    	            break;
    
		case FileAccess:
		    result = g_fileAccess( (FileAccessCode)function, oopParameters);
		    break;

                default:
        	    gerr << "geniecall: Unknown facility number = " << facility << endl;
        	    result = nilOOP;
    	            break;
          }
    }
    else // error return 
    {
	gerr << "*** Caught GenieException (code = " << (int)exception_type << ") ***" << endl;
	ginf << GenieException::message(exception_type) << endl;
	result = nilOOP;
    }
	
#if defined(GC_TORTURE)
    if (gcTortureInCallout && (facility == gcTortureFacility))
    {
	stopGCTorture();
    }
#endif /* GC_TORTURE */

    return result;
}



/*
 * Simple heuristic check to ensure parameters are being passed from GENIE correctly
 */

static int validParameters( OOP parameters ) 
{
    int is_oop(0), is_array(0), field_count_ok(0);
    is_oop = isOOP(parameters);               /*** First check proper OOP, ***/
    if (is_oop)
    {
	is_array = isAKindOf(oopClass(parameters),arrayClass);      /*** then for an array, ***/
        field_count_ok = numIndexableFields(parameters) == 2;       /*** and finally for correct number of elements ***/
    }
    return is_oop && is_array && field_count_ok;
}



/*
 * Gets a 'C' string from a symbol (no dynamic allocation done)
 */

char *symToCString( OOP oopSym )
{

	static int len;
	static char buff[SHORTCHARBUFF];

	len = stringOOPLen(oopSym);
	if (len >= sizeof(buff))
	{
	    gerr << "Buffer too small in symToCString() for " << len << endl;
	    len = 0;
	}
	strncpy(buff, (char *) stringOOPChars(oopSym), len);
	buff[len] = '\0';

	return buff;
}



/*
* this routine is called from a geniecall callout function 
* and is used to link a GenieVariable class object to the
* underlying smalltalk OOP
* 
* returns 0 on success, 1 on failure (invalid parameters)
*
* The variable argument list of the form <char*> <GenieVariable*> and is terminated by NULL
*/

/* #define DISPLAY_PROCESSING  /*   */

// Freeup storage in "names" using parsenamesFree()
static int parseNames(OOP oopParameters, char**& names, int& param_count)
{
    OOP symbol;
    if (!validParameters(oopParameters))
    {
	gerr << "parseNames(invalid parameters)" << endl;
	return 1;
    }
    OOP paramNames = arrayAt(oopParameters, 1);
    param_count = numIndexableFields(paramNames);
    names = new char*[param_count+1];
    char* item_name;
// Parse parameter names into an array
    names[0] = names[1] = 0;	// just for safety
        for (int i=2; i <= param_count; i++ ) 
	{
            symbol = arrayAt(paramNames,i);             /*** symbol containing field name ***/
	    item_name = symToCString(symbol);
	    names[i] = new char[strlen(item_name) + 1];
	    strcpy(names[i], item_name);
            UnGarble(names[i]);
#ifdef DISPLAY_PROCESSING
	    gerr << "parseNames(GCL parameter " << i << " = " << names[i] << ")" << endl;
#endif	/* DISPLAY_PROCESSING */
        }    
    return 0;
}


static void parseNamesFree(char**& names, int& param_count)
{
    for(int i=2; i<=param_count; i++)
    {
	delete[] names[i];
	names[i] = 0;
    }
    delete[] names;
    names = 0;
    return;
}


#if 0
int parametersToWorkspace(OOP oopParameters, GenieWorkspace& workspace)
{
    int param_count;
    char** names;
    if (parseNames(oopParameters, names, param_count) != 0)
    {
	return 1;
    }
    workspace.alloc(param_count);
    OOP paramValues = arrayAt(oopParameters, 2);
    for(int i=2; i<=param_count; i++)
    {
        item_value = instVarAt(paramValues, i);           /*** field value being passed ***/
        gvp->init(item_value, paramValues, location);
    parseNamesFree(names, paramcount);
}
#endif


// Return 0 on success, 1 on error
int associateGenieVariables(OOP oopParameters, ... )
{
    va_list ap;
    int location;
    char** names;
    GenieVariableBase* gvp;
    OOP item_value;
    int param_count;
// Parse parameter names into an array
    if (parseNames(oopParameters, names, param_count) != 0)
    {
	return 1;
    }
    OOP paramValues = arrayAt(oopParameters, 2);
    int i;
    char* item_name;
// Now select out the ones required
    va_start(ap, oopParameters);
    while((item_name = va_arg(ap, char*)) != NULL)
    {
#ifdef DISPLAY_PROCESSING
	gerr << "associateGenieVariables(Looking up " << item_name << ")" << endl;
#endif	/* DISPLAY_PROCESSING */
        gvp = va_arg(ap, GenieVariableBase*);    
// now find which parameter is appropriate
        for(i=2; i<=param_count; i++)
	{ 
	    if (Strcasecmp(names[i], item_name) == 0)
	    {
		break;
	    }
	}
        if (i <= param_count)
	{
            item_value = instVarAt(paramValues, i);           /*** field value being passed ***/
	    location = i;
#ifdef DISPLAY_PROCESSING
    	    gerr << "associateGenieVariables(name " << item_name << " found at position " << location << ")" << endl;
#endif
	}
	else
	{
	    gerr << "associateGenieVariables(name " << item_name << " not found)" << endl;
	    item_value = nilOOP;
	    location = 0;	// variable will not be written back by the GenieVariable class destructor
	}
        gvp->init(item_value, paramValues, location);
    }
    va_end(ap);
    parseNamesFree(names, param_count);
    return 0;
}
