/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 * File: 	getsect.cc - A fast interface to the old GenieII get routines.
 * Author:	Freddie Akeroyd, ISIS computer group
 *
 * Bassically, we just read the whole file into "file_buffer" and then catch
 * calls to the old fortran GETSECT routine
 */
#include "geniedefs.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#ifdef __VMS_POSIX
#	include <vms_posix_extensions.h>	/* for _translate() */
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <gnustring.h>

#ifdef _WIN32
#	include <io.h>
#	include <fcntl.h>
	char _itoa_buff[18];	/* _itoa can return 17 characters */
#	define dec(x) itoa(x, _itoa_buff, 10)	/* Seems to do the same thing !! */
#else
#	include <sys/file.h>
#	include <stream.h>
#endif /* _WIN32 */

#include "Genie_functions.h"

static const char* rcsid = "@(#)$Id: getsect.cxx,v 1.26 1997/04/28 07:54:43 faa Exp $";
USE_RCSID(rcsid);

#if !defined(USE_OPEN) && !defined(USE_FOPEN)
#define USE_FOPEN
#endif

// Sort out external symbol names
#if defined(HAVE_LIBF2C)
#define fastget_init fastget_init__
#define getsect getsect_
#define fort_file fort_file__
#elif defined(__unix)
#define fastget_init fastget_init_
#define getsect getsect_
#define fort_file fort_file_
#endif

extern "C" 
{
    int fastget_init(const char* name, const int& iunit);
    void getsect(const int& istart, const int& ilong, int ivalue[], const int& iunit, int& ierr);
    void fort_file(const int& iunit, char* file_name);
}

static int enlarge_buffer(off_t new_buffer_size);

enum FastgetBufferOption { ReadWholeFile, UseFOPEN };

static char file_name[61];					// Name of the file to buffer (passed from fortran)
static char* file_buffer = NULL;				// Where we store "file_name"
static off_t file_size = 0;					// How big "file_name" is
static off_t file_buffer_size = 0;				// How big our buffer is ( file_size <= file_buffer_size )
static long io_block_size = 0;					// Optimal read block size for the file system containing file_name
static long buffer_iunit = 0;					// Fortran unit associated with "file_buffer"
static FastgetBufferOption buffer_option = ReadWholeFile;	// how "file buffer" is being used
// G_MAX_BUFFER_SIZE is the maximum size the buffer may be
// G_FOPEN_BUFFER_SIZE is the buffer size used if file is larger than G_MAX_BUFFER_SIZE
static off_t MAX_BUFFER_SIZE = (getenv("G_MAX_BUFFER_SIZE") != 0 ? atol(getenv("G_MAX_BUFFER_SIZE")) : 4000000);
static off_t FOPEN_BUFFER_SIZE = (getenv("G_FOPEN_BUFFER_SIZE") != 0 ? atol(getenv("G_FOPEN_BUFFER_SIZE")) : 200000);
static FILE* file_fd = NULL;

// Returns 1 on error, 0 on success
// name is character*60 fortran variable
int fastget_init(const char* name, const int& iunit)
{
    String error_message;
    struct stat stat_buffer;
//    stat_t stat_buffer;
    int i;
// First null terminate our FORTRAN string
    for(i=59; name[i] == ' '; i--)
	;
    if (i < 0)
    {
	error_add("FASTGET_INIT", "null file name given", "");
	return 1;
    }
    strncpy(file_name, name, i+1);
    file_name[i+1] = '\0';
#ifdef __VMS_POSIX
    const char* trans_file_name = _translate(file_name, _TO_PATHNAME);
    if (trans_file_name != NULL)
    {
	strcpy(file_name, trans_file_name);
    }
#endif /* __VMS_POSIX */
    buffer_iunit = iunit;
// Now attempt to open the file and get some information about it
#if defined(__VMS) && !defined(__VMS_POSIX)
    int fd = open(file_name, O_RDONLY, 0, "shr=put");
#else
    int fd = open(file_name, O_RDONLY, 0);
#endif /* __VMS && !__VMS_POSIX */
    if (fd < 0) 
    {
	error_message = String("Cannot open \"") + file_name + String("\" --- ") + String(strerror(errno));
	error_add("FASTGET_INIT", error_message, "Check file exists and is readable");
	return 1;
    }
/***** ???   if (stat(file_name, &stat_buffer) != 0)   seemed to fail sometimes ??? *****/
    if (fstat(fd, &stat_buffer) != 0)
    {
	error_message = String("Cannot fstat \"") + file_name + String("\" --- ") + String(strerror(errno)); 
	error_add("FASTGET_INIT", error_message, "Check file is not corrupted");
        close(fd);
	return 1;
    }
    if (close(fd) != 0)
    {
	error_message = String("Cannot close \"") + file_name + String("\" --- ") + String(strerror(errno)); 
	error_add("FASTGET_INIT", error_message, "This shouldn't happen!");
	return 1;
    }
    file_size = stat_buffer.st_size;
#if defined(__VMS) || defined(_WIN32)
    io_block_size = 512;			// optimal IO block size for this file system
#else
    io_block_size = stat_buffer.st_blksize;	// optimal IO block size for this file system
#endif /* __VMS */
    if (file_size > MAX_BUFFER_SIZE)
    {
	buffer_option = UseFOPEN;
	enlarge_buffer(FOPEN_BUFFER_SIZE);
    }
    else
    {
	buffer_option = ReadWholeFile;
	if (file_size > file_buffer_size)
	{
	    enlarge_buffer(file_size);
	}
    }

#ifdef USE_FOPEN
    if (file_fd != 0)
    {
	fclose(file_fd);
	file_fd = 0;
    }
// Now fill up our buffer
#if defined(__VMS) && !defined(__VMS_POSIX)
    if ((file_fd = fopen(file_name, "r", "shr=put", "ctx=stm")) == NULL)
#else
    if ((file_fd = fopen(file_name, "r")) == NULL)
#endif /* __VMS && !__VMS_POSIX */
    {
	error_message = String("Cannot fopen \"") + file_name + String("\" --- ") + String(strerror(errno));
	error_add("FASTGET_INIT", error_message, "Check file exists and is readable");
	return 1;
    }
    if ((buffer_option == ReadWholeFile) && (fread(file_buffer, 1, file_size, file_fd) != (size_t)file_size))
    {
	error_message = String("Cannot read ") + String(dec(file_size)) +
			String(" bytes from \"" ) + file_name + String("\" --- ") + String(strerror(errno));
	error_add("FASTGET_INIT", error_message, "Check file is not corrupted or has been deleted");
	return 1;
    }          
    else if (buffer_option == UseFOPEN)
    {
	if (setvbuf(file_fd, file_buffer, _IOFBF, FOPEN_BUFFER_SIZE) != 0)
	{
	    perror("setvbuf");
	    return 1;
	}
    }
#endif /* USE_FOPEN */
#ifdef USE_STREAMS
    ifstream the_file;
    the_file.open(file_name);
    the_file.read(file_buffer, file_size);
    the_file.close();
#endif /* USE_STREAMS */
    return 0;
}

/*
 * Originally declared as SUBROUTINE GETSECT(ISTART,ILONG,IVALUE,IUNIT,IERR) in FORTRAN
 * 
 * The routine read a certain number of 512 byte blocks from a direct access file on unit "iunit"
 */
void getsect(const int& istart, const int& ilong, int ivalue[], const int& iunit, int& ierr)
{
    static char buffer[256];
    static const int sz = sizeof(int);
#if 0
    char fort_file_name[60];		// fortran file attached to unit "iunit"
    if (iunit != buffer_iunit)		// If we've changed file, reinitialize
    {
	fort_file(iunit, fort_file_name);
	fastget_init(fort_file_name, iunit);
    }
#endif
    if ((ilong < 0) || ((istart + ilong - 1) * sz > file_size))
    {
	sprintf(buffer, "Attempt to read invalid part of file buffer: istart = %d words, ilong = %d words, file size = %ld bytes",
			istart, ilong, file_size);
	error_add("GETSECT", buffer, "This shouldn't happen!");
	ierr = -1;
	return;
    }
    ierr = 0;
    if (buffer_option == ReadWholeFile)
    {
    	memcpy(ivalue, file_buffer+(istart-1)*sz, ilong*sz);
    }
    else
    {
	if (fseek(file_fd, (istart-1)*sz, SEEK_SET) == -1)
	{
	    sprintf(buffer, "Error seeking file size %ld bytes at %d: %s", 
			file_size, (istart-1)*sz, strerror(errno));
	    error_add("GETSECT", buffer, "This shouldn't happen!");
    	    ierr = -1;
	    return;
	}
        if (fread(ivalue, sz, ilong, file_fd) != (size_t)ilong)
        {
	    sprintf(buffer, "Error reading file size %ld bytes for %d words from %d: %s", 
			file_size, ilong, istart-1, strerror(errno));
	    error_add("GETSECT", buffer, "This shouldn't happen!");
	    ierr = -1;
	    return;
	}
    }
    return; 
}

static int enlarge_buffer(off_t new_buffer_size)
{
    String error_message;
    if (new_buffer_size > file_buffer_size)
    {
	delete[] file_buffer;
	file_buffer = new char[new_buffer_size];
	if (file_buffer == NULL)
	{
	    error_message = String("Unable to allocate ") + String(dec(file_size)) + String(" bytes for file buffer");
	    error_add("FASTGET_INIT", error_message, "Check your page file quota and/or the current machine load");
	    file_buffer_size = 0;
	    return 1;
	}
	file_buffer_size = new_buffer_size;
    }        
    return 0;
}
