/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef __IOFUNCS__
#define __IOFUNCS__

/*
//
// The Streams and IO routines defined in this file are used throughout Open
// genie to control normal IO, error reporting & debug messages.
//


// GENIE STREAMS (C++)

// All IO must go through one of output, information, error, or debug streams
// Note: that output, informational, warning & debug are all on stdout, only
// error actually goes to stderr
//
// All output the user might see
//  gout
//  ginf
//  gerr
//  gdbg
//
// Automatically compressed input/output where necessary
//  gzin
//  gzout
//
// IO routines for C
//  gprintf
//  gprintif
//  gprintef
//  gprintdf


//
// Handy error reporting macro for system call failures
// 
//   Always use this instead of perror as it helps track down the line
//   and file in which the error occured (useful when debugging
//   over a phone line!)
//
//
*/

#ifdef perror
#   undef perror
#endif

#ifdef __cplusplus
#   define perror(routine) gerr << (routine) << strerror(errno) << endl
#   define PERROR(routine) gerr << (routine) << strerror(errno) << endl
#else
#   define perror(routine) gprintef("%s%s\n",(routine), strerror(errno))
#   define PERROR(routine) gprintef("%s%s\n",(routine), strerror(errno))
#endif /* __cplusplus */


/* I/O descriptor numbers (capitalized to distinguish from C FILE streams stdin, stdout etc.) */
enum { Stdin=0, Stdout=1, Stderr=2 };

#ifdef __cplusplus


// define some useful ANSI sequences (for setting up genie streams).
static const char BLUE_ON[]  = "\033[34m\033[1m";
static const char BLUE_OFF[] = "\033[0m";
static const char RED_ON[]   = "\033[31m\033[1m";
static const char RED_OFF[]  = "\033[0m";
static const char PINK_ON[]  = "\033[35m\033[1m";
static const char PINK_OFF[] = "\033[0m";


// C++ Streams to take all output and filter if necessary.
// these are also used by the GENIE C I/O routines to avoid
// synchronization problems.

// Note: gerr and gprintef are #defined differently so that the current line number
// and file from which they are called get reported automatically at the start of
// the error message.

extern istream              *g_in;       // No special processing needed on input
extern genie_ostream        *g_out;      // equivalent to cout
extern genie_ostream        *g_inf;      // filter for informational messages/warnings
extern genie_ostream        *g_err;      // filter for error messages (to stderror)
extern genie_ostream        *g_dbg;      // filter for debug output (to stdlog)

#define gin  (*g_in)
#define gout (*g_out)
#define ginf (*g_inf)
#define gerr (*g_err << __LINE__ << "/\"" << GENIE_FILE << "\": " )
#define gdbg (*g_dbg)


// MACROS to trap use of all non-genie I/O streams
#define cin     (gdbg << __LINE__ << "/\"" << GENIE_FILE << "\": Please use \"gin\" here" << endl, cin)
#define cout    (gdbg << __LINE__ << "/\"" << GENIE_FILE << "\": Please use \"gout\" here" << endl, cout)
#define cerr    (gdbg << __LINE__ << "/\"" << GENIE_FILE << "\": Please use \"gerr\" here" << endl, cerr)
#define clog    (gdbg << __LINE__ << "/\"" << GENIE_FILE << "\": Please use \"glog\" here" << endl, clog)
 
// Printing routines called from C (mainly by the parser)

extern "C" {

#endif /* __cplusplus */

extern FILE *__si;    /* temps for stdin, stdout & stderr */
extern FILE *__so;
extern FILE *__se;

/* The consts here are just to let me get round the preprocessor where #define stdin X stdin X fails.*/
#define printf  (gprintdf("%d/%s%s\n", __LINE__, GENIE_FILE,": Please use \"gprintf()\" here" ), &printf)

/* Protect this bit when initializing __si, etc. in iofuncs.cc !! */
#ifndef IN_IOFUNCS_CC
#   ifdef stdin
#       undef stdin
#       undef stdout
#       undef stderr
#   endif
#   define stdin   (gprintdf("%d/%s%s\n", __LINE__, GENIE_FILE,": Please use \"gprintXf()\" here" ), __si)
#   define stdout  (gprintdf("%d/%s%s\n", __LINE__, GENIE_FILE,": Please use \"gprintXf()\" here" ), __so)
#   define stderr  (gprintdf("%d/%s%s\n", __LINE__, GENIE_FILE,": Please use \"gprintXf()\" here" ), __se)
#endif /* IN_IOFUNCS_CC */

/* logs errors */
void logit( const char callinfo[] );
void gwrite(const void *data, size_t size, size_t num_items);
void gwritei(const void *data, size_t size, size_t num_items);
void g_writee(const void *data, size_t size, size_t num_items);
void gwrited(const void *data, size_t size, size_t num_items);
void gprintf(const char *fmt, ...);
void gprintif(const char *fmt, ...);
void g_printef(const char *fmt, ...);
void gprintdf(const char *fmt, ...);

#define gprintef (g_printef("%d/\"%s\"/", __LINE__, GENIE_FILE), &g_printef)
#define gwritee (g_printef("%d/\"%s\"/", __LINE__, GENIE_FILE), &g_writee)

#ifdef __cplusplus
}


// "cout << setindent(n)" will indent output by n spaces for that line

ostream& g_indent(ostream& os, int n);

#if defined(__GNUG__)
    inline omanip<int> setindent(int n)
    {
        return omanip<int>(g_indent, n);
    }
#elif defined(OMANIP)
    inline OMANIP(int) setindent(int n)
    {
        return OMANIP(int)(g_indent, n);
    }
#else
#error Cannot create omanips
#endif /* __GNUG__ */

#endif /* __cplusplus */

#endif /* __IOFUCNCS__ */
