/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <iostream.h>
#ifdef _WIN32
#	include <io.h>
	char _itoa_buff[18];	/* _itoa can return 17 characters */
#	define dec(x) itoa(x, _itoa_buff, 10)	/* Seems to do the same thing !! */
#else
#	include <stream.h>
#endif /* _WIN32 */
#include <math.h>
#include "geniedefs.h"
#include "genievars.h"
#include "genie_types.h"
#include "g_fortextern.h"
#include "isis_raw_file.h"
#include "Genie_functions.h"
#include "units_p.h"

static const char* rcsid = "$Id: isis_raw_file.cxx,v 1.11 1997/04/18 09:08:38 cmm Exp $";
USE_RCSID(rcsid);

ISISRawFile::ISISRawFile() : m_spec_work(0), m_idat(0), m_delt_work(0), m_tthe_work(0), m_l2_work(0), 
	m_tcb(0), m_user_tables(0), m_xdata(0)
{
}

// 0 on success, 1 on failure
int ISISRawFile::open(const char* runid)
{
    m_runid = runid;
    translatePath(m_runid, 'f');
    if (checkAndOpen(m_runid, m_ntc, m_ndet, m_nuse) != 0)
    {
	return 1;
    }
    m_ndetmax = m_ndet + 1;
    m_ntcmax = m_ntc + 1;
    m_spec_work = new fort_int[m_ndetmax];
    m_idat = new fort_int[m_ntcmax];              // for counts
    m_delt_work = new fort_real[m_ndetmax];
    m_tthe_work = new fort_real[m_ndetmax];
    m_l2_work = new fort_real[m_ndetmax];
    m_tcb = new fort_real[m_ntcmax];
    m_user_tables = new fort_real[m_nuse];
    m_xdata = new g_real[m_ntcmax];
    return (readSpectrum(0, 1) == -1);
}

// Return 0 on success, -1 on read error
int ISISRawFile::readSpectrum(GenieWorkspace* pspectrum, fort_int spec_no)
{
    fort_int errcode, quick(0);
    fort_real l1, l2, tthe, delt, duration, phi;
    double denom;
    g_real yval, eval;
    String run_title, comb_time, user_name, inst_name, run_no;
    READ_DATA(m_runid,errcode,spec_no,m_delt_work,m_spec_work,
                      m_tthe_work,m_l2_work,m_ndetmax,m_tcb,m_idat,m_ntcmax,
                      l1,l2,tthe,delt,phi,run_title,duration,
                      comb_time,user_name,inst_name,run_no,
                      m_user_tables, m_nuse, quick);
    if (errcode != 0)
    {
	gerr << "Error reading \"" << m_runid << "\"" << endl;
	return -1;
    }
    for(int i=0; i<m_ntc+1; i++)
	m_xdata[i] = m_tcb[i];

    if (pspectrum != 0)
    {
            GenieRealArray ydata(m_ntc);
            GenieRealArray edata(m_ntc);
            int ii;
            for(int i=0; i<m_ntc; i++)
            {
/* first Y point is dummy in get routines */
                ii = i + 1;
                denom = m_tcb[i+1] - m_tcb[i];
                yval = (double)m_idat[ii] / denom;
                eval = sqrt((double)m_idat[ii]) / denom;
                ydata[i] = yval;
                edata[i] = eval;
            }

	    GenieWorkspace& spectrum = *pspectrum;
            spectrum("NTC", m_ntc);
            spectrum("FI", phi);
            spectrum("NUT", m_nuse);
            spectrum("L1", l1);
            spectrum("L2", l2);
            spectrum("YMAX", 0.0);
/*
 *  assign units on x and y axes --- these will be TOF (microseconds)
 *  and neutron counts per microsecond respectively
 */
            spectrum("XLABEL", unit_name(2, 'x'));
            spectrum("YLABEL", unit_name(2, 'y'));

            spectrum("X", GenieRealArray(m_tcb, m_ntc+1));
            spectrum("Y", ydata);
            spectrum("E", edata);
            spectrum("UT", GenieRealArray(m_user_tables, m_nuse));

            spectrum("TITLE", run_title);
            spectrum("TIME", comb_time);
            spectrum("HISTORY", " ");
            spectrum("USER_NAME", user_name);
            spectrum("INST_NAME", inst_name);
            spectrum("RUN_NO", run_no);
            spectrum("FILE", m_runid);
            spectrum("DELTA", delt);
            spectrum("RUN_DURATION", duration);
            spectrum("SPEC_NO", dec(spec_no));
            spectrum("EFIXED", 0.0);
            spectrum("TWOTHETA", tthe);
            spectrum("EMODE", 0);
            spectrum("BIN_GROUPING", 0);
    }
    return 0;
}

ISISRawFile::~ISISRawFile()
{
    delete []m_tcb;
    delete []m_delt_work;
    delete []m_tthe_work;
    delete []m_l2_work;
    delete []m_spec_work;
    delete []m_idat;
    delete []m_xdata;
    delete []m_user_tables;
}

int ISISRawFile::checkAndOpen(String& runid, fort_int& ntc, fort_int& ndet, fort_int& nuse)
{
    fort_int errcode;
    OPEN_DATA_FILE(runid, ntc, ndet, nuse, errcode);
    if (errcode != 0)
    {
        gerr << "Cannon open file \"" << runid << "\"" << endl;
        return 1;
    }
    else
    {
        return 0;
    }
}

// Return 0 on success, -1 on error
int ISISRawFile::readSpectra(GenieWorkspace& spectrum, const GenieIntegerArray& spectrum_numbers, int nspec)
{
    if (!spectrum_numbers.defined()  || nspec == 0)
	return -1;
    fort_int spec_no = spectrum_numbers[0];
    fort_int errcode, quick(0);
    fort_real l1, l2, tthe, delt, duration, phi;
	    double denom;
	    g_real yval, eval;
	    String run_title, comb_time, user_name, inst_name, run_no;
// Read base stuff
            READ_DATA(m_runid,errcode,spec_no,m_delt_work,m_spec_work,
                      m_tthe_work,m_l2_work,m_ndetmax,m_tcb,m_idat,m_ntcmax,
                      l1,l2,tthe,delt,phi,run_title,duration,
                      comb_time,user_name,inst_name,run_no,
                      m_user_tables, m_nuse, quick);
	    if (errcode != 0)
	    {
	        return -1;
	    }
            spectrum("NTC", m_ntc);
            spectrum("NUT", m_nuse);
            spectrum("L1", l1);
            spectrum("YMAX", 0.0);

            GenieRealArray ydata(nspec,m_ntc);
            GenieRealArray edata(nspec,m_ntc);
	    GenieRealArray l2_array(nspec);
	    GenieRealArray two_theta_array(nspec);
	    GenieRealArray delt_array(nspec);
	    GenieRealArray phi_array(nspec);
	    GenieRealArray user_tables_array(nspec,m_nuse);

    quick = 1;
    int i,j,k,ii;
    for(j=0; (j < nspec) && (errcode == 0); j++)
    {
	spec_no = spectrum_numbers[j];
        READ_DATA(m_runid,errcode,spec_no,m_delt_work,m_spec_work,
                      m_tthe_work,m_l2_work,m_ndetmax,m_tcb,m_idat,m_ntcmax,
                      l1,l2,tthe,delt,phi,run_title,duration,
                      comb_time,user_name,inst_name,run_no,
                      m_user_tables, m_nuse, quick);
    	if (errcode == 0)
	{
	    l2_array[j] = l2;
	    two_theta_array[j] = tthe;
	    delt_array[j] = delt;
	    phi_array[j] = phi;
	    for(k=0; k<m_nuse; k++)
  	    {
	        user_tables_array(j, k) = m_user_tables[k];
	    }
 
            for(i=0; i<m_ntc; i++)
            {
/* first Y point is dummy in get routines */
                ii = i + 1;
                denom = m_tcb[i+1] - m_tcb[i];
                yval = (double)m_idat[ii] / denom;
                eval = sqrt((double)m_idat[ii]) / denom;
                ydata(j,i) = yval;
                edata(j,i) = eval;
            }
	}
    }
/*
 *  assign units on x and y axes --- these will be TOF (microseconds)
 *  and neutron counts per microsecond respectively
 */
            spectrum("XLABEL", unit_name(2, 'x'));
            spectrum("YLABEL", unit_name(2, 'y'));

            spectrum("X", GenieRealArray(m_tcb, m_ntc+1));
            spectrum("Y", ydata);
            spectrum("E", edata);
            spectrum("UT", user_tables_array);

            spectrum("TITLE", run_title);
            spectrum("TIME", comb_time);
            spectrum("HISTORY", " ");
            spectrum("USER_NAME", user_name);
            spectrum("INST_NAME", inst_name);
            spectrum("RUN_NO", run_no);
            spectrum("FILE", m_runid);
            spectrum("DELTA", delt_array);
            spectrum("RUN_DURATION", duration);
            spectrum("SPEC_NO", dec(spec_no));
            spectrum("EFIXED", 0.0);
            spectrum("TWOTHETA", two_theta_array);
            spectrum("EMODE", 0);
            spectrum("BIN_GROUPING", 0);
            spectrum("FI", phi_array);
            spectrum("L2", l2_array);
    return 0;
}


// Return 0 on success, -1 on error
int ISISRawFile::readSpectrum(g_real* yarray, g_real* earray, fort_int spec_no)
{
    fort_int errcode(0), nos(1);
    STRING runid_i;
    runid_i.length = m_runid.length();
    strcpy(runid_i.spot, m_runid);
    GETDAT(runid_i, spec_no, nos, m_idat, m_ntc, errcode);

    if (errcode != 0)
	return -1;

    double denom;
    int ii;
    for(int i=0; i<m_ntc; i++)
    {
/* first Y point is dummy in get routines */
        ii = i + 1;
        denom = m_tcb[i+1] - m_tcb[i];
        yarray[i] = (double)m_idat[ii] / denom;
        earray[i] = sqrt((double)m_idat[ii]) / denom;
    }
    return 0;
}
