/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* 
 *	Module:		local_units
 * 	Author:		Freddie Akeroyd, ISIS, 1/3/96
 * 	Purpose:	Allow user to define and manipulate new unit types
 */

#include "geniedefs.h"
#include <iostream.h>
#include <gnustring.h>
#include <math.h>
#include "assoc.h"
#include "local_units.h"
//#include "Genie_functions.h"
#include "genievars.h"

static const char* rcsid = "@(#)$Id: local_units.cxx,v 1.5 1997/03/26 16:26:33 cmm Exp $"; 
USE_RCSID(rcsid);

static Assoc<DerivedUnit> units_list;		// Associative array (indexed by unit name) to hold units information

/*
 * Create or replace an instance of a unit, returning an "encoded unit" string that will
 * be used later to index the unit or recreate it if needed
 */
GenieWorkspace create_unit(const char* name, const char* label, const char* components, double scale)
{
    SPString key(name);
    units_list(key) = DerivedUnit(name, label, components, scale);
    return (GenieWorkspace)units_list(key);
}

/*
 * Look up an encoded unit string, returning a reference to the "DerivedUnit" structure.
 * If the unit does not exist, an entry is created and a reference to the new instance returned
 */
const DerivedUnit& lookup_unit(const String& encoded_unit)
{
    static const int MAXITEMS(4);	// "name", "label", "components" and "scale", separated by ';'
    String temp(encoded_unit);
    SPString key(String(temp.before(';')));
    if (!units_list.isin(key))		// Is it on the list?
    {
	String items[MAXITEMS];
	int nitems = split(encoded_unit, items, MAXITEMS, ';'); 
	if (nitems == MAXITEMS)
 	{
	    create_unit(items[0], items[1], items[2], atof(items[3]));
	}
	else
	{
	    gerr << "lookup_unit: corrupted encoded_unit string, \"" << encoded_unit << "\"" << endl;
	}
    }
    return units_list(key);
}

const DerivedUnit& lookup_unit(const GenieWorkspace& gw)
{
    SPString key(gw("name"));
    if (!units_list.isin(key))		// Is it on the list?
    {
	create_unit(gw("name"), gw("label"), gw("components"), gw("scale"));
    }
    return units_list(key);
}

static String unit_properties(const String& key_string)
{
    String res;
    SPString key(key_string);
    if (units_list.isin(key))
    {
        DerivedUnit& unit = units_list(key);
        cat("Name: ", unit.name(), res);
	res += "\nDimensions: ";
        res += unit.dimensions().asString(); 
    }
    else
    {
	cat("No matching unit for  ", key_string, res);
//	error_add("get_unit_properties", res, "");
    }
    return res;
}

// Return information about the unit, such as it's name and dimensions
String get_unit_properties(const String& encoded_unit)
{
    String temp(encoded_unit);
    return unit_properties(String(temp.before(';')));	// key (name) is string up to first ';'
}

String get_unit_properties(const GenieWorkspace& gw)
{
    return unit_properties((GenieString)gw("name"));
}

double unit_convert(const GenieWorkspace& unit1, const GenieWorkspace& unit2)
{
    return lookup_unit(unit1).convertTo(lookup_unit(unit2));
}

GenieWorkspace unit_multiply(const GenieWorkspace& unit1, const GenieWorkspace& unit2)
{
    return (GenieWorkspace)(lookup_unit(unit1) * lookup_unit(unit2));
}

GenieWorkspace get_unit_dimensions(const GenieWorkspace& unit)
{
    return (GenieWorkspace)lookup_unit(unit).dimensions();
}

String get_known_units()
{
    String list;
    for(int i=0; i<units_list.scalar(); i++)
    {
	const DerivedUnit& u = units_list[i].value();
	list += u.name();
	list += '\n';
    }
    return list;
}

#ifdef __GNUG__
template class Assoc<DerivedUnit>;
template class Binar<DerivedUnit>;
template class SPList<DerivedUnit>;
template class SPListBase<DerivedUnit>;
template class SPList< Binar<DerivedUnit> >;
template class SPListBase< Binar<DerivedUnit> >;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template Assoc<DerivedUnit>
#pragma define_template Binar<DerivedUnit>
#pragma define_template SPList<DerivedUnit>
#pragma define_template SPListBase<DerivedUnit>
#pragma define_template SPList< Binar<DerivedUnit> >
#pragma define_template SPListBase< Binar<DerivedUnit> >
#endif /* __DECCXX */

#if 0
// Angstroms
static const SIUnitElement angstrom_list[] = 
{ 
    {SIUnit::metre, SIUnit::NoPrefix, 1.0} 
};

static int len_angstrom_list = sizeof(angstrom_list) / sizeof(SIUnitElement);

static DerivedUnit Angstrom("Angstrom", angstrom_list, len_angstrom_list, 10.0);

ostream junk = (gerr << Angstrom << endl);
#endif
