/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         module
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Class to encapsulate dynamic loading of user programs
 *
 */
#include "geniedefs.h"
#include <fstream.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gnustring.h>
#if HAVE_DLFCN_H
#   include <dlfcn.h>
#elif defined(__VMS)
#   include "dlfcn_vms.h"
#elif defined(__hpux)
#   include <dl.h>
#   define RTLD_NOW 1
static void* dlopen(const char* path, int val)
{
    int flags;
    if (val == RTLD_NOW)
	flags = BIND_IMMEDIATE;
    else
	flags = BIND_DEFERRED;
// Need BIND_NOSTART ?
    flags |= BIND_VERBOSE;
    flags |= DYNAMIC_PATH;
    return shl_load(path, flags, 0L);
}

static const char* dlerror() 
{ 
    return strerror(errno); 
}

static int dlclose(void* handle) 
{
    return shl_unload((shl_t)handle);
}

static void* dlsym(void* handle, const char* name)
{
    void* value = NULL;
    shl_t temp_handle = (shl_t)handle;
    if (handle == 0)
	temp_handle = PROG_HANDLE;
// Could use TYPE_UNDEFINED to pick up anything
    if (shl_findsym(&temp_handle, name, TYPE_PROCEDURE, &value) == 0)
    {
	return value;
    }
    else
    {
	return 0;
    }
} 
#else
#   define RTLD_NOW 1
    static void* dlopen(const char* path, int val) { return 0; }
    static const char* dlerror() { static const char* mess = "Sorry - dynamic module loading not available"; return mess; }
    static int dlclose(void* handle) { return -1; }
    static void* dlsym(void* handle, const char* name) { return 0; }
#endif	/* HAVE_DLFCN_H */
#include "module.h"
#include "Genie_functions.h"

static const char* rcsid = "@(#)$Id: module.cxx,v 1.15.2.5 1997/09/01 07:41:45 faa Exp $";
USE_RCSID(rcsid);

static const char* get_dlerror()
{
    const char* temp = dlerror();
    return temp == 0 ? "No dlerror() message" : temp;
}

    
// Attempt to fix LD_LIBRARY_PATH to search the current directory if it doesn't already
//static int fix_ld_path()
//{
//    const char* path = getenv("LD_LIBRARY_PATH");
//    if (path == 0)
//    {
//	return putenv(strcpy((char*)malloc(1+strlen("LD_LIBRARY_PATH=.")), "LD_LIBRARY_PATH=."));
//    }
//    String new_path;
//    cat("LD_LIBRARY_PATH=", path, new_path);
//    if (!new_path.contains(":.:"))
//    {
//	new_path += ":.:";
//	return putenv(strcpy((char*)malloc(1+strlen(new_path)), new_path));
//    }
//    return 0;
//}
//    
//static int fix_ld_path_ret = (fix_ld_path() != 0 ? printf("WARNING: fix_ld_path() failed\n") : 0);
//

// return -1 on error, else position in list
int Module::load(const char* pathname, const char* description)
{
    module_item item;
    void* handle;
    String mod_pathname;
// position == -1 -> module not yet loaded
    int position = getModulePosition(pathname);
    if (position != -1)
    {
	unload(position);	
    }
    if ((handle = dlopen(pathname, RTLD_NOW)) == 0)
    {
// On some machine the current directory is not searched unless 
// the name begins with "./"
    	cat(CURRENT_DIRECTORY_PATH, pathname, mod_pathname);
	handle = dlopen(mod_pathname, RTLD_NOW);
    }
    if (handle == 0)
    {
	error_add("Module/Load", get_dlerror(), "");
	return -1;
    }
    item.handle = handle;
    item.path = pathname;
    item.description = description;
    item.load_time = time(0);
    if (position == -1)
    {
	m_list.push(item);
	return m_list.count() - 1;
    }
    else
    {
	m_list[position] = item;
	return position;
    }
}

// returns pointer to function, NULL on error
void* Module::getFunctionPointer(const char* name, ModuleFunctionType type)
{
    int i;
    String local_func(name);
#ifdef HAVE_LIBF2C
    if (type == FortranModuleFunction) 
    { 
	if (local_func.contains("_"))	// f2c mangles names containing "_" differetly
	    local_func += "__";
	else
	    local_func += "_"; 
    }
#elif !defined(__VMS)
    if (type == FortranModuleFunction)
    { 
	local_func += "_"; 		// Usual UNIX name mangling	
    }
#endif /* HAVE_LIBF2C */
    void* funcptr(0);
    for(i=m_list.count()-1; i >= 0; i--)    
    {
	if ( m_list[i].handle != 0 && 
             (funcptr = dlsym(m_list[i].handle, (const char*)local_func)) != 0 )
	{
	    return funcptr;
	}
    }
// Try again with an all lowercase function name
    local_func.downcase();
    for(i=m_list.count()-1; i >= 0; i--)    
    {
	if ( m_list[i].handle != 0 && 
             (funcptr = dlsym(m_list[i].handle, (const char*)local_func)) != 0 )
	{
	    ginf << "MODULE: function name \"" << local_func << "\" used for \"" << name << "\"" << endl;
	    ginf << "MODULE: If this is correct, please lowercase your reference to this function" << endl; 
	    return funcptr;
	}
    }
    *g_err << "MODULE: Unable to locate function \"" << name << "\" in loaded modules" << endl;
    *g_err << "MODULE: " << get_dlerror() << endl;
    return 0;
}

// unload module - return 0 in success, -1 on error
int Module::unload(const char* pathname)
{
    int position = getModulePosition(pathname);
    if (position != -1)
    {
	return unload(position);	
    }
    else
    {
	return 0;
    }
}

// unload module - return 0 in success, -1 on error
int Module::unload(int i)
{
    if (m_list[i].handle == 0)
    {
	return 0;	// nothing to unload
    }
    if (dlclose(m_list[i].handle) == -1)
    {
	error_add("Module/Unload", get_dlerror(), "");
	m_list[i].handle = 0;
	return -1;
    }
    else
    {
	m_list[i].handle = 0;
	return 0;
    }
}

int Module::list(ostream& output_stream)
{
    int j(0);
    struct stat stat_buffer;
// File, Description,  load time
    for(int i=0; i<m_list.count(); i++)
    {
	if (m_list[i].handle != 0)
	{
	    output_stream << "[" << ++j << "] " << m_list[i].path << " (" << m_list[i].description << ") loaded on " 
		      << ctime(&m_list[i].load_time);
	    if ((stat(m_list[i].path, &stat_buffer) == 0) && (stat_buffer.st_mtime > m_list[i].load_time))
	    {
	        output_stream << "WARNING: \"" << m_list[i].path << "\" modified, but not re-loaded\n";
   	    }
	}
    }
    return 0;
}

const char* Module::getPath(int i)
{
    return m_list[i].path;
}

// return position of module path, or -1 if not loaded
int Module::getModulePosition(const char* path)
{
    for(int i=m_list.count()-1; i >= 0; i--)    
    {
	if (!strcmp(path, getPath(i)))
	{
	    return i;
	}
    }
    return -1;
}

Module::~Module()
{
    for(int i=0; i<m_list.count(); i++)
    {
	if (unload(i) != 0)
	    break;
    }
    m_list.reset();
}

// Just needed to satisfy SPList - not used at all
int operator<(const module_item& a, const module_item& b)
{
    return ((long)a.handle - (long)b.handle) < 0;
}

// Instantiate templates
#ifdef __GNUG__
template class SPList<module_item>;
template class SPListBase<module_item>;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template SPList<module_item>
#pragma define_template SPListBase<module_item>
#endif /* __DECCXX */
