/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <assert.h>
#include <iostream.h>
#include <stdio.h>

#if defined (HAVE_UNISTD_H)
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include <stdlib.h>
#include <ctype.h>
#ifdef _WIN32
#	include <io.h>
#else
#	include <sys/wait.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <gnustring.h>
#include <signal.h>
#include <setjmp.h>
#if !(defined(__VMS) && __VMS_VER < 70000000) && !defined(_WIN32)
#	include <sys/utsname.h>
#endif /* !(defined(__VMS) && __VMS_VER < 70000000) */
#include <time.h>
#include <limits.h>
#include <gnuregex.h>

#include "parsedefs.h"
#include "gsignals.h"
#include "genielog.h"
#include "pgkill.h"
#include "smalltalk.h"

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

#if defined(__ultrix)
extern "C" 
{
    int putenv(const char* var);
}
#include "strcasecmp.h"
#endif /* __ultrix */

#ifdef _WIN32
#	include "strcasecmp.h"
#endif

#if defined(__VMS) && __VMS_VER < 70000000
extern "C" 
{
    int putenv(const char* var);
#   define unlink(__path)	remove(__path)
#   define strdup(__a) strcpy((char*)malloc(1+strlen((const char*)__a)), (const char*)__a)
}
#include "vms_uname.h"
#include "strcasecmp.h"
#endif /* defined(__VMS) && __VMS_VER < 70000000 */

#if defined(__VMS) && !defined(__VMS_POSIX)
#   define TTYNAME() ttyname()
#else
#ifdef _WIN32
#	define TTYNAME() "CON:"
#else
#   define TTYNAME() ttyname(Stdout)
#endif
#endif

static const char* rcsid = "@(#)$Id: parse.cxx,v 1.130.2.5 1997/09/03 16:53:36 faa Exp $";
USE_RCSID(rcsid);

// GNU Readline
extern "C"
{
#    include <readline/readline.h>
#    include <readline/history.h>
}

static void initialize_readline();
static void stripwhite(char* string);

/*
 *     CPPFunction *genie_command_completion;
 *     Function *genie_command_generator;
 */

#include "genie_commands.h"	// Automatically generated list of commands

// Somewhere to store the command history
String history_file;

// This session id is initialized by the first call to logit and reset by
// Kill_smalltalk.  Each kept session has a different session id.
static time_t session_id = 0;

/*** GCL boot files constituting the smalltalk GENIE ****/
const char *basefiles[][2] =
{
    /* Basic GENIE system for running procedures etc. */
    { "_STFIXES", "stfixes.gcl" },
    { "_STEXTENSIONS", "stextensions.gcl" },
    { "_EXTERNAL", "external.gcl" },

    /* Predefined variables, constants */
    { "_SETUP", "setup.gcl" },

    /* Load the PROCEDURE/FUNCTION running subsystem */
    { "_FUNCTION", "function.gcl" },

    /* Genie & smalltalk types/extensions */
    { "_MATHS", "maths.gcl" },					    /* Basic mathematical functions (generic) */
    { "_WORKSPACE", "workspace.gcl" },				    /* Workspace class */
    { "_MULTIDIMENSIONALARRAY", "multidimensionalarray.gcl" },	    /* needed by array.gcl */
    { "_ARRAY", "array.gcl" },
    { "_INTEGERARRAY", "integerarray.gcl" },
    { "_REALARRAY", "realarray.gcl" },
    { "_GENIETYPES", "genietypes.gcl" },			    /* GXreal, GXstring & GXinteger */

    /* System primitive procedures */
    { "_BUILTINS", "builtins.gcl" },
    { "_BASEPRIMITIVES", "baseprimitives.gcl" },
    { "_DIRECTORY", "directory.gcl" },

    { "_ERRORS", "errors.gcl" },				    /* GCL <-> C++ error handling functions */

    /* Neutron/science specific GENIE from now on */
    { "_DATAMODEL", "datamodel.gcl" },				    /* Data model for Time-of-Flight data */
    { "_UNITS", "units.gcl" },					    /* Units and conversion functions for measurements */
    { "_EXPERIMENT", "experiment.gcl" },                            /* Units and conversion functions for measurements */
    { "_WORKSPACE_USER", "workspace_user.gcl" },		    /* Workspace add ons */
    { "_ASCIIFILE", "asciifile.gcl" },

    /* Mainly hacking away in GCL until the end */
    { "_DRIVERS", "drivers.gcl" },				    /* Auto generated from ../pgplot/makefile */
    { "_IO", "io.gcl" },
    { "_OBJECTGRAPHICS", "objectgraphics.gcl" },
    { "_GRAPHICALOBJECTS","GraphicalObjects.gcl"},
    { "_GRAPHICSGLOBALS", "GraphicsGlobals.gcl" },
    { "_COLOUR", "colour.gcl" },
    { "_DEVICE","Device.gcl"},
    { "_DRAWABLEOBJECTS","DrawableObjects.gcl"},
    { "_DRAWTWOD","DrawTwoD.gcl"},
    { "_TWODCONTOUR","TwoDContour.gcl"},
    { "_TWODCELL","TwoDCell.gcl"},
    { "_DRAWTWODCONTOUR","DrawTwoDContour.gcl"},
    { "_DRAWTWODCONTOURARRAY","DrawTwoDContourArray.gcl"},
    { "_DRAWTWODCONTOURFUNCTION","DrawTwoDContourFunction.gcl"},
    { "_DRAWTWODCONTOURLABEL","DrawTwoDContourLabel.gcl"},
    { "_DRAWTWODCELL","DrawTwoDCell.gcl"},
    { "_DRAWTWODCELLARRAY","DrawTwoDCellArray.gcl"},
    { "_DRAWTWODCELLFUNCTION","DrawTwoDCellFunction.gcl"},
    { "_DRAWTWODCELLWEDGE","DrawTwoDCellWedge.gcl"},
    { "_WINDOW","Window.gcl"},
    { "_SCALEDWINDOW","ScaledWindow.gcl"},
    { "_TWODWINDOW","TwoDWindow.gcl"},
    { "_MULTIPLOT","Multiplot.gcl"},
    { "_PICTURE","Picture.gcl"},
    { "_GRAPHICALOBJECTSMISC","GraphicalObjectsMisc.gcl"},
    { "_DEVICEMISC","DeviceMisc.gcl"},
    { "_ADDTOLIST","AddToList.gcl"},
    { "_DRAWONED","DrawOneD.gcl"},
    { "_DRAWONEDGRAPH","DrawOneDGraph.gcl"},
    { "_DRAWMARKERS","DrawMarkers.gcl"},
//    { "_DRAWORIGINALWINDOW","drawOriginalWindow.gcl"},
    { "_ISINAWINDOWII","isInAWindowII.gcl"},
    { "_ISOVERLAPPED","isOverlapped.gcl"},
    { "_ISOVERLAPPEDII","isOverlappedII.gcl"},
    { "_ISINAWINDOW","isInAWindow.gcl"},
    { "_FINDOVERLAPPINGPOINT","FindOverLappingPoint.gcl"},
    { "_MAKENEWWINDOW","makeNewWindow.gcl"},
    { "_GETNEWCOORDINATES","GetNewCoordinates.gcl"},
    { "_GETCOORDINATES","GetCoordinates.gcl"},
    { "_FINDWINDOWS","FindWindows.gcl"},
    { "_ZOOM","zoom.gcl"},
    { "_DRAWPOLYLINE","DrawPolyLine.gcl"},
    { "_DRAWTEXT","Drawtext.gcl"},
    { "_POLYGON","Polygon.gcl"},
//    { "_ALLOCATE","Allocate.gcl"},
    { "_DRAWANNOTATION","DrawAnnotation.gcl"},
    { "_DRAWONEDAXES","DrawOneDAxes.gcl"},
    { "_DRAWONEDGRATICULE","DrawOneDGraticule.gcl"},
    { "_DRAWONEDLABELS","DrawOneDLabels.gcl"},
    { "_DRAWONEDTITLE","DrawOneDTitle.gcl"},
    { "_DRAWVERRORS","DrawVerrors.gcl"},
    { "_DRAWHERRORS","DrawHErrors.gcl"},
    { "_DRAWONEDERRORS","DrawOneDErrors.gcl"},
    { "_DRAWMULTIPLOT","DrawMultiplot.gcl"},
    { "_DRAWHISTOGRAM","Drawhistogram.gcl"},
    { "_WHICHWINDOW","whichWindow.gcl"},
    { "_NEWGRAPHICS", "newGraphics.gcl" },
    { "_MISCNEWGRAPHICS", "MiscNewGraphics.gcl" },
    { "_SELECT", "select.gcl"},
    { "_OLD_GRAPHICS", "old_graphics.gcl" },
    { "_GRAPHICS_V2", "graphics_v2.gcl" },
    { "_GENIE_V2_COMPAT", "genie_v2_compat.gcl" },
    { "_PEAK", "peak.gcl"},
    { "_MODULE", "module.gcl"},
    { "_INFO", "info.gcl"},
    { "_HELP", "help.gcl" },					           /* procedures to implement /HELP qualifier on commands */ 
    { NULL, NULL }						    /* terminator */
};

/* Establish a signal return point where this is used */
#define CHECKPOINT_RETURN   if (!Test_option(CRASHDEBUG)) Handle_signal(sigsetjmp(context, 1))

/*** Command line options ***/

enum	options
{
    YACCDEBUG=1,
    GENDEBUG=2,
    CRASHDEBUG=4,
    SAVEIMAGE=8,
    CAPTIVE_START=16,
    CAPTIVE_STOP=32,
    FILECHECK=64,
    NOSTOP=128
};

static	int options = 0;						/* Genie options word */
static	char image_file_name[MAXVARLEN];
static	int Parse_options(char *argv[]);
static	int Test_option(int flag );
static	void Help_options(void);

// Starting and controlling smalltalk
static void	Start_smalltalk( char *exec_arge[] );
int		Send_to_smalltalk(node *message);
int		Load_smalltalk(const char file[], int mode);

// Parser control
int	    	Compile_input(int input, int mode);
void		Interpret_input(int input);

// Miscellaneous
void		save_file( char filename[] );
char		*Append(const char *str1, const char *str2);
time_t		mdate(const char *filename);
static void	initialize(void);
static int	Genie_load(void);
static void	buffer_diagnostics(int lines);

extern int malloc_count;

/* Stream for redirecting output to during a build */
FILE *debugout = 0;
FILE *g_type_file = 0;

static int genie_nostop = 0;	/* do we stop on smalltalk error */

/*
 * Main parser control program Starts here
 * =======================================
 */
int parser_main(int argc, char *argv[], char *arge[],
                istream &in,
                genie_ostream &out, genie_ostream &inf,
                genie_ostream &err, genie_ostream &dbg)
{

    // equivalence I/O streams for use in the rest of Open GENIE
    g_in = &in;
    g_out = &out;
    g_inf = &inf;
    g_err = &err;
    g_dbg = &dbg;

    if ( ! Parse_options(argv) )
        PERROR("Invalid command line option(s) specified");

    if (Test_option(NOSTOP))
	genie_nostop = 1;

    Set_signals();
    CHECKPOINT_RETURN;

    if (Test_option(CAPTIVE_START))
    {
        // Start up a captive Open GENIE for data access or for control
        // from a TCL/TK session

        if (!session_id)                // check for already active session
        {
            logit("Starting Kept");     // Internet Log this invocation of GENIE
            Start_smalltalk(arge);      // Also sets smalltalk signal handling

            // Now boot GENIE from image files if necessary
            if ( Genie_load() == -1 )
	            gprintef("Essential system files not found, cannot start GENIE\n	Check definition of symbol GENIE_DIR\n");
        }
        else
            gerr << "Session " << session_id << " is still active, only one session allowed";
    }

    else if (Test_option(CAPTIVE_STOP))
    {
        // Close down an Open GENIE Kept session
        // (from a TCL/TK session or Data access)

        // close down session but keep running
        Kill_smalltalk(0);
    }

    else
    {
        // Start up a fully interactive Open GENIE for data access or for control
        // from a TCL/TK session

        logit("Starting Interactive");  	// Internet Log this invocation of GENIE
        initialize();
        Start_smalltalk(arge);

        // Now boot GENIE from image files if necessary
        if ( Genie_load() == -1 )
 	{
	    gprintef("Essential system files not found, cannot start GENIE\n	Check definition of symbol GENIE_DIR\n");
	}
	else
	{
	    if (!Test_option(SAVEIMAGE))
                Interpret_input(Stdin);
	}
        // Stop the interactive session here
        Kill_smalltalk(1);
    }
    if ( Test_option(GENDEBUG) )
    {
	fclose(debugout);
    }

    return 0;
}


/*

    Standard symbols or logical names for Open GENIE components.  This list is intended
    to be minimal (OK I know there isn't a Java GENIE yet!).

    GENIE_DIR                   - Essential system files (genie can get most from this anyway)
    GENIE_TCL_DIR		- startup files for TCL interface
    GENIE_LIBRARY_DIR           - General modules, bits of GCL etc.
    GENIE_INCLUDE_DIR		- where module headers etc are
    GENIE_USER_LIBRARY_DIR      - Pointer to user written routines for automatic search

    GENIE_EXE                   - Location of startup executable
    GENIE_SHARED_LIBRARY        - Bulk of GENIE code shared library/DLL

    GENIE_SMALLTALK_IMAGE       - User Language specific (compiled) image
    GENIE_TCL_IMAGE             - eg ICE compiled TCL
    GENIE_JAVA_IMAGE

    GENIE_SMALLTALK_INIT        - User Language specific initialization (needing interpretation)
    GENIE_GCL_INIT              - User GCL file (eg GENIEINIT.GCL)
    GENIE_TCL_INIT              - User TCL interface initializtion file

    GENIE_DEFAULT               - User set default directory for running genie in
*/

// Functions stinit() and gclinit()
// execute any GCL or smalltalk code which must be re-done on each
// invocation of Open GENIE.  Examples are initialization of workspaces,
// global parameters, interrupt enabling.  Remember, most globals etc are
// saved with the image but this may be an appropriate place to reinitialize
// them.

static void stinit()		// smalltalk bits
{
    char buffer[SHORTCHARBUFF];

    sprintf(buffer, "Smalltalk at: #gXUUlocalUtty put: '%s'!", TTYNAME());
    evalSmalltalkCommand(buffer);
    if (Test_option(SAVEIMAGE)) evalSmalltalkCommand("Smalltalk growTo: 20 * 1024 * 1024!");
    evalSmalltalkCommand("Smalltalk enableGC: true!");

    return;
}

static void gclinit()		// GCL bits
{
    evalGCLCommand("__initgraphics");
    evalGCLCommand("__initio");

    return;
}

/*
 * Build a running genie up from all the appropriate bits of smalltalk or restore a
 * saved image file instead.
 */
static int Genie_load()
{
    int i, retval;
    time_t date;
    char *cp, loadpath[SHORTCHARBUFF];
    enum { SYMBOL=0, FILENAME=1 };

    (cp=getenv("GENIE_DIR")) ? strcat(strcpy(loadpath, cp), DIRECTORY_SEPARATOR) : 
			       strcpy(loadpath, CURRENT_DIRECTORY_PATH);

    date = mdate((cp=getenv("GENIE_SMALLTALK_IMAGE"))?cp:Append(loadpath, "genie.im"));

    CHECKPOINT_RETURN;

    if (Test_option(FILECHECK|SAVEIMAGE))
       for ( i=0; basefiles[i][0]; i++ )
  	{
//	      evalSmalltalkCommand("(Array new: 10) at: 1 put: nil!");	// Hack for copyWith: problem
          retval = load_file(basefiles[i][SYMBOL], Append(loadpath, basefiles[i][FILENAME]), STARTUP, date );
	  if (retval > 0)		// errors on loading file
	  {
	      gprintef("GENIELOAD: Error on loading file \"%s\"\n", basefiles[i][FILENAME]);
	  }
	  if (retval == -1)		// File not found
	 	return -1;
	}

    gclinit();			// run any GCL initialisation

    // User customizeable file which is automatically loaded if present.
    load_file("GENIE_GCL_INIT", "genieinit.gcl", BATCH, (time_t) 0, 0);

    if ( Test_option(SAVEIMAGE) )
    {
        // Do this here as it will be the first thing after a non-synchronized load
        evalSmalltalkCommand("Smalltalk snapshot: 'genie.im' !\n");
        unlink("gst.im");                               // Remove unnecessary files
        gprintif("Saved image file to \"genie.im\" and boot file to \"boot.st\"\n");
	gprintif("Also save procedure definitions to \"procdefs.out\"\n");
        fclose(debugout);
	fclose(g_type_file);
    }

    return 0;
}



time_t mdate(const char *filename)
{
    static struct stat buf;
    if ( stat(filename, &buf) == -1 )
	return (time_t) 0;
    else
	return buf.st_mtime;
}



char *Append(const char *str1, const char *str2)
{
    static char buffer[SHORTCHARBUFF];

    strcpy(buffer, str1);
    strcat(buffer, str2);

    return buffer;
}


/*
 * Log the startup and disable buffering ( any other version control things here too)
 */
static void initialize(void)
{

    gprintf("%s\n", (const char*)&version[4]);		/* (Don't print the @(#) bit) */
    gout << "(Running on " << getenv("GENIE_MACHINE")<< ")" << '\n';    /* (env var set by logit() ) */
    gout << '\n';
    gout << "   ******************************************************************" << '\n';
    gout << "   *  Copyright (C) 1996,1997 CCLRC Rutherford Appleton Laboratory  *" << '\n';
    gout << "   *  Open GENIE comes with ABSOLUTELY NO WARRANTY; for details     *" << '\n';
    gout << "   *  type `warranty'.  This is free software, and you are welcome  *" << '\n';
    gout << "   *  to redistribute it under certain conditions; type `licence'   *" << '\n'; 
    gout << "   *  for details.                                                  *" << '\n';
    gout << "   ******************************************************************" << '\n';
    gout << endl;

    using_history();				/* set up for GNU readline history */

#if defined(__VMS) && !defined(__VMS_POSIX)
    cat(getenv("SYS$LOGIN"), ".genie_history", history_file);
#else
    cat(getenv("HOME"), "/.genie_history", history_file);
#endif
    if (read_history((char*)(const char*)history_file) != 0) /* read in previous session commands */
    {
        gprintif("WARNING: unable to read commands from %s\n", (const char*)history_file);
        gprintif("--------- creating a new history file\n");
	add_history("");
        if (write_history((char*)(const char*)history_file) != 0) /* read in previous session commands */
            gprintif("WARNING: cannot create a history file", (const char*)history_file);
    }

    // For a forced save, remove "genie.im" & "gst.im" before the load begins
    // reassign the debugging output to build "boot.st"
    // This must be done before the smalltalk interpreter is started.
    String env_temp;
    if ( Test_option(SAVEIMAGE) )
    {
	cat("GENIE_GCL_INIT=", NULL_OUTPUT_DEVICE, env_temp);
        putenv(strdup(env_temp));
        rename("genie.im", "genie.im.old");
	unlink("gst.im");
        debugout = fopen("boot.st","w");
        rename("procdefs.out", "procdefs.out.old");
	g_type_file = fopen("procdefs.out", "w");
    }
    if ( Test_option(GENDEBUG) )
    {
	gprintif("%s\n", "Writing Smalltalk to screen and \"debug.st\"");
        debugout = fopen("debug.st","w");
    }

    return;
}


static int Parse_options( char *argv[] )
{
    char *flags;

    options = 0;
    for ( ; *++argv; )
	if (argv[0][0] == '-')
	    for (flags = &argv[0][1]; *flags; flags++)
		switch (*flags)
		{
		case 'g':			/* Debug smalltalk output */
		    options |= GENDEBUG;
		    break;
		case 's':			/* Save an image file */
		    options |= SAVEIMAGE;
		    break;
		case 'y':			/* Debug yacc parser */
		    options |= YACCDEBUG;
		    break;
		case 'd':			/* Debug yacc parser */
		    options |= CRASHDEBUG;
		    break;
		case 'B':			/* Debug yacc parser */
		    options |= CAPTIVE_START;
		    break;
		case 'E':			/* Debug yacc parser */
		    options |= CAPTIVE_STOP;
		    break;
		case 'c':			/* Debug yacc parser */
		    options |= FILECHECK;
		    break;
		case 'n':			/* Don't stop on error */
		    options |= NOSTOP;
		    break;
		default:
		    Help_options();
		    return 0;
		}
    return 1;
}

static void Help_options()
{
    gprintf("Options understood by GENIE are:\n");
    gprintf("\t-g\tDebug code generation\n");
    gprintf("\t-s\tSave an execution image\n");
    gprintf("\t-y\tDebug parser\n\n");
    gprintf("Symbols translated:\n");
    gprintf("\tGENIE_GCL_INIT\t\tUser initialization file ( default ./genieinit.gcl )\n");
    gprintf("\tGENIE_SMALLTALK_IMAGE\tSmalltalk saved image file ( default ./genie.im )\n");

    return;
}



static int Test_option( int flag )
{
    return (flag & options) != 0;
}



/*
 * This function defines the streams and defaults required for
 * a subprocess running the smalltalk interpreter.
 */
static void Start_smalltalk(char *exec_arge[])
{
    /* in the absence of any other info, this should be sufficient to kick of a default process */
    const char *cp, *exec_argv[5];

    /* Construct exec_argv from symbols or default values */
    exec_argv[0] = "gmain";	    	/* default will be "./gmain -qg -I ./genie.im" */
    exec_argv[1] = "-qg";
    exec_argv[2] = "-I";
    exec_argv[3] = (cp=getenv("GENIE_SMALLTALK_IMAGE"))?cp:CURRENT_DIRECTORY_PATH "genie.im" ;
    exec_argv[4] = NULL;

    smalltalkArgs(4, (char**)exec_argv);
    initSmalltalk();
    initializeSmalltalkSignals();		// Include smalltalk in our signal handler
    Set_signals();
    pseudoTopLevelLoop();
    stinit();			// Genie specific smalltalk initialisation
    return;
}



/*
 * Start off a child process and connect up three streams for input, output
 * and error (the pipes should be already opened).
 */

/*
 * Kill off the smalltalk subprocess cleanly and check for anything unusual
 * (eg it has already gone !)
 */
void Kill_smalltalk(int reallyquit)
{
    Unset_signals();						/* best not to catch ourselves with this one ! */

    logit("Stopping cleanly");				/* Log the end of this session */
    session_id = 0;                         // (Need to zero this here for kept sessions)

    if (history_truncate_file((char*)(const char*)history_file, 500) != 0) 		/* save commands from this session */
	    gprintif("WARNING: error closing session log file %s\n", (const char*)history_file);

    unlink("gst.im");	// In case we did a save this session

// We do not want to prompt on a "-s" option as we are probably running non-interactively
    if (!Test_option(SAVEIMAGE) && reallyquit)
        kill_pgplot_server();					/* Kill any "pgxwin_server" process */

    if (reallyquit)
        exit(0);
}



/*
 * Sends a message to the smalltalk interpreter process either as a synchronized operation
 * or as a datagram style string.
 */
int Send_to_smalltalk(node *message)
{
    char *buffer;
    int retval;

    buffer = (char *)malloc(nlength(message)+1);
    buffer[0] = '\0';

    printNode(buffer, message);

    retval = evalSmalltalkCommand(buffer);	/* send the generated code to smalltalk */

    if ( Test_option(GENDEBUG|SAVEIMAGE) )
        if ( Test_option(SAVEIMAGE) )
            fprintf(debugout, "%s", buffer);
        else
	{
            fprintf(debugout, "%s", buffer);
            gprintdf("%s", buffer);
	}

    free(buffer);
    return retval;
}


// Return -1 on open failure, else number of errors on execution of file
int Load_smalltalk(const char path[], int mode)
{
    int fdesc, status(-1);

    if ( (fdesc=open((char*)path, O_RDONLY, 0)) != (-1) )
    {
        gprintif("Loading file %s ", path);
	    status = Compile_input(fdesc, mode);	    /* Lex takes its input from the open file */
	    if ( status > 0 )
	        g_printef("\n%d Error(s) occured while compiling file %s\n", status, path);

	    close(fdesc);
    }

    return status;			    /* This will return -1 on an open failure */
}


static void buffer_diagnostics(int lines)
{
    int bspace;

    gprintif("\t%d lines scanned/compiled", lines);
    if( (bspace = check_buffers()) > 0 )
    {
        gprintdf("\tbuffer space remaining = %d, ", lines, bspace);
        p_stats();
        gprintif("\n");
    }
    else
    {
    	gprintef("Buffer space overflowed on compilation - code not loaded properly\n");
        p_stats();
    	gprintif("\tOne procedure is probably too long or complex,\n\ttry splitting it into smaller procedures\n");
    }
}

static int getc_readline()
{
    // Procedure ostensibly to get a character directly from the standard
    // input, but to actually do this via a readline buffer instead.

    static char old_line[MAXSTRLEN] = "";
    static char *lp = old_line + 1;                     // Sets initial condition to be ready to read
    static bool end_of_line_to_send = FALSE;             // Keep a note of when an EOL is needed
    static char *lp_unexpanded = 0, *lp_expanded = 0;
    int res, expand;
    bool done;

        if( *lp )                                       // Return characters from the buffer
            res = *lp++;

        else if (end_of_line_to_send)                   // Put back end of lines to keep lexer happy
        {
            end_of_line_to_send = FALSE;
            res = '\n';
        }

        else                                            // get a new (complete) input line from the user
        {
	        done = FALSE;
	        while(!done)
	        {
	            if (lp_expanded != 0)
	            {
		        free(lp_expanded);
		        lp_expanded = 0;
	            }
                    while( (lp_unexpanded = readline(prompt_mode()?">> ":"")) == 0 )
		    {
			crlf();
			rl_on_new_line();	// Trap NULL returns from e.g. ^D
		    }
#if 0	/* I've turned off ! history subs as they cause problems */
	            expand = history_expand(lp_unexpanded, &lp_expanded);
		    free(lp_unexpanded);
#else
		    expand = 0;
		    lp_expanded = lp_unexpanded;
#endif
		    lp_unexpanded = 0;
		    switch(expand)
		    {
		        case 0:		        // no expansion too place, so no need to echo result
			    done = TRUE;
			    break;

		        case 1:		        // history expansion took place
		            done = TRUE;	// and now fall though to next statement ...
		        case 2:		        // line should be displayed, but not executed (e.g. it was a "!string:p")
		        case -1:		    // error in expanding, error message returned
		            fprintf(rl_outstream, "%s\n", lp_expanded);
	 	            rl_on_new_line();
			    break;

		        default:
		            fprintf(rl_outstream, "Unusual code returned from history_expand(): %i\n", expand);
	 	            rl_on_new_line();
			    break;
		    }
	    }
	    lp = lp_expanded;
        
            if (lp)
            {
                end_of_line_to_send = TRUE;             // make sure we send an EOL later on
                stripwhite(lp);                         
                if (*lp)
                {
                    if (strcmp(lp, old_line))           // Only store if non-null & different
                    {
                        add_history(lp);                        
                        append_history(1, (char*)(const char*)history_file); 		/* save commands from this session */
                    }
                        strcpy(old_line, lp);               // Keep a copy to avoid history duplication
                    res = *lp++;
                }
                else
                    res = ' ';                          // Send this back as a filler (can't return a NULL)
            }
            else
                res = EOF;
        }

    return res;
}

static char* evalGCLCommand_lp;

// String may be altered to strip white space
int evalGCLCommand(const char* comm)
{
    int status(0);
    evalGCLCommand_lp = 0;
    if (comm)
    {
        char *command = strcpy((char*)malloc(1+strlen(comm)), comm);
        stripwhite(command);
        if (*command)
	{
	    evalGCLCommand_lp = command;
	    status = Compile_input(-1, BUFFER);			// -1 is just an invalid file descriptor
	    if ( status > 0 )
	        gprintef("\n%d Error(s) occured while evaluating string \"%s\"\n", status, command);
	}
        free(command);
    }
    return status;
}


// Returns 0 on success, 1 on error
int evalSmalltalkCommand(const char* comm)
{
    static int call_count = 0;	// to check for recursion
    // Pass responsibility for clean signal recovery to the smalltalk here
    // ie, let it longjump around inside itself etc. and trust it to return
    // gracefully after evalCode!
//    if (call_count)
//    {
//	gprintef("ERROR: Smalltalk cannot be called recursively!!\n");
//	return 1;
//    }

    jumpsigs = 0;

    ++call_count;
    int retval = evalCode((char*)comm);
    --call_count;

    // take back responsibility for signal recovery & our own longjumping
    jumpsigs = 1;

    return retval;
}


// Return characters from buffer allocated by evalGCLCommand()
static int getc_buffer()
{
    if (evalGCLCommand_lp == 0)			// Called again after returning '\n', so return EOF
    {
	return EOF;
    }
    else if (*evalGCLCommand_lp)		// in middle of string, return next character
    {
	return *evalGCLCommand_lp++;
    }
    else 					// End of string, so put back the stripped '\n'
    {
	evalGCLCommand_lp = 0;
	return '\n';
    }
}

int getc_parser(FILE *fp, int mode)
{
    if (mode == INTERACTIVE)
        return getc_readline();
    else if (mode == BUFFER)
	return getc_buffer();
    else
        return getc(fp);
}


/*
 * This function communicates with yyparse() and yylex() to parse and compile
 * a complete file into smalltalk code.
 */
int Compile_input(int Input, int mode )
{
    int errors(0), lines, status=1;
    node *code;
    FILE *in_channel;

    in_channel = fdopen(Input, "r");				/* Get file descriptor for stream */

    fill_buffers();						            /* Fill parser buffers with sentinel values */
    initialize_parser_stats();					    /* Set line and error counts to 0 */

    while ( status )
    {
    	initialize_parser(mode, in_channel, Test_option(YACCDEBUG));	    /* Set up the parser */

    	status = yyparse();					            /* Start the parser */
        code = interrogate_parser(&errors, &lines);		            /* Get the results */

    	if (status)
    	{
            assert(code->valid);
    	    if ( nlength(code) && (mode != BUFFER) ) gprintf(".");
    	    if (Send_to_smalltalk(code))			    	    /* call back with each chunk of code */
	    {
		errors++;
		if (!genie_nostop)
		    status = 0;
	    }
    	}
        if (code->valid);
            deleteNode(code);
    }

//    sweep_nodes();
    if (mode != BUFFER)
    {
        gprintf("\n");
        buffer_diagnostics(lines);					    /* Check no errors occurred and print stats */
    }
    if (in_channel != 0)
	fclose(in_channel);
    return errors;
}



void Interpret_input(int Input)
{
    int errors, lines;
    volatile int status = 1;
    node *code;
    FILE *in_channel;
    int tmp_count;
//    static int Chris = (getuid() == 24753 ? 1 : 0);		/* Are we Chris Moreton-Smith? */

    initialize_readline();                                      /* prepare readline for action */
    in_channel = fdopen(Input, "r");		                /* get the standard input descriptor */

    CHECKPOINT_RETURN;

    while ( status )
    {
	initialize_parser_stats();
        initialize_parser(INTERACTIVE, in_channel, Test_option(YACCDEBUG));	                /* reset parser variables */

        // Reset dynamic call depth for debugging
        evalSmalltalkCommand("Function new depth: 0!");

        tmp_count = malloc_count;

	status = yyparse();				                                                    /* Start the parser */
	code = interrogate_parser(&errors, &lines);				                            /* get the results */

        if (status)
        {
            assert(code->valid);
            Send_to_smalltalk(code);        		                                        /* call back with each chunk of code */
        }
        if (code->valid)
            deleteNode(code);

//        if (Chris && (tmp_count = malloc_count - tmp_count + 2))	/* Only Chris gets these messages! */
//            gprintdf("Mallocator leaked %d bytes on the last command!\n", tmp_count);

    }

    fclose(in_channel);
    return;
}


// Returns 0 on success, -1 on error (no file), else number of errors on executing smalltalk in file
int load_file( const char symbol[], const char filename[], int mode, const time_t date , int warn)
{
    const char *cp, *filespec;
    time_t filedate;
    int retval(0);

    /* Load if the file is more recent than the date passed */
    filespec = (cp=getenv(symbol))?cp:filename;
    filedate = mdate(filespec);
    if(filedate==0) filedate=INT_MAX;			/* Go for very large date if file doesn't exist */

    if ( (filedate > date) || (date == 0) )
    {
  	retval = Load_smalltalk( filespec, mode );
	if ( (retval == -1) && warn )
	{
	    g_printef("\tError - Could not open file \"%s\"\n", filespec );
	    if ( cp ) gprintif("\t\tWas symbol %s defined correctly ?\n", symbol);
	}
    }
    return retval;
}



void save_file( char filename[] )
{
    node *tmp_node;

    tmp_node = newLeaf("");
    gprintif("Saving GENIE in file \"%s\" ...\n", filename);
    nprintf(&tmp_node, "Smalltalk snapshot: '%s'!\n", filename);
    Send_to_smalltalk(tmp_node);
    deleteNode(tmp_node);

    return;
}


/*
 * Log details to the internet on invocation or stopping of GENIE
 */

void logit( const char callinfo[] )
{
    static char buffer[LONGCHARBUFF];

#ifndef _WIN32
    struct utsname details;
#endif

    if (session_id == 0) session_id = time(NULL);

#ifndef _WIN32
    uname(&details);			/* Get OS and machine info */
// session_id is now done by genielog() internally
    sprintf(buffer, "Event = %s\nEnvironment = #%s#%s#%s#%s#%s\nVersion = %s",
    callinfo, details.nodename, details.machine, details.sysname, details.release, details.version, version);
#endif

    genielog(buffer);			/* Now send the information */

    /* set up a handy evironment variable with this information */
#ifdef __DECCXX
#	define MAJV 10000000
#	define MINV 100000
#	ifdef __ultrix
#		define __DECCXX_VER	10300000	/* Ultrix CXX doesn't define this - I use 1.3 so we'll assume that */
#	endif
    sprintf(buffer, "GENIE_MACHINE=%s:%s:%s (built with %s V%.1f)",
        details.machine, details.sysname, details.release, "DEC C++",
        __DECCXX_VER/MAJV + ((__DECCXX_VER-(__DECCXX_VER/MAJV)*MAJV)/MINV) /10.0);
#else
#	ifdef _WIN32
    sprintf(buffer, "GENIE_MACHINE=%s:%s:%s (built with %s V%d.%d)",
        "_M_IX86", "Windows/NT", "4.0", "Visual C++", (int)_MSC_VER, 2);
#	else
    sprintf(buffer, "GENIE_MACHINE=%s:%s:%s (built with %s V%d.%d)",
        details.machine, details.sysname, details.release, "GNU C++", (int)__GNUC__, (int)__GNUC_MINOR__);
#	endif
#endif
// according to AES, putenv may keep the pointer so make a copy first
    int temp_len = strlen(buffer);
    char* temp_str = (char*)malloc(temp_len + 1);
    putenv(strcpy(temp_str, buffer));
    return;
}


/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char* genie_command_generator(char* text, int state)
{
  static int list_index, len;
  const char *name;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      list_index++;

      if (strncasecmp (name, text, len) == 0)
        return strcpy((char*)malloc(1+strlen(name)), name);		/* the memory is free'd by readline */
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


// text should be in lower case
char* genie_keyword_generator(char* text, int state)
{
  static int list_index, len;
  static String help_string;
  const char *help;
  const char *name;
  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }
  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      help = genie_commands[list_index][COMMENT];
      cat(name, " ", help_string);
      cat(help_string, help, help_string);
      help_string.downcase();
      list_index++;
      if (help_string.contains(text))
      {
	cat(BLUE_ON, name, help_string);
        cat(help_string, BLUE_OFF, help_string);
        cat(help_string, " ", help_string);
        cat(help_string, help, help_string);
        return strcpy((char*)malloc(1+strlen(help_string)), help_string);	/* the memory is free'd by readline */
      }
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

char* genie_help_generator(char* text, int state)
{
  static int list_index, len;
  const char *name;
  const char *help;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      help = genie_commands[list_index][COMMENT];
      list_index++;

      if (strncasecmp (name, text, len) == 0)
      {
        return strcpy((char*)malloc(1+strlen(help)), help);		/* the memory is free'd by readline */
      }
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* 
 * Attempt to complete on the contents of TEXT.  START and END show the
 * region of TEXT that contains the word to complete.  We can use the
 * entire line in case we want to do some simple parsing.  Return the
 * array of matches, or NULL if there aren't any. 
 */
char** genie_command_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
	matches = completion_matches(line, genie_command_generator);
    }
    return matches;
}

char** genie_help_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
	matches = completion_matches(line, genie_help_generator);
    }
    return matches;
}

char** genie_keyword_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
        matches = completion_matches(line, genie_keyword_generator);
    }
    return matches;
}

int genie_keyword_search(int dummy1, int dummy2)
{
    static int count;
    static char buffer[132];
    static char* prompt = "(command-keyword-search) ";
    int c, done=0;
    int rl_save_point = rl_point, rl_save_end = rl_end;
    rl_modifying(0, rl_end);
    rl_extend_line_buffer(132);
    strcpy(rl_line_buffer, prompt);
    rl_point = rl_end = strlen(prompt);
    rl_redisplay(); 
    count = 0;
    while(!done)
    {
//	c = rl_read_key();
	c = rl_getc(rl_instream);
	switch(c)
	{
	    case RETURN:
	    case NEWLINE:
		done=1;
		break;

	    default:
		buffer[count] = tolower(c);
		count++;
		break;
	}
	fputc(c, rl_outstream);
	fflush(rl_outstream);
    }
    buffer[count] = '\0';
    char** matches = genie_keyword_completion(buffer, 0, count);
    int i = 0;
    if (matches != 0)
    {
    	while(matches[i] != 0)
    	{
	    fprintf(rl_outstream, "%s", matches[i]);
	    free(matches[i]);
	    i++;
	    crlf();
    	}
	free(matches);
    }
    else
    {
	crlf();
    }
    fflush(rl_outstream);
// Remove junk left over from previous line
    for(i=rl_save_end; i<rl_end; i++)
    {
	rl_line_buffer[i] = ' ';
    }
    rl_do_undo();
    rl_on_new_line();
    rl_point = rl_save_point;
    rl_end = rl_save_end;
    rl_redisplay(); 
    return 0;
}

/*
 * Print a list of genie commands followd by the "comment" field
 * from Smalltalk onto standard output. This procedure is invoked by the
 * <Ctrl>-h key on the command line, which is bound in initialize_readline()
 */
int genie_command_help(int dummy1, int dummy2)
{
    unsigned i, j, maxlen;
    int end = rl_point;
// Strip off any spaces at the end - we may have pressed ^H after completing the command
    while((end > 0) && whitespace(rl_line_buffer[end-1]))
	end--;
    
    char** matches_command = genie_command_completion(rl_line_buffer, 0, end);
    char** matches_help = genie_help_completion(rl_line_buffer, 0, end);
    if (matches_command != 0)
    {
	i = maxlen = 0;
// Calculate the max length of a completed command (used for "pretty printing")
	while(matches_command[i] != 0)
	{
	    if (strlen(matches_command[i]) > maxlen) 
	    { 
		maxlen = strlen(matches_command[i]); 
	    }
	    i++;
	}
// If we couldn't complete the command, first element of "matches" is the incomplete command - we don't want this printed
// However if we _did_ complete the command, the first element is what we want (and all we're given in fact!)
	if (i == 1)
	{
	    i = 0;
	}
	else
	{
	    i = 1;
	    free(matches_command[0]);		// The incomplete command we matched on
	    free(matches_help[0]);
	}
	maxlen += 4;		// add in number of blank spaces printed between command and help
	crlf();
	while(matches_command[i] != 0)
	{
	    fprintf(rl_outstream, "%s", matches_command[i]);
	    for(j=0; j<maxlen-strlen(matches_command[i]); j++)
	    {
		putc(' ', rl_outstream);
	    }
	    fprintf(rl_outstream, "%s", matches_help[i]);
	    free(matches_command[i]);
	    free(matches_help[i]);
	    crlf();
	    i++;
	}
	free(matches_command);
	free(matches_help);
	fflush(rl_outstream);
        rl_on_new_line();	// start a new prompt line
    }
    return 0;
}

// This is where we will set up options such as copmmand name completion
static void initialize_readline()
{
    /* Allow conditional parsing of the ~/.inputrc file. */
    rl_readline_name = (char*)"Genie";

    /* Tell the completer that we want a crack first. */
    rl_attempted_completion_function = (CPPFunction *)genie_command_completion;

    /* Register the <Ctrl>-h key as out help key */
    rl_add_defun("genie-command-help", (Function*)genie_command_help, CTRL('H'));

    /* Register the <Ctrl>-k key as out keyword search key */
    rl_add_defun("genie-keyword-search", (Function*)genie_keyword_search, CTRL('K'));

    return;
}

/* Strip whitespace from the start and end of STRING. */
static void stripwhite(char* string)
{
    int i = 0;

    while(whitespace(string[i]))
    {
        i++;
    }

    if (i)
    {
      strcpy(string, string + i);
    }

    i = strlen(string) - 1;

    while(i > 0 && whitespace(string[i]))
    {
        i--;
    }

    string[++i] = '\0';
    return;
}

