/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 * Module: 	read_ascii
 * Author:	Freddie Akeroyd, ISIS computer group
 * Purpose:	Read in columns of ascii numbers from an external file, placing each
 *		column into an array and the arrays into a workspace
 */
#include "geniedefs.h"
#include <stdlib.h>
#ifndef _WIN32
#	include <sys/time.h>
#else
#	include <time.h>
#endif
#include "genievars.h"
#include "smalltalk_object.h"
#include "read_ascii.h"
#include "Genie_functions.h"
#ifdef _WIN32
	char _itoa_buff[18];	/* _itoa can return 17 characters */
#	define dec(x) itoa(x, _itoa_buff, 10)	/* Seems to do the same thing !! */
#else
#	include <stream.h>
#endif /* _WIN32 */

static const char* rcsid = "@(#)$Id: read_ascii.cxx,v 1.14 1997/06/05 20:24:45 faa Exp $";
USE_RCSID(rcsid);


// What we can handle in the data file
enum ItemType { UnknownItem, IntegerItem, RealItem };

// Items spearated by "delim". Blank lines and those starting with "comment" are discarded
int read_ascii(const char* file, const String& field_names, GenieWorkspace& result, char comment, char delim)
{
    String message;
    fstream fs;
    fs.open(file, ios::in);
    if (!fs)
    {
        cat("Cannot open ", file, message);
        error_add("read_ascii", message, "");
        return 1;
    }
    fs.seekg(0, ios::beg);
    int nlines = lines_in_file(fs, comment, delim);
// "field_names" is a comma separated list of names to call the workspace fields 
#if defined(__DECCXX) || defined(_WIN32)
    const int MAXFIELDS(10);
#else
    const int MAXFIELDS(field_names.freq(',') + 1);
#endif
    String field_labels[MAXFIELDS];
    split(field_names, field_labels, MAXFIELDS, ',');
    ItemType field_type[MAXFIELDS];
    char buffer[SHORTCHARBUFF];
// Read first line
    if (skip_comments(fs, comment, delim))
    {
        fs.getline(buffer, sizeof(buffer), delim);
    }
    if (!fs.good())
    {
        error_add("read_ascii", "cannot read first line of file", "check file contains valid data lines");
        return 1;
    }
// Now parse line to see what type of data we have
    int i, j, nitems;
    Regex item_separator("[ \n\t,]+", 1), real_item("[\\.eE]", 1);
    String line(buffer), items[MAXFIELDS];
    void* result_array[MAXFIELDS];
    nitems = split(line, items, MAXFIELDS, item_separator);
    for(i=0; i<nitems; i++)
    {
	if (items[i].contains(real_item))
	{
	    field_type[i] = RealItem;
	    result_array[i] = new g_real[nlines];
 	}
	else
	{
	    field_type[i] = IntegerItem;
	    result_array[i] = new g_integer[nlines];
	}
    }
// Now go back to beginning and read whole file
    fs.seekg(0, ios::beg);
    g_real* gra;
    g_integer* gia;
    for(i=0; i<nlines; i++)
    {
        if (skip_comments(fs, comment, delim))
        {
            fs.getline(buffer, sizeof(buffer), delim);
        }
        if (fs.good())
	{
            nitems = split(buffer, items, MAXFIELDS, item_separator);
            for(j=0; j<nitems; j++)
	    {
	        switch(field_type[j])
	        {
	            case RealItem:
			gra = (g_real*)result_array[j];
		        gra[i] = atof(items[j]);
		        break;

	            case IntegerItem:
			gia = (g_integer*)result_array[j];
		        gia[i] = atol(items[j]);
		        break;

		    default:
		        error_add("read_ascii", "unknown item type in input file", "");
		        return 1;
    		        break;
  	        }
	    }
        }
    }
    for(j=0; j<nitems; j++)
    {
        switch(field_type[j])
        {
            case RealItem:
		gra = (g_real*)result_array[j];
    		result(field_labels[j], GenieRealArray(gra, nlines));
		break;

            case IntegerItem:
		gia = (g_integer*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(gia, nlines));
		break;

	    default:
	        error_add("read_ascii", "unknown item type in input file", "");
	        return 1;
    	        break;
	}
	delete []result_array[j];
    }
    fs.close();
    return 0;
}

// number of data (i.e. less comments and blank) lines in a file
// returns -1 on error, else number of lines
int lines_in_file(const char* file, char comment, char delim)
{
    String message;
    fstream ifs(file, ios::in);
    if (!ifs)
    {
	cat("Cannot open " , file, message);
 	error_add("LinesInFile", message, "Check file exists and is readable");
	return -1;
    }
    ifs.seekg(0, ios::beg);
    int nlines = lines_in_file(ifs, comment, delim);
    ifs.close();
    return nlines;
}

// lines from current position to end
int lines_in_file(fstream& ifs, char comment, char delim)
{
    int nlines(0);
    streampos mark = ifs.tellg();
    while(skip_comments(ifs, comment, delim))
    {
	ifs.ignore(50000, delim);
	nlines++;
    }
    ifs.clear();	// set stream status back to good
    ifs.seekg(mark);
    return nlines;
} 

// 0 = Error, 1 = OK
int skip_comments(fstream& input_stream, char comment, char delim)
{
    static int MAXLINE = 50000;		// just needs to be a big number
    int done = 0;
    int i = EOF;
    char c;
    while(input_stream.good() && !done)
    {
// Skip lines starting with a 'comment'
	if ((i = input_stream.peek()) == comment)
	{
	    input_stream.ignore(MAXLINE, delim);
	    continue;
	}
	if (i == EOF)
	    break;
// Now skip blank lines
	input_stream >> c;
	if (c == comment)
	{
	    input_stream.ignore(MAXLINE, delim);
	    continue;
	}
	if (!input_stream.good())
	    break;
	input_stream.putback(c);
	done = 1;
    }
    if ((i == EOF) || !input_stream.good() || !done)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}

#if 0
get_line(fstream& fs, char* buffer, int len_buffer, char comment, char delim)
{
    if (skip_comments(input_source, comment, delim))
    {
	fs.getline(buffer, len_buffer, delim);
    }
    if (!fs.good())
    {
        error_add("get_ascii", "cannot read line of file", "check file contains valid data lines");
        return 1;
    }
    return 0;
}

int read_ascii_stream(fstream& fs, GenieVariable& result, char comment, char delim)
{
    char buffer[SHORTCHARBUFF];
// First line is item type followed by dimensions information
    get_line(fs, buffer, sizeof(buffer), delim);
    String line(buffer), items[MAXFIELDS];
    Regex item_separator("[ \n\t,]+", 1);
    int nitems;
    nitems = split(buffer, items, MAXFIELDS, item_separator);
// [0] Item type
    if (!strcmp(items[0], "Realarray"))
    {
	error_add("read_ascii_stream", "Unknown input data type", "");
	return 1;
    }
// [1] length of dimensions array
    int ndims = atol(items[1]);
// We may be missing a few array data elements - this stops a big error
    if (nitems < ndims+2) { ndims = nitems - 2; }	
    int* dims_array = new int[ndims];
// [2+] contents of dimensions array
    for(i=0; i<ndims; i++)
    {
	dims_array[i] = atol(items[i+2]);
    }
    GenieRealArray gra;
    gra.alloc(dims_array);
    delete[] dims_array;
// Rest of lines are data
    for(i=0; i<gra.length(); i++)
    {
	
    }
    return 0;
}

// [ItemType] [ndims] [dims array] [data values]
int read_ascii_stream(fstream& fs, GenieVariable& result)
{
    int i;
    String type, message;
    fs >> type;
    if (!fs.good() || (type != "Realarray"))
    {
	cat("Unknown input data type: ", type, message);
	error_add("read_ascii_stream", message, "Sorry, Only \"Realarray\" supported at the moment");
	return 1;
    }
    int ndims(-1);
    fs >> ndims;
    if (!fs.good() || (ndims <= 0))
    {
	error_add("read_ascii_stream", "Invalid number of dimensions", "");
	return 1;
    }
    int* dims_array = new int[ndims];
    for(i=0; i<ndims; i++)
    {
	fs >> dims_array[i];
    }
    GenieRealArray gra;
    gra.alloc(dims_array, ndims);
    delete[] dims_array;
    for(i=0; i<gra.length(); i++)
    {
	fs >> gra[i];
    }
    result.become(gra);
    return 0;
}
#endif

// write out variables in ASCII columns
// Return 0 on success, 1 on error
int write_ascii(const char* file, const GenieString& comment, const GenieString& format, SmalltalkObject& gv)
{
    int i, j, gv_len;
    int ncolumns(0);
    String message;
    fstream fs;
    long field_width(0);		// default - no width limitation
    fs.open(file, ios::out);
    if (!fs)
    {
        cat("Cannot open ", file, message);
        error_add("write_ascii", message, "");
        return 1;
    }
    if (format.defined()) 
    {
	field_width = atol(format);
    }
    fs.flags(ios::showpoint);		// show decimal point on reals
    int len = INT_MAX;
    int nlines = INT_MAX;
// Determine number of lines - use shortest array length found
    gv_len = gv.arrayLength();
    GenieVariable** gv_list = new GenieVariable*[gv_len+1];
    GenieVariable* pgv;
    for(j=1; j<=gv_len; j++)
    {
	if (gv.arrayAt(j) != nilOOP)
	{
	    ++ncolumns;
	    pgv = new GenieVariable(gv.arrayAt(j));
	    switch(pgv->type())
	    {
		case GenieIntegerArrayType:
		    len = ((GenieIntegerArray*)*pgv)->length();
		    break;

		case GenieRealArrayType:
		    len = ((GenieRealArray*)*pgv)->length();
		    break;

		default:
		    len = 0;		// Indicates the type is not supported - trapped later
		    break;
	    }
	    if (len < nlines)
		nlines = len;
	    gv_list[j] = pgv;
	}
 	else
	{
	    gv_list[j] = 0;
	}
    }
    if (nlines == 0)
    {
	error_add("write_ascii", "Attempt to write out non-array or unsupported type", "");
	return 1;
    }
    if (comment.defined())
    {  
	char buffer[128];
        time_t clock;
	time(&clock);
	struct tm* tm_struct = localtime(&clock);
	if (strftime(buffer, sizeof(buffer), "%a %d %h at %H:%M:%S %Z", tm_struct) == 0)
        {
            strcpy(buffer, "<unknown>");
        }
	fs << "#\n";
//        fs << "# \"" << file << "\" created by Open GENIE V" << hex << __GENIE_VERSION_NO << " on " << buffer << "\n";
        fs << "# \"" << file << "\" created by Open GENIE on " << buffer << "\n";
	fs << "# This file contains " << nlines << " data line(s) and " << ncolumns << " column(s)\n";
	fs << "#\n";
// now user supplied comment - put a '#' in front of each line
   	fs << "# ";
  	const char* user_comment = comment;
	for(j=0; (size_t)j<strlen(user_comment); j++)
	{
	    fs.put(user_comment[j]);
	    if (user_comment[j] == '\n')
	 	fs << "# ";
	}
	fs << "\n#\n";
    }
    ginf << "WRITE/ASCII: Number of lines written = " << dec(nlines) << endl;
// Now write out data
    for(i=0; i<nlines; i++)
    {
	for(j=1; j<=gv_len; j++)
	{
	    pgv = gv_list[j];
	    if (pgv != 0)
	    {
	        switch(pgv->type())
	        {
		    case GenieIntegerArrayType:
			if (field_width != 0)
			    fs << setw(field_width) << (*(GenieIntegerArray*)*pgv)[i]; 
			else
			    fs << (*(GenieIntegerArray*)*pgv)[i] << " ";
		        break;

		    case GenieRealArrayType:
			if (field_width != 0)
			    fs << setw(field_width) << (*(GenieRealArray*)*pgv)[i];
			else
			    fs << (*(GenieRealArray*)*pgv)[i] << " ";
		        break;

		    default:
			fs << "<Invalid Array Type> ";		// should get caught earlier, so never executed
		        break;	
	        }
	    }
	}
	fs << "\n";
    }
    for(j=1; j<=gv_len; j++)
    {
	delete gv_list[j];
    }
    delete[] gv_list;
    return 0;
}
