/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         si_unit
 *      Author:         Freddie Akeroyd, ISIS, 1/3/96
 *      Purpose:        Allow user to define and manipulate new unit types
 */

#include "geniedefs.h"
#include <iostream.h>
#include <string.h>
#include <gnustring.h>
#include <math.h>
#include "si_unit.h"

#if defined(__ultrix) || (defined(__VMS) && __VMS_VER < 70000000)
#include "strcasecmp.h"
#endif

static const char* rcsid = "@(#)$Id: si_unit.cxx,v 1.8 1997/03/27 22:49:02 cmm Exp $";
USE_RCSID(rcsid);

// must null terminate the lists - it is used for searching 
const char* SIUnit::CodeName[] = { "", "metre", "gram", "second", "ampere", "kelvin", "mole", "candela", "radian",
        "steradian", "hertz", "newton", "pascal", "joule", "watt", "coulomb", "volt", "farad", "ohm",
	"siemens", "weber", "tesla", "henry", "degrees celsius", "lumen", "lux", "becquerel", "gray", "sievert", 0 };

// must null terminate the lists - it is used for searching 
const char* SIUnit::CodeAbbreviatedName[] = { "", "m", "g", "s", "A", "K", "Mol", "Cd", "rad", 
	"", "Hz", "N", "Pa", "J", "W", "C", "V", "F", "omega"
	"S", "W", "T", "H", "C", "", "", "", "", "", 0 };
    
// must null terminate the lists - it is used for searching 
const char* SIUnit::PrefixName[] = { "exa", "peta", "tera", "giga", "mega", "kilo", "hecto", "deca", "", "deci",
	"centi", "milli", "micro", "nano", "pico", "femto", "atto", 0 };

// must null terminate the lists - it is used for searching 
const char* SIUnit::PrefixAbbreviatedName[] = { "E", "P", "T", "G", "M", "K", "H", "D", "", "d", 
	"c", "m", "mu", "n", "p", "f", "a", 0 }; 

const int SIUnit::PrefixExponent[] = { 18, 15, 12, 9, 6, 3, 2, 1, 0, -1, -2, -3, -6, -9, -12, -15, -18 };

// What i must multiply data by to go from "m_code" to "s.m_code"
double SIUnit::convertTo(const SIUnit& s) const
{
    if (m_code != s.m_code)
    {
//	gerr << "Cannot convert " << codeName() << " to " << s.codeName() << endl;
	return 0.0;
    }
    else
    {
	return s.prefixValue() / prefixValue();
    }
}

String SIUnit::name(Code code, Prefix prefix)
{
    String res;
    cat(prefixName(prefix), codeName(code), res);
    return res;
}

SIDimensionalExponents SIUnit::dimensions(Code code)
{
    switch(code)
    {
	case metre:
	    return SIDimensionalExponents(1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

 	case gram:
	    return SIDimensionalExponents(0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0);

	case second:
	    return SIDimensionalExponents(0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0);

	case ampere:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0);

	case kelvin:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0);

	case mole:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

	case candela:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);

	case radian:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

	case steradian:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

	case hertz:
	    return SIDimensionalExponents(0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0);

	case newton:
	    return SIDimensionalExponents(1.0, 1.0, -2.0, 0.0, 0.0, 0.0, 0.0);

	// capital P here for MSVC++
	case Pascal:
	    return SIDimensionalExponents(-1.0, 1.0, -2.0, 0.0, 0.0, 0.0, 0.0);

	case joule:
	    return SIDimensionalExponents(2.0, 1.0, -2.0, 0.0, 0.0, 0.0, 0.0);

	case watt:
	    return SIDimensionalExponents(2.0, 1.0, -3.0, 0.0, 0.0, 0.0, 0.0);

	case coulomb:
	    return SIDimensionalExponents(0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0);

	case volt:
	    return SIDimensionalExponents(2.0, 1.0, -3.0, -1.0, 0.0, 0.0, 0.0);

	case farad:
	    return SIDimensionalExponents(-2.0, -1.0, 4.0, 1.0, 0.0, 0.0, 0.0);

	case ohm:
	    return SIDimensionalExponents(2.0, 1.0, -3.0, -2.0, 0.0, 0.0, 0.0);

	case siemens:
	    return SIDimensionalExponents(-2.0, -1.0, 3.0, 2.0, 0.0, 0.0, 0.0);

	case weber:
	    return SIDimensionalExponents(2.0, 1.0, -2.0, -1.0, 0.0, 0.0, 0.0);

	case tesla:
	    return SIDimensionalExponents(0.0, 1.0, -2.0, -1.0, 0.0, 0.0, 0.0);

	case henry:
	    return SIDimensionalExponents(2.0, 1.0, -2.0, -2.0, 0.0, 0.0, 0.0);

	case degrees_celsius:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0);

	case lumen:
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);

	case lux:
	    return SIDimensionalExponents(-2.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);

	case becquerel:
	    return SIDimensionalExponents(0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0);

	case gray:
	    return SIDimensionalExponents(2.0, 0.0, -2.0, 0.0, 0.0, 0.0, 0.0);

	case sievert:
	    return SIDimensionalExponents(2.0, 0.0, -2.0, 0.0, 0.0, 0.0, 0.0);

	default:
	    gerr << "Invalid SI unit, code = " << code << endl;
	    return SIDimensionalExponents(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
    }
}

const char* SIUnit::codeName(Code code)
{ 
    return CodeName[code]; 
}

SIUnit::Code SIUnit::lookupCode(const char* name)
{
    if (name == 0)
	return NoCode;

    for(int i=0; CodeName[i] != 0; i++)
    {
// Look up both full name and abbreviation
	if ( (strcasecmp(CodeName[i], name) == 0) || (strcasecmp(CodeAbbreviatedName[i], name) == 0) )
	{
	    return (Code)i;
	}
    }
    return NoCode;
}
        
SIUnit::Prefix SIUnit::lookupPrefix(const char* name)
{
    if (name == 0)
	return NoPrefix;

    for(int i=0; PrefixName[i] != 0; i++)
    {
	if ( (strcasecmp(PrefixName[i], name) == 0) || (strcasecmp(PrefixAbbreviatedName[i], name) == 0) )
	{
	    return (Prefix)i;
	}
    }
    return NoPrefix;
}
