/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         transform_utils
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Template routines for transforming data
 *
 */
#include "geniedefs.h"
#include <math.h>
#include <stdio.h>
#include "g_types.h"
#include "transform_utils.h"
#include "genievars.h"

static const char* rcsid = "@(#)$Id: transform_utils.cxx,v 1.13 1997/06/25 20:38:14 faa Exp $";
USE_RCSID(rcsid);

/* 
 * take a run number, and add leading zeros to return a 
 * pointer to the string
 *
 * returns NULL pointer on new failure
 */
char* crunid(g_integer id)
{
    char* ctemp = new char[6];
    if (ctemp != 0)
    {
	(void)sprintf(ctemp, "%05li", (long)id);
    }
    return(ctemp);
}

/*
 * xbounds[]    contains the new boundaries
 * xsteps[]     contains the bin size to use within the limits deined
 *              by xbounds[]
 *
 * e.g. step xstep[0] between xbounds[0] and xbounds[1]
 *      step xstep[1] between xbounds[1] and xbounds[2]
 *
 * if assign=0, does not assign to xnew
 *    assign=1, assigns results to previously allocated array xnew
 *
 * if itype = 1, constant bin sizes
 *          = 2, logarithmic
 *
 * returns number of elements in xnew[], -1 if error
 */

template <class __g_real, class __g_integer>
int newAxis(__g_real xbounds[], __g_integer ibounds, __g_real xsteps[], 
              __g_integer isteps, __g_real xnew[], int assign,
	      __g_integer itype)
{
    __g_real xcurr, xs;
    __g_integer i(1), j(0), inew(1);
    if ( assign && xnew == NULL )
    {
	return -1;
    }
    xcurr = xbounds[0];
    if ( assign )
    {
	xnew[0] = xcurr;
    }
    while( (i < ibounds) && (j < isteps) )
    {
	if ( itype == 1 )
	{
            xs = xsteps[j];
	}
	else if ( itype == 2 )
	{
	    xs = xcurr * xsteps[j];
	}
	else
	{
	    return -1;		/* invalid itype specified */
	}
/* continue stepping unless we get to almost where we want to */
	if ( (xcurr + xs) < (xbounds[i] - (xs * 1.E-6)) )
	{
	    xcurr += xs;
	}
	else
	{
	    xcurr = xbounds[i];
	    i++;
	    j++;
	}
	if ( assign )
	{
	    xnew[inew] = xcurr;
	}
	inew++;
    }
    return( (i == ibounds) && (j == isteps) ? inew : -1 );
}


/*
 * 
 * returns 0 on success, 1 on error
 *
 */

template <class __g_real, class __g_integer>
int rebin(__g_real xold[], __g_real yold[], __g_real eold[], __g_integer ntcold, 
	  __g_real xnew[], __g_real ynew[], __g_real enew[], __g_integer ntcnew, __g_real& ymax)
// ymax no longer used
{
    __g_integer i,iold(0),inew(0);
    __g_real xo_low, xo_high, xn_low, xn_high, delta, width;
    for(i=0; i<ntcnew; i++)
    {
	ynew[i] = 0.0;
    }
    memcpy(enew, ynew, ntcnew*sizeof(__g_real));
    while((inew < ntcnew) && (iold < ntcold))
    {
	xo_low = xold[iold];
	xo_high = xold[iold+1];
	xn_low = xnew[inew];
	xn_high = xnew[inew+1];
	if ( xn_high <= xo_low )
	{
	    inew++;		/* old and new bins do not overlap */
	}
	else if ( xo_high <= xn_low )
	{
	    iold++;		/* old and new bins do not overlap */
	}
	else
	{
/*
 *        delta is the overlap of the bins on the x axis
 */
	    delta = min(xo_high, xn_high) - max(xo_low, xn_low);
	    width = xo_high - xo_low;
	    if ( (delta <= 0.0) || (width <= 0.0) )
	    {
		return 1;	/* something wrong */
	    }
/*
 *        yold contains counts/unit time, ynew is just counts at moment
 *	  enew contains counts**2
 */
	    if ( GenieReal::isDefined(yold[iold]) && GenieReal::isDefined(ynew[inew]) )
	    {
	    	ynew[inew] += yold[iold]*delta;
	        enew[inew] += eold[iold]*eold[iold]*delta*width;
	    }
	    else
	    {
		GenieReal::undefine(ynew[inew]);
		GenieReal::undefine(enew[inew]);
	    }
	    if ( xn_high > xo_high )
	    {
		iold++;
	    }
	    else
	    {
		inew++;
	    }
	}
    }
/*
 * convert back to counts/unit time
 */
    ymax = ynew[0];
    for(i=0; i<ntcnew; i++)
    {
	if (GenieReal::isDefined(ynew[i]))
	{
	    width = xnew[i+1]-xnew[i];
	    ynew[i] /= width;
	    enew[i] = (__g_real)sqrt((double)enew[i]) / width;
	    ymax = max(ymax, ynew[i]);
	}
    }
    return 0;
}

/*
 * returns NTCNEW on success, -1 on failure
 *
 * may safely be called with xarray and xnew the same etc.
 */


template <class __g_real_old, class __g_real_new, class __g_integer>
int group_bins(__g_real_old xarray[], __g_real_old yarray[], __g_real_old earray[],
	       __g_integer ntc, __g_integer nspec, __g_real_new xnew[], __g_real_new ynew[],
	       __g_real_new enew[], __g_integer nbin)
{
    __g_integer iold, inew, igroup, i, j, ntcnew(0);
    double step, ytemp, etemp;
// Note: ntcnew is incorrect on first pass of loop, but this doesn't matter
// as j is 0
    for(j=0; j<nspec; j++)
    {    
    iold = inew = 0;
    igroup = nbin;
    while( (igroup == nbin) && (iold < ntc) )
    {
	igroup = ( iold + nbin > ntc ? ntc - iold : nbin );
	etemp = ytemp = 0.0;
	for(i=iold; i < iold+igroup; i++)
	{
	    step = xarray[i+1] - xarray[i];
	    ytemp += yarray[i+j*ntc] * step;
	    etemp += earray[i+j*ntc] * earray[i+j*ntc] * step * step;
	}
	step = xarray[iold+igroup] - xarray[iold];
	ynew[inew+j*ntcnew] = (ytemp / step);
	enew[inew+j*ntcnew] = (sqrt(etemp) / step);
	xnew[inew] = xarray[iold];
	iold += igroup;
	inew++;
    }
    ntcnew = inew;
    xnew[inew] = xarray[iold];	
    }
    return ntcnew;			/* new NTC value */
}


/*
 * return -1 on error (e.g. xval not in range)
 * returns i such that xarray[i] <= xval < xarray[i+1] unless xval=x[np-1] when returns i=np-2
 */
template <class __g_real, class __g_integer>
int bracket(__g_real xval, const __g_real xarray[], __g_integer np)
{
    int ilow(0), ihigh(np), imid;
    if ( xval > xarray[np-1] || xval < xarray[0] )
    {
	return(-1);
    }
    if ( xval == xarray[np-1] )
    {
	return(np-2);
    }
    while( (ihigh-ilow) > 1 )
    {
	imid = (ilow+ihigh) / 2;
	(xval < xarray[imid]) ? (ihigh = imid) : (ilow = imid);
    }
    return(ilow);
}

/*
 * returns 0 on success, 1 on a rebin error
 * xnew is set to be xold, but clipped by (xmin, xmax)
 */
template <class __g_real, class __g_integer>
int rebinRange(__g_real xold[], __g_real yold[], __g_real eold[], 
	       __g_real xnew[], __g_real ynew[], __g_real enew[],
		__g_integer& ntc, __g_real xmin, __g_real xmax, __g_real& ymax)
{
    __g_real diff1, diff2, xtol(1E-8);
    __g_integer i, j(1);

    xnew[0] = xmin;

    for(i=0; i<ntc+1; i++)
    {
	diff1 = xold[i] - xmin;
	diff2 = xmax - xold[i];
	if ( (diff1 > xtol) && (diff2 > xtol) )
	{
	    xnew[j] = xold[i];
	    j++;
	}
    }
    xnew[j] = xmax;
    if ( rebin(xold, yold, eold, ntc, xnew, ynew, enew, j, ymax) != 0 )
    {
	return 1;
    }

    ntc = j;
    return 0;
}

/*
 * integarte a 1D spectrum array between xmin and xmax 
 * ntc is length of yarray (length of xarray will be ntc+1)
 * return 0 on success, 1 on error
 */
template <class __g_real, class __g_integer>
int integrateSpectrum(const __g_real xarray[], const __g_real yarray[], const __g_real earray[],
			__g_integer ntc, __g_real xmin, __g_real xmax, 
			__g_real& sum, __g_real& error)
{
    double shift, dtemp;
    int imax, imin, i;
    sum = error = 0.0;
// Determine which bins the given range covers
    if ( (imin = bracket(xmin, xarray, ntc+1)) == -1 )
    {
	gerr << "INTEGRATE: xmin value out of range" << endl;
	return 1;
    }
    if ( (imax = bracket(xmax, xarray, ntc+1)) == -1 )
    {
	gerr << "INTEGRATE: xmax value out of range" << endl;
	return 1;
    }
// First add up complete bins
    for(i=imin+1; i<imax; i++)
    {
        dtemp = xarray[i+1] - xarray[i];
        sum += yarray[i] * dtemp;
        error += earray[i] * earray[i] * dtemp * dtemp;
    }
// Now the partial bins at the lower end
	i=imin;
	dtemp = xarray[i+1] - xarray[i];
	shift = xarray[imin+1] - (double)xmin;
	if ( shift < 0.0 )
	{
	    gerr << "INTEGRATE: negative shift value" << endl;
	    return 1;
	}
	sum += yarray[i] * shift;
	error += earray[i] * dtemp * shift;
// Ditto at the higher end
	i=imax;
	dtemp = xarray[i+1] - xarray[i];
	shift = xmax - xarray[imax];
	if ( shift < 0.0 )
	{
	    gerr << "INTEGRATE: negative shift value" << endl;
	    return 1;
	}
	sum += yarray[i] * shift;
	error += earray[i] * dtemp * shift;
	error = sqrt(error);
    return 0;
}

/*
 * integarte several 1D spectra in a 2D array between xmin and xmax 
 * ntc is length of yarray (length of xarray will be ntc+1)
 * nspec is the number of spectra
 * return 0 on success, 1 on error
 */
template <class __g_real, class __g_integer>
int integrateSpectra(__g_real xarray[], __g_real yarray[], __g_real earray[],
			__g_integer ntc, __g_integer nspec, __g_real xmin[], 
			__g_integer nmin, __g_real xmax[], __g_integer nmax, 
			__g_real sum[], __g_real error[])
{
    int status(0), imin, imax;
    for(int i=0; (i < nspec) && (status == 0); i++)
    {
	imin = i % nmin;
	imax = i % nmax;
	status = integrateSpectrum(xarray, yarray + i*ntc, earray + i*ntc,
				ntc, xmin[imin], xmax[imax], sum[i], error[i]);
    }
    return status;
}
	
// Instantiate the templates
#ifdef __DECCXX
#pragma define_template group_bins<float, float, int>
#pragma define_template group_bins<double, float, long>
#pragma define_template group_bins<double, double, long>
#pragma define_template newAxis<float, int>
#pragma define_template newAxis<double, long>
#pragma define_template bracket<float, int>
#pragma define_template bracket<double, long>
//#pragma define_template rebinRange<float, int>
#pragma define_template rebinRange<double, long>
//#pragma define_template rebin<float, int>
#pragma define_template rebin<double, long>
#pragma define_template integrateSpectrum<double, long>
#pragma define_template integrateSpectra<double, long>
#endif /* __DECCXX */

#ifdef __GNUG__
template int bracket(float, const float[], int);
template int bracket(double, const double[], long);
template int newAxis(float[], int, float[], int, float[], int, int);
template int newAxis(double[], long, double[], long, double[], int, long);
//template int rebin(float[], float[], float[], int, float[], float[], float[], int, float&);
template int rebin(double[], double[], double[], long, double[], double[], double[], long, double&);
template int rebinRange(double[], double[], double[], double[], double[], double[], long&, double, double, double&);
//template int rebinRange(float[], float[], float[], float[], float[], float[], int&, float, float, float&);
template int group_bins(double[], double[], double[], long, long, float[], float[], float[], long);
template int group_bins(double[], double[], double[], long, long, double[], double[], double[], long);
template int group_bins(float[], float[], float[], int, int, float[], float[], float[], int);
template double max(double, double);
template float max(float, float);
template int max(long, long);
template int max(int, int);
template double min(double, double);
template float min(float, float);
template int min(long, long);
template int min(int, int);
template int integrateSpectrum(const double xarray[], const double yarray[], const double earray[],
			long ntc, double xmin, double xmax, 
			double& sum, double& error);
template int integrateSpectra(double xarray[], double yarray[], double earray[],
			long ntc, long nspec, double xmin[], long nmin, 
			double xmax[], long nmax, double sum[], double error[]);
#endif /* __GNUG__ */
