/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         units
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Routines for changing the units of spectra
 *
 */
#include "geniedefs.h"
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "genie_types.h"
#include "units.h"
#include "units_p.h"

#define	MAX(__a, __b)	((__a) < (__b) ? (__b) : (__a))

#define DEFINED(__a)	((__a) != DBL_MAX)	/* to handle undefined real genie variables */

/*
 * error codes:
 *-1 = points truncated in array
 * 0 = success
 * 1 = input unit code invalid
 * 2 = output unit code invalid
 * 3 = inelastic, not specified emode/efixed properly
 * 4 = input unit is channel number --- cannot convert
 * 5 = invalid l1, l2 or TwoTheta
 */

static const char* rcsid = "@(#)$Id: units.cxx,v 1.10 1997/05/06 09:57:17 cmm Exp $";
USE_RCSID(rcsid);

int convert_units(g_real xarray[], g_real yarray[], g_real earray[], 
		  g_integer *ntcp, const char* xunit_in, const char** xunit_out, 
		  const char* yunit_in, const char** yunit_out, g_integer unit_convert,
		  g_real l1, g_real l2,
		  g_integer emode, g_real efixed, g_real TwoTheta, g_real *ymax, g_real emin, 
		  g_real emax, g_real delta)
{
    g_integer ntc;
    int i, j, error_code = 0;
    g_real factor, FourPi, SinTheta, ltot, step, t1, e2, t2;
    unit_code_struct xunit_code_in, xunit_code_out, yunit_code_in,
		yunit_code_out;
    ntc = *ntcp;
/* only y can be dimensionless */
    if ( (xunit_code_in = (unit_code_struct) unit_code(xunit_in, 'x')) <= 0)
    {
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
    }
    if ( xunit_code_in == 1 )
    {
	error_code = 4;
	gerr << "UNITS Error: cannot change units from channel number" << endl;
	return(error_code);		
    }
    if ( (yunit_code_in = (unit_code_struct) unit_code(yunit_in, 'y')) == -1)
    {
	error_code = 1;
	gerr << "UNITS Error: invalid yunit_in = " << yunit_in << endl;
	return(error_code);		/* invalid unit_in */
    }
    if ( unit_convert < 1 || unit_convert >= NUNIT )
    {
	error_code = 2;
	gerr << "UNITS Error: invalid output unit code = " << unit_convert << endl;
	return(error_code);
    }
    xunit_code_out = (unit_code_struct) unit_convert;
    yunit_code_out = ((yunit_code_in == 0) ? (unit_code_struct)0 : xunit_code_out);
/* 
 * deal with some easy options 
 */
    if ( xunit_code_in == xunit_code_out )
    {
	*xunit_out = xunit_in;
	*yunit_out = yunit_in;
	sprintf(error_message_buffer, "\n");
	return(error_code);
    }

/*
 * convert y and e to counts if not dimensionless
 */
    if ( yunit_code_in != 0 )
    {
	for(i=0; i<ntc; i++)
    	{
	    step = xarray[i+1] - xarray[i];
	    if DEFINED(yarray[i]) ( yarray[i] = yarray[i] * step );
	    if DEFINED(earray[i]) ( earray[i] = earray[i] * step );
        }
    }
    ltot = l1 + l2;
    SinTheta = sin(fabs(TwoTheta) * DEG2RAD / 2.0);
    FourPi = 4.0 * PI;
    if ( ltot == 0.0 )
    {
	error_code = 5;
	gerr << "UNITS Error: total flight path L = 0" << endl;
	return(error_code);
    }
//    if ( SinTheta == 0.0 )   // Move to later
//    {
//	error_code = 5;
//	sprintf(error_message_buffer, "Error: SinTheta = 0\n");
//	return(error_code);
//    }
    if (( xunit_code_in >= 9 || xunit_code_out >= 9 ) && ( emode <= 0 || emode >= 3 || efixed <= 0.0 ))
    {
	error_code = 3;
	gerr << "UNITS Error: either emode (" << emode << ") or efixed (" << efixed << ") wrongly specified" << endl;
	return(error_code);
    }

/* 
 * now convert units to a common base of TOF in microseconds 
 * C1 and C2 defined in units.h
 */

    switch(xunit_code_in) {

    case Dimensionless:
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
	break;

    case ChannelNo:
	error_code = 4;
	gerr << "UNITS Error: cannot change units from channel number" << endl;
	return(error_code);
	break;

    case TOF:
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] - delta;
	}
	break;

    case Wavelength:
	factor = ltot * C1;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] * factor;
	}
	break;

    case dSpacing:
	factor = 2.0 * ltot * SinTheta * C1;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] * factor;
	}
	break;

    case Energy:
	factor = C1 * ltot * sqrt(C2);
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = factor / sqrt(xarray[i]);
	}
	break;

    case MomentumTransfer:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = factor / xarray[i];
	}
	break;

    case Qsquared:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = factor / sqrt(xarray[i]);
	}
	break;

    case ReciprocalTOF:
	factor = ltot;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = factor * xarray[i];
	}
	break;

    case DeltaE:
	energyToTOF(1.0, emode, efixed, xarray, ntc, l1, l2);
	break;

    case DeltaE_inWaveNumber:
	energyToTOF(MEV2WAVENUMBER, emode, efixed, xarray, ntc, l1, l2);
	break;

    case PrimaryWavelength:
	if ( emode == 2 )
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = l1 * C1;
	    for(i=0; i<ntc+1; i++)
	    {
	    	t1 = factor * xarray[i];
		xarray[i] = t1 + t2;
	    }
	}
	break;

    default:
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
	break;
    }

// Now checked in individual commands - caused problems on instruments with "funny" hold off values
#if 0
/*
 * TOF cannot be -ve
 */
    for(i=0; i<ntc+1; i++)
    {
	if (xarray[i] <= 0.0)
	{
	    error_code = 1;
	    gerr << "UNITS Error: zero or negative time-of-flight with x[" << i+1 << "] = " << xarray[i] << endl;
	    gerr << "             check that the \"hold off\" (DELTA) value of " << delta << " is appropriate" << endl;
	    return(error_code);		
	}
    }
#endif

/*
 * Now convert TOF to required units
 */
    switch(xunit_code_out) {

    case Dimensionless:
	error_code = 3;
	gerr << "UNITS Error: converting to dimensionless X" << endl;
	return(error_code);
	break;

    case ChannelNo:
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = (g_real)i + 1.0;
	}
	break;

    case TOF:
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] + delta;
	}
	break;

    case Wavelength:
	factor = ltot * C1;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] / factor;
	}
	break;

    case dSpacing:
	factor = 2.0 * ltot * SinTheta * C1;
    	if (factor == 0.0)
    	{
	    error_code = 5;
	    gerr << "UNITS Error: L1+L2 or Sin(TwoTheta) is 0.0" << endl;
	    return(error_code);
    	}
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] / factor;
	}
	break;

    case Energy:
	factor = C1 * C1 * ltot * ltot * C2;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = factor / (xarray[i] * xarray[i]);
	}
	break;

    case MomentumTransfer:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc+1; i++)
	{
	    if (xarray[i] == 0.0)
		xarray[i] = (factor > 0.0 ? BIGVAL : -BIGVAL);
	    else
	        xarray[i] = factor / xarray[i];
	}
	break;

    case Qsquared:
	factor = FourPi * C1 * ltot * SinTheta;
	factor = factor * factor;
	for(i=0; i<ntc+1; i++)
	{
	    if (xarray[i] == 0.0)
		xarray[i] = BIGVAL;
	    else
	        xarray[i] = factor / (xarray[i] * xarray[i]);
	}
	break;

    case ReciprocalTOF:
	factor = ltot;
	for(i=0; i<ntc+1; i++)
	{
	    xarray[i] = xarray[i] / factor;
	}
	break;

    case DeltaE:
	TOFToEnergy(1.0, emode, efixed, xarray, ntc, l1, l2);
	break;

    case DeltaE_inWaveNumber:
	TOFToEnergy(MEV2WAVENUMBER, emode, efixed, xarray, ntc, l1, l2);
	break;


    case PrimaryWavelength:
	if ( emode == 2 )
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = l1 * C1;
	    for(i=0; i<ntc+1; i++)
	    {
		t1 = xarray[i] - t2;
		xarray[i] = (( t1 <= 0.0 ) ? ( -BIGVAL ) : ( t1 / factor ));
	    }
	}
	break;

    default:
	error_code = 2;
	gerr << "UNITS Error: invalid output unit code = "  << xunit_code_out << endl;
	return(error_code);		
	break;
    }
/*
 * flip order 
 */
    if (xarray[0] > xarray[ntc])
    {
	for(i=0; i < ntc/2; i++)
	{
	    j = ntc - i - 1;

	    factor = yarray[i];
	    yarray[i] = yarray[j];
	    yarray[j] = factor;

	    factor = earray[i];
	    earray[i] = earray[j];
	    earray[j] = factor;
	}
//	yarray[ntc] = 0.0;	/* not needed now */
//	earray[ntc] = 0.0;
	for(i=0; i < (ntc+1)/2; i++)
	{
	    j = ntc - i ;
	    factor = xarray[i];
	    xarray[i] = xarray[j];
	    xarray[j] = factor;
	}
    }

    if ( xunit_code_out == DeltaE ) 	//		|| xunit_code_out == PrimaryWavelength )
    {
	ntc = truncate(xarray, yarray, earray, ntc+1, emin, emax) - 1;
    }

/*
 * now scale back y and e if not dimensionless
 */
    if (yunit_code_out != 0)
    {
    	for(i=0; i<ntc; i++)
    	{
	    step = xarray[i+1] - xarray[i];
	    if DEFINED(yarray[i]) (yarray[i] = yarray[i] / step);
	    if DEFINED(earray[i]) (earray[i] = earray[i] / step);
        }
    }

    *ymax = 0.0;
    for(i=0; i<ntc; i++)
    {
	if ( DEFINED(yarray[i]) && (yarray[i] > *ymax) )
    	{
	    *ymax = yarray[i];
    	}
    }

    *xunit_out = unit_name( (int)xunit_code_out, 'x');
    *yunit_out = unit_name( (int)yunit_code_out, 'y');

    if ( (xunit_out == NULL) || (yunit_out == NULL) )
    {
	gerr << "UNITS Error: xunit_out or yuit_out is NULL" << endl;
	error_code = 2;
    }
    *ntcp = ntc;
    return(error_code);
}


/* returns -1 on error */
int unit_code(const char* cunit, char axis)
{
    int i;
    if ( axis == 'x' )
    {
    	for(i=0; i<NUNIT; i++)
    	{
	    if (strcmp(cunit, xunit_names[i]) == 0)
	    {
	        break;
	    }
    	}
    }
    else if ( axis == 'y' )
    {
    	for(i=0; i<NUNIT; i++)
    	{
	    if (strcmp(cunit, yunit_names[i]) == 0)
	    {
	        break;
	    }
    	}
    }
    else
    {
	i = NUNIT;
    }


    if ( i == NUNIT  )
    {
	return(-1);		/* invalid unit string */
    }
    else
    {
	return(i);
    }
}


/* returns NULL pointer on error */
const char* unit_name(int iunit, char axis)
{
    static const char* invalid_unit = "Invalid Unit";
    if ( (iunit < 0) || (iunit >= NUNIT) )
    {
	return(invalid_unit);		/* invalid unit code */
    }
    else if ( axis == 'x' )
    {
	return(xunit_names[iunit]);
    }
    else if ( axis == 'y' )
    {
	return(yunit_names[iunit]);
    }
    else
    {
	return(invalid_unit);
    }

}

char* convert_units_errmsg(void)
{
    return(error_message_buffer);
}

/* remove all points outside range */

g_integer truncate(g_real xarray[], g_real yarray[], g_real earray[], g_integer np, g_real xmin, g_real xmax)
{
    g_integer i,j=0;
    for(i=0; i<np; i++)
    {
	if (xarray[i] >= xmin && xarray[i] <= xmax)
	{
	    xarray[j] = xarray[i];
	    yarray[j] = yarray[i];
	    earray[j] = earray[i];
	    j++;
	}
    }
    return(j);	/* new np value */
}


/*
 * 0 success 
 * 1 invalid emode
 */
int energyToTOF(g_real scale, g_integer emode, g_real efixed, g_real xarray[], 
		g_integer ntc, g_real l1, g_real l2)
{
    g_real factor, e1, e2, t1, t2;
    g_integer i;
	if (emode == 1)
	{
	    e1 = efixed;
	    t1 = C1 * l1 * sqrt(C2/e1);
	    factor = C1 * l2 * sqrt(C2);
	    for(i=0; i<ntc+1; i++)
	    {
		e2 = e1 - xarray[i] / scale;	// xarray[i] / scale --> from whatever to meV
		t2 = factor / sqrt(e2);
		xarray[i] = t1 + t2;
	    }
	}
	else if (emode == 2)
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = C1 * l1 * sqrt(C2); 
	    for(i=0; i<ntc+1; i++)
	    {
		e1 = e2 + xarray[i] / scale;	// xarray[i] / scale --> from whatever to meV
		t1 = factor / sqrt(e1);
		xarray[i] = t1 + t2;
	    }
	}
	else
	{
	    return 1;
	}
	return 0;
}

/*
 * 0 success
 * 1 invalid emode
 */
int TOFToEnergy(g_real scale, g_integer emode, g_real efixed, g_real xarray[], 
		g_integer ntc, g_real l1, g_real l2)
{
    g_real factor, e1, e2, t1, t2;
    g_integer i;
	if (emode == 1)
	{
	    e1 = efixed;
	    t1 = C1 * l1 * sqrt(C2/e1);
	    factor = C1 * C1 * l2 * l2 * C2;
	    for(i=0; i<ntc+1; i++)
	    {
		t2 = xarray[i] - t1;
		if ( t2 <= 0.0 )
		{
		    xarray[i] = -BIGVAL;
		}
		else
		{
		    e2 = factor / (t2 * t2);
		    xarray[i] = (e1 - e2) * scale;	// e1 * scale --> from meV to whatever
		}
	    }
	}
	else if (emode == 2)
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = C1 * C1 * l1 * l1 * C2;
	    for(i=0; i<ntc+1; i++)
	    {
		t1 = xarray[i] - t2;
		if ( t1 <= 0.0 )
		{
		    xarray[i] = BIGVAL;
		}
		else
		{
		    e1 = factor / (t1 * t1);
		    xarray[i] = (e1 - e2) * scale;	// e1 * scale --> from meV to whatever
		}
	    }
	}
        else 
	{
	    return 1;
	}
    return 0;
}
